#include "TCPSockets.h"

#include <app/Message.h>
#include <app/AppDefs.h>
#include <kernel/OS.h>

#include <errno.h>
#include <net/netdb.h>

#define TCP_LOOPER_PRIORITY B_LOW_PRIORITY

TCPLooper::TCPLooper(): BLooper("TCPLooper", TCP_LOOPER_PRIORITY)
{
	socket_read = 0;
	tcp_socket = new TCPSocket;
}

TCPLooper::~TCPLooper()
{
	if (socket_read)
		kill_thread(socket_read);
	if (tcp_socket)
		delete tcp_socket;
}

void TCPLooper::MessageReceived(BMessage *msg)
{
	switch (msg->what) {
	case TCP_INIT: {
		if (msg->FindPointer(HANDLER_ARG, &msg_handler) != B_OK) {
			BMessage reply(TCP_ERR);
			reply.AddInt8(ERROR_ARG, NO_LOOPER_MSG);
			msg->SendReply(&reply);
			return;
		}
		const char *hostname;
		if (msg->FindString(HOST_ARG, &hostname) != B_OK) {
			// What do you want?
			BMessage reply(TCP_ERR);
			reply.AddInt8(ERROR_ARG, NO_HOST_MSG);
			msg->SendReply(&reply);
			return;
		}
		const char *port;
		if (msg->FindString(PORT_ARG, &port) != B_OK) {
			// Information.
			BMessage reply(TCP_ERR);
			reply.AddInt8(ERROR_ARG, NO_PORT_MSG);
			msg->SendReply(&reply);
			return;
		}
		int tcp_err;
		if (tcp_err = tcp_socket->init(hostname, port)) {
			// You won't get it.
			BMessage reply(TCP_ERR);
			reply.AddInt8(ERROR_ARG, tcp_err);
			msg->SendReply(&reply);
			return;
		}
		if ((socket_read = spawn_thread(read_thread_entry, "Socket Reader",
										 TCP_LOOPER_PRIORITY, this)) < 0) {
			// By hook or by crook, we will.
			BMessage reply(TCP_ERR);
			reply.AddInt8(ERROR_ARG, NO_THREAD_MSG);
			msg->SendReply(&reply);
			return;
		}
		resume_thread(socket_read);

		// Now do the layer 4 setup.
		InitConnection();

		// I am not a number, I'm a free man.
		msg->SendReply((uint32)B_OK);
		msg_handler->Looper()->PostMessage(TCP_CONNECTED, msg_handler);
	}; break;
	case TCP_SENDDATA: {
		int32 nbytes;
		void *dataptr;
		if (msg->FindData(DATA_ARG, B_RAW_TYPE, &dataptr, &nbytes) != B_OK)
			return;
		void *data = malloc(nbytes);
		if (!data) {
			// TODO: use PostMessage to the handler we belong to.
			BMessage reply(B_ERROR);
			reply.AddInt8(ERROR_ARG, NO_MEM_MSG);
			msg->SendReply(&reply);
			return;
		}
		memcpy(data, dataptr, nbytes);
		SendRemote(data, nbytes);
	}; break;
	case TCP_RECVDATA: {
		int32 nbytes;
		void *dataptr;
		if (msg->FindData(DATA_ARG, B_RAW_TYPE, &dataptr, &nbytes) != B_OK)
			return;
		void *data = malloc(nbytes);
		if (!data) {
			// TODO: use PostMessage to the handler we belong to.
			BMessage reply(B_ERROR);
			reply.AddInt8(ERROR_ARG, NO_MEM_MSG);
			msg->SendReply(&reply);
			return;
		}
		memcpy(data, dataptr, nbytes);
		SendLocal(data, nbytes);
	}; break;
	default:
		BLooper::MessageReceived(msg);
	}
}


void TCPLooper::InitConnection()
{
}


void TCPLooper::SendLocal(void *data, int32 nbytes)
{
	BMessage send_msg(TCP_RECVDATA);
	send_msg.AddData(DATA_ARG, B_RAW_TYPE, data, nbytes);
	msg_handler->Looper()->PostMessage(&send_msg, msg_handler);
}

void TCPLooper::SendRemote(void *data, int32 nbytes)
{
	if (tcp_socket->write((char *)data, nbytes) < 0) {
		if (!msg_handler->Looper()) {
			fprintf(stderr, "DEBUG: Oh my! Looper() is NULL.\n");
			return;
		}
		msg_handler->Looper()->PostMessage(TCP_NOCONN, msg_handler);
	}
}

long TCPLooper::read_thread()
{
	int32 nbytes;
	char data[2048];

	while ((nbytes = tcp_socket->read(data, sizeof(data))) > 0) {
		BMessage msg(TCP_RECVDATA);
		msg.AddData(DATA_ARG, B_RAW_TYPE, data, nbytes);
		PostMessage(&msg);
	}

	// Oops, the read died before we did. Inform the relative parties that we've
	// kicked the bucket.
	msg_handler->Looper()->PostMessage(TCP_NOCONN, msg_handler);

	return 0;
}

long TCPLooper::read_thread_entry(void *arg)
{
	TCPLooper *socket = (TCPLooper *)arg;
	return socket->read_thread();
}

Socket::Socket()
{
	status = CLOSED;
}

Socket::~Socket()
{
	if (status == OPEN)
		closesocket(fd);
}

tcp_status_t Socket::get_status()
{
	return status;
}

char *Socket::get_hostname()
{
	return hostname;
}

struct hostent *Socket::get_hostent()
{
	return host_info;
}

TCPSocket::TCPSocket(): Socket()
{
}

int TCPSocket::init(const char *dest_host, const char *dest_port)
{
	strncpy(hostname, dest_host, sizeof(hostname));
	char *foo;
	int port_int = strtol(dest_port, &foo, NULL);
	
	if (*foo) {
		service_info = getservbyname(dest_port, "tcp");
		port_int = service_info->s_port;
	}
	
	host_socket.sin_family = AF_INET;
	host_socket.sin_port = htons(port_int);
	if (!(host_info = gethostbyname(dest_host)))
		return NO_HOST;

	if ((fd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
		return NO_SOCKET;

	bcopy(host_info->h_addr, &host_socket.sin_addr, host_info->h_length);
	
	if (connect(fd, (struct sockaddr *) &host_socket, sizeof(host_socket)) < 0)
		return NO_CONNECT;

	status = OPEN;
	
	return 0;
}

int32 TCPSocket::read(char *buf, int32 buf_len)
{
	fd_set read_fd_set;
	struct timeval tv = { 0, 0 };
	
	FD_ZERO(&read_fd_set);
	FD_SET(fd, &read_fd_set);

	int32 read_bytes;
	while ((read_bytes = recv(fd, buf, buf_len, 0)) < 0 && errno == EINTR);
	if (read_bytes < 0)
		status = CLOSED;
	return read_bytes;
}

int32 TCPSocket::write(char *buf, int32 buf_len)
{
	int32 wrote = send(fd, buf, buf_len, 0);
	if (wrote < 0)
		status = CLOSED;
	return wrote;
}

int32 TCPSocket::write(char c)
{
	char buf[2] = "\0";
	
	buf[0] = c;
	return write(buf, 2);
}


bool TCPSocket::block_read(char *buf, long buf_len)
{
	long i=0;
	
	while (i < buf_len) {
		long j;
		if ((j = read(buf + i, buf_len - i)) < 0)
			return FALSE;
		i += j;
	}

	return TRUE;
}

bool TCPSocket::block_write(char *buf, long buf_len)
{
	long i=0;
	
	while (i < buf_len) {
		long j;
		if ((j = write(buf + i, buf_len - i)) < 0)
			return FALSE;
		i += j;
	}

	return TRUE;
}
