/*
 * Amanda, The Advanced Maryland Automatic Network Disk Archiver
 * Copyright (c) 1991 University of Maryland
 * All Rights Reserved.
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of U.M. not be used in advertising or
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.  U.M. makes no representations about the
 * suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * U.M. DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL U.M.
 * BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
 * OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Author: James da Silva, Systems Design and Analysis Group
 *			   Computer Science Department
 *			   University of Maryland at College Park
 */
/*
 * $Id$
 *
 * read client configuration file
 */
/*
 *
 * XXX - I'm not happy *at all* with this implementation, but I don't
 * think YACC would be any easier.  A more table based implementation
 * would be better.  Also clean up memory leaks.
 */
#include "amanda.h"
#include "arglist.h"

#include "conffile.h"

#define BIGINT	1000000000		/* 2 million yrs @ 1 per day */

/* internal types and variables */

/*
 * XXX - this is used by the krb4 stuff.
 * Hopefully nobody will need this here.  (not very likely).  -kovert
 */
#if defined(INTERFACE)
#  undef INTERFACE
#endif

typedef enum {
    UNKNOWN, ANY, COMMA, LBRACE, RBRACE, NL, END,
    IDENT, INT, BOOL, REAL, STRING,

    /* config parameters */
    INCLUDEFILE,

    /* disk type */
    DISK, EDUMP, EGNUTAR, BACKUP, INDEX, SIZE, COMMENT
} tok_t;

typedef struct {	/* token table entry */
    char *keyword;
    tok_t token;
} keytab_t;

keytab_t *keytable;

typedef union {
    int i;
    double r;
    char *s;
} val_t;

static val_t tokenval;
static tok_t tok;

/* configuration parameters */

static dumpdisk_t ddcur;
static dumpprog_t dpcur;

static int line_num, got_parserror, allow_overwrites;
static dumpdisk_t *disklist = NULL;
static FILE *conf = (FILE *)NULL;
static char *confname = NULL;

/* predeclare local functions */

static void init_defaults P((void));
static void read_conffile_recursively P((char *filename));

static int read_confline P((void));
static void get_dumpdisk P((void));
static void init_dumpdisk_defaults P((void));
static void save_dumpdisk P((void));
static void get_program P((void));
static void copy_program P((dumpprog_t*));

static void get_simple P((val_t *var, int *seen, tok_t type));
static void ckseen(int*);
static void parserror P((char *format, ...))
    __attribute__ ((format (printf, 1, 2)));
static tok_t lookup_keyword P((char *str));
static void get_conftoken P((tok_t exp));

/*
** ------------------------
**  External entry points
** ------------------------
*/

int read_conffile(filename)
char *filename;
{
    init_defaults();

    /* We assume that confname & conf are initialized to NULL above */
    read_conffile_recursively(filename);

    return got_parserror;
}


/*
** ------------------------
**  Internal routines
** ------------------------
*/


static void init_defaults()
{
    allow_overwrites = 0;
    /* not used yet */
}

static void read_conffile_recursively(filename)
char *filename;
{
    extern int errno;

    /* Save globals used in read_confline(), elsewhere. */
    int  save_line_num  = line_num;
    FILE *save_conf     = conf;
    char *save_confname = confname;

    confname = stralloc(filename);

    if((conf = fopen(confname, "r")) == NULL) {
        free((void*)confname);
        return;
    }

    line_num = 0;

    /* read_confline() can invoke us recursively via "includefile" */
    while(read_confline());
    afclose(conf);

    afree(confname);

    /* Restore globals */
    line_num = save_line_num;
    conf     = save_conf;
    confname = save_confname;
}


/* ------------------------ */


keytab_t main_keytable[] = {
    { "DISK", DISK },
    { NULL, IDENT }
};

static int read_confline()
{
    keytable = main_keytable;

    line_num += 1;
    get_conftoken(ANY);
    switch(tok) {
    case INCLUDEFILE:
	{
	    char *fn;

	    get_conftoken(STRING);
	    fn = tokenval.s;
	    read_conffile_recursively(fn);
	}
	break;

    case DISK: {
        get_dumpdisk();
	break;
    }
    case NL:	/* empty line */
	break;
    case END:	/* end of file */
	return 0;
    default:
	parserror("configuration keyword expected");
    }
    if(tok != NL) get_conftoken(NL);
    return 1;
}

keytab_t dumpdisk_keytable[] = {
    { "COMMENT", COMMENT },
    { "DUMP", EDUMP },
    { "GNUTAR", EGNUTAR },
    /* I would rather have this in a program_keytable, but debugging */
    { "BACKUP", BACKUP },
    { "INDEX", INDEX },
    { "SIZE", SIZE },
    { NULL, IDENT }
};

static void get_dumpdisk()
{
    int done;
    int save_overwrites, tmp;
    keytab_t *save_kt;

    save_kt = keytable;
    keytable = dumpdisk_keytable;
    save_overwrites = allow_overwrites;
    allow_overwrites = 1;

    init_dumpdisk_defaults();

    get_conftoken(STRING);
    ddcur.diskroot = stralloc(tokenval.s);

    get_conftoken(LBRACE);
    get_conftoken(NL);

    done = 0;
    do {
	line_num += 1;
	get_conftoken(ANY);
        tmp = 0;
	switch(tok) {

	case RBRACE:
	    done = 1;
	    break;
	case EDUMP:
	    dbprintf(("get_dumpdisk: got dump\n"));
            get_program();
	    copy_program(&ddcur.dump);
            break;
        case EGNUTAR:
	    dbprintf(("get_dumpdisk: got gnutar\n"));
            get_program();
            copy_program(&ddcur.gnutar);
            break;
	case COMMENT:
	    dbprintf(("get_dumpdisk: got comment\n"));
	    get_simple((val_t *)&ddcur.comment, &tmp, STRING);
	    break;
	case NL:	/* empty line */
	    break;
	case END:	/* end of file */
	    done = 1;
	default:
	    parserror("dump disk parameter expected");
	}
	if(tok != NL && tok != END) get_conftoken(NL);
    } while(!done);

    save_dumpdisk();

    allow_overwrites = save_overwrites;
    keytable = save_kt;
}

static void init_dumpdisk_defaults()
{
    ddcur.comment = "";
    ddcur.diskroot = "/";
    ddcur.dump.backup = 0;
    ddcur.dump.size = 0;
    dpcur.backup = 0;
    dpcur.size = 0;
}

static void save_dumpdisk()
{
    dumpdisk_t *dp;

    dp = alloc(sizeof(dumpdisk_t));
    *dp = ddcur;
    dp->next = disklist;
    disklist = dp;
}

dumpdisk_t *lookup_dumpdisk(char *pathname)
{
    dumpdisk_t *dp, *ret;
    char *iter1, *iter2;
    int lenmatch, curmatch;

    lenmatch = 0;
    ret = 0;
    dbprintf(("lookup_dumpdisk: %s\n", pathname));
    for (dp = disklist; dp; dp = dp->next) {
	curmatch = 0;
	iter1 = pathname; iter2 = dp->diskroot;
	while (*iter1 && *iter2 && (*iter1 == *iter2)) {
	    iter1++;
	    iter2++;
	    curmatch++;
	}
	if (!*iter2 && (curmatch > lenmatch)) {
	    if (*iter1 && *iter1 != '/') {
		lenmatch = curmatch;
		ret = dp;
	    }
	}
    }
    return(ret);
}
/*
keytab_t program_keytable[] = {
    { "BACKUP:", BACKUP },
    { "INDEX:", INDEX },
    { "SIZE:", SIZE },
    { NULL, IDENT }
};
*/
static void get_program()
{
    get_conftoken(ANY);
    switch (tok) {
    case BACKUP:
	get_conftoken(STRING);
	afree(dpcur.backup);
	dpcur.backup = stralloc(tokenval.s);
	dbprintf(("got backup: %s\n", tokenval.s));
	break;
    case INDEX:
	get_conftoken(STRING);
	afree(dpcur.index);
	dpcur.index = stralloc(tokenval.s);
	dbprintf(("got index: %s\n", tokenval.s));
	break;
    case SIZE:
	get_conftoken(STRING);
	afree(dpcur.size);
	dpcur.size = stralloc(tokenval.s);
	dbprintf(("got size: %s\n", tokenval.s));
	break;
    default:
	parserror("expecting \"index:\", \"backup:\" or \"size\".\n");
    }
    get_conftoken(NL);
}

static void copy_program(dumpprog_t *dp)
{
#define pcopy(s) if (!dp->s) dp->s = dpcur.s; dpcur.s = 0
    pcopy(backup);
    pcopy(index);
    pcopy(size);
#undef pcopy
}


/* ------------------------ */


static void get_simple(var, seen, type)
val_t *var;
int *seen;
tok_t type;
{
    ckseen(seen);

    switch(type) {
    case STRING:
    case IDENT:
	get_conftoken(type);
	var->s = stralloc(tokenval.s);
	break;
    case INT:
	var->i = get_number();
	break;
    case BOOL:
	var->i = get_bool();
	break;
    case REAL:
	get_conftoken(REAL);
	var->r = tokenval.r;
	break;
    default:
	error("error [unknown get_simple type: %d]", type);
	/* NOTREACHED */
    }
    return;
}

static void ckseen(seen)
int *seen;
{
    if(*seen && !allow_overwrites) {
	parserror("duplicate parameter, prev def on line %d", *seen);
    }
    *seen = line_num;
}

arglist_function(static void parserror, char *, format)
{
    va_list argp;

    /* print error message */

    fprintf(stderr, "\"%s\", line %d: ", confname, line_num);
    arglist_start(argp, format);
    vfprintf(stderr, format, argp);
    arglist_end(argp);
    fputc('\n', stderr);

    got_parserror = 1;
}

static tok_t lookup_keyword(str)
char *str;
{
    keytab_t *kwp;

    /* switch to binary search if performance warrants */

    for(kwp = keytable; kwp->keyword != NULL; kwp++) {
	if(strcmp(kwp->keyword, str) == 0) break;
    }
    return kwp->token;
}

static char tkbuf[4096];

static void get_conftoken(exp)
tok_t exp;
{
    int ch, i, d;
    char *buf;
    int token_overflow;

    ch = getc(conf);

    while(ch != EOF && ch != '\n' && isspace(ch)) ch = getc(conf);
    if(ch == '#') {		/* comment - eat everything but eol/eof */
        while((ch = getc(conf)) != EOF && ch != '\n') {}
    }

    if(isalpha(ch)) {		/* identifier */
        buf = tkbuf;
        token_overflow = 0;
        do {
	    if(islower(ch)) ch = toupper(ch);
	    if(buf < tkbuf+sizeof(tkbuf)-1) {
		*buf++ = ch;
	    } else {
	        *buf = '\0';
		if(!token_overflow) {
		    parserror("token too long: %.20s...", tkbuf);
		}
		token_overflow = 1;
	    }
	    ch = getc(conf);
	} while(isalnum(ch) || ch == '_' || ch == '-');

	ungetc(ch, conf);
	*buf = '\0';

	tokenval.s = tkbuf;

	if(token_overflow) tok = UNKNOWN;
	else if(exp == IDENT) tok = IDENT;
	else tok = lookup_keyword(tokenval.s);
    }
    else if(isdigit(ch)) {	/* integer */
	tokenval.i = 0;
	do {
	    tokenval.i = tokenval.i * 10 + (ch - '0');
	    ch = getc(conf);
	} while(isdigit(ch));
	if(ch != '.') {
	    if(exp != REAL)
		    tok = INT;
	    else {
		/* automatically convert to real when expected */
		i = tokenval.i;
		tokenval.r = (double) i;
		tok = REAL;
	    }
	}
	else {
	    /* got a real number, not an int */
	    i = tokenval.i;
	    tokenval.r = (double) i;
	    i=0; d=1;
	    ch = getc(conf);
	    while(isdigit(ch)) {
		i = i * 10 + (ch - '0');
		d = d * 10;
		ch = getc(conf);
	    }
	    tokenval.r += ((double)i)/d;
	    tok = REAL;
	}
	ungetc(ch,conf);
    }
    else switch(ch) {

    case '"':			/* string */
	buf = tkbuf;
	token_overflow = 0;
	ch = getc(conf);
	while(ch != '"' && ch != '\n' && ch != EOF) {
	    /* understand escapes */
	    if (ch == '\\')
		ch = getc(conf);
	    if(buf < tkbuf+sizeof(tkbuf)-1) {
	        *buf++ = ch;
	    } else {
		*buf = '\0';
		if(!token_overflow) {
		    parserror("string too long: %.20s...", tkbuf);
		}
		token_overflow = 1;
	    }
	    ch = getc(conf);
	}
	if(ch != '"') {
	    parserror("missing end quote");
	    ungetc(ch, conf);
	}
	*buf = '\0';
	tokenval.s = tkbuf;
	if(token_overflow) tok = UNKNOWN;
	else tok = STRING;
	break;

    case ',':  tok = COMMA; break;
    case '{':  tok = LBRACE; break;
    case '}':  tok = RBRACE; break;
    case '\n': tok = NL; break;
    case EOF:  tok = END; break;
    default:   tok = UNKNOWN;
    }

    if(exp != ANY && tok != exp) {
	char *str;
	keytab_t *kwp;

	switch(exp) {
	case LBRACE: str = "\"{\""; break;
	case RBRACE: str = "\"}\""; break;
	case COMMA:  str = "\",\""; break;

	case NL: str = "end of line"; break;
	case END: str = "end of file"; break;
	case INT: str = "an integer"; break;
	case REAL: str = "a real number"; break;
	case STRING: str = "a quoted string"; break;
	case IDENT: str = "an identifier"; break;
	default:
	    for(kwp = keytable; kwp->keyword != NULL; kwp++)
		if(exp == kwp->token) break;
	    if(kwp->keyword == NULL) str = "token not";
	    else str = kwp->keyword;
	}
	parserror("%s expected", str);
	tok = exp;
	if(tok == INT) tokenval.i = 0;
	else tokenval.s = "";
    }

    return;
}


/* ------------------------ */


#ifdef TEST
dump_configuration(filename)
{
    tapetype_t *tp;
    dumptype_t *dp;
    interface_t *ip;
    holdingdisk_t *hp;
    time_t st;
    struct tm *stm;

    printf("AMANDA CONFIGURATION FROM FILE \"%s\":\n\n", filename);

    printf("conf_org = \"%s\"\n", getconf_str(CNF_ORG));
    printf("conf_mailto = \"%s\"\n", getconf_str(CNF_MAILTO));
    printf("conf_dumpuser = \"%s\"\n", getconf_str(CNF_DUMPUSER));
    printf("conf_tapedev = \"%s\"\n", getconf_str(CNF_TAPEDEV));
    printf("conf_tpchanger = \"%s\"\n", getconf_str(CNF_TPCHANGER));
    printf("conf_chngrdev = \"%s\"\n", getconf_str(CNF_CHNGRDEV));
    printf("conf_chngrfile = \"%s\"\n", getconf_str(CNF_CHNGRFILE));
    printf("conf_labelstr = \"%s\"\n", getconf_str(CNF_LABELSTR));
    printf("conf_tapelist = \"%s\"\n", getconf_str(CNF_TAPELIST));
    printf("conf_infofile = \"%s\"\n", getconf_str(CNF_INFOFILE));
    printf("conf_logdir = \"%s\"\n", getconf_str(CNF_LOGDIR));
    printf("conf_diskfile = \"%s\"\n", getconf_str(CNF_DISKFILE));
    printf("conf_tapetype = \"%s\"\n", getconf_str(CNF_TAPETYPE));

    printf("conf_dumpcycle = %d\n", getconf_int(CNF_DUMPCYCLE));
    printf("conf_runtapes = %d\n", getconf_int(CNF_RUNTAPES));
    printf("conf_tapecycle = %d\n", getconf_int(CNF_TAPECYCLE));
    printf("conf_bumpsize = %d\n", getconf_int(CNF_BUMPSIZE));
    printf("conf_bumpdays = %d\n", getconf_int(CNF_BUMPDAYS));
    printf("conf_bumpmult = %f\n", getconf_real(CNF_BUMPMULT));
    printf("conf_netusage = %d\n", getconf_int(CNF_NETUSAGE));
    printf("conf_inparallel = %d\n", getconf_int(CNF_INPARALLEL));
    /*printf("conf_timeout = %d\n", getconf_int(CNF_TIMEOUT));*/
    printf("conf_maxdumps = %d\n", getconf_int(CNF_MAXDUMPS));
    printf("conf_etimeout = %d\n", getconf_int(CNF_ETIMEOUT));

    /*printf("conf_diskdir = \"%s\"\n", getconf_str(CNF_DISKDIR));*/
    /*printf("conf_disksize = %d\n", getconf_int(CNF_DISKSIZE));*/
    printf("conf_indexdir = \"%s\"\n", getconf_str(CNF_INDEXDIR));
    printf("num_holdingdisks = %d\n", num_holdingdisks);
    for(hp = holdingdisks; hp != NULL; hp = hp->next) {
	printf("\nHOLDINGDISK %s:\n", hp->name);
	printf("	COMMENT \"%s\"\n", hp->comment);
	printf("	DISKDIR \"%s\"\n", hp->diskdir);
	printf("	SIZE %d\n", hp->disksize);
    }

    for(tp = tapelist; tp != NULL; tp = tp->next) {
	printf("\nTAPETYPE %s:\n", tp->name);
	printf("	COMMENT \"%s\"\n", tp->comment);
	printf("	LENGTH %u\n", tp->length);
	printf("	FILEMARK %u\n", tp->filemark);
	printf("	SPEED %d\n", tp->speed);
    }

    for(dp = dumplist; dp != NULL; dp = dp->next) {
	printf("\nDUMPTYPE %s:\n", dp->name);
	printf("	COMMENT \"%s\"\n", dp->comment);
	printf("	PROGRAM \"%s\"\n", dp->program);
	printf("	PRIORITY %d\n", dp->priority);
	printf("	DUMPCYCLE %d\n", dp->dumpcycle);
	st = dp->start_t;
	if(st) {
	    stm = localtime(&st);
	    printf("	STARTTIME %d:%02d:%02d\n",
	      stm->tm_hour, stm->tm_min, stm->tm_sec);
	}
	if(dp->exclude) {
	    printf("	EXCLUDE");
	    if(dp->exclude_list) {
		printf("-LIST");
	    }
	    printf(" \"%s\"\n", dp->exclude);
	}
	printf("	FREQUENCY %d\n", dp->frequency);
	printf("	MAXDUMPS %d\n", dp->maxdumps);
	printf("	STRATEGY ");
	switch(dp->strategy) {
	case DS_SKIP:
	    printf("SKIP");
	    break;
	case DS_STANDARD:
	    printf("STANDARD");
	    break;
	case DS_NOFULL:
	    printf("NOFULL");
	    break;
	case DS_NOINC:
	    printf("NOINC");
	    break;
	case DS_HANOI:
	    printf("HANOI");
	    break;
	}
	putchar('\n');
	printf("	COMPRATE %f, %f\n", dp->comprate[0], dp->comprate[1]);

	printf("	OPTIONS: ");

	switch(dp->compress) {
	case COMP_NONE:
	    printf("NO-COMPRESS ");
	    break;
	case COMP_FAST:
	    printf("COMPRESS-FAST ");
	    break;
	case COMP_BEST:
	    printf("COMPRESS-BEST ");
	    break;
	case COMP_SERV_FAST:
	    printf("SRVCOMP-FAST ");
	    break;
	case COMP_SERV_BEST:
	    printf("SRVCOMP-BEST ");
	    break;
	}

	if(!dp->record) printf("NO-");
	printf("RECORD");
	if(dp->auth == AUTH_BSD) printf(" BSD-AUTH");
	else if(dp->auth == AUTH_KRB4) printf(" KRB4-AUTH");
	else printf(" UNKNOWN-AUTH");
	if(dp->skip_incr) printf(" SKIP-INCR");
	if(dp->skip_full) printf(" SKIP-FULL");
	if(dp->no_hold) printf(" NO-HOLD");
	if(dp->kencrypt) printf(" KENCRYPT");
	/* an ignored disk will never reach this point */
	assert(!dp->ignore);
	if(dp->index) printf(" INDEX");
	putchar('\n');
    }

    for(ip = interface_list; ip != NULL; ip = ip->next) {
	printf("\nINTERFACE %s:\n", ip->name);
	printf("	COMMENT \"%s\"\n", ip->comment);
	printf("	USE %d\n", ip->maxusage);
    }
}

main(argc, argv)
int argc;
char *argv[];
{
  int result;
  int fd;

  for(fd = 3; fd < FD_SETSIZE; fd++) {
    /*
     * Make sure nobody spoofs us with a lot of extra open files
     * that would cause an open we do to get a very high file
     * descriptor, which in turn might be used as an index into
     * an array (e.g. an fd_set).
     */
    close(fd);
  }

  startclock();
  if (argc>1)
    chdir(argv[1]);
  result = read_conffile(CONFFILE_NAME);
  dump_configuration(CONFFILE_NAME);
  return result;
}

char *pname = "conffile";

#endif /* TEST */
