/*
 * feedtree -- walk a news-spool hierarchy sending every article in it to
 * a foreign news host.  Puts in an Xref: header if it's missing.
 *
 * Copyright (c) 1999 Aidan Cully
 */
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <dirent.h>
#include <fcntl.h>
#include <string.h>
#include <stdio.h>

static void send_file(char*, char*, struct stat*);

static int dirskip;
static char *pathhost;

int main(int argc, char *argv[])
{
    char inbuf[1024], pwd[1024], *slash;
    struct stat st;

    /* parse argv (Hah!) */
    if (argc != 3) {
	fprintf(stderr, "Usage:\n\tfeedtree <pathhost> <spooldir>\n");
	exit(1);
    }
    pathhost = argv[1];
    dirskip = strlen(argv[2]);

    pwd[0] = '\0';
    inbuf[sizeof(inbuf)-1] = '\0';
    while (fgets(inbuf, sizeof(inbuf)-1, stdin) != 0) {
	inbuf[strlen(inbuf)-1] = '\0';
	slash = strrchr(inbuf, '/');
	if (!slash) {
	    fprintf(stderr, "Not a directory name: %s\n", inbuf);
	    continue;
	}
	*slash++ = '\0';
	if (strcmp(inbuf, pwd) != 0) {
	    if (chdir(inbuf) < 0) {
		fprintf(stderr, "Couldn't chdir to %s\n", inbuf);
		exit(1);
	    }
	    strcpy(pwd, inbuf);
	}
	if (lstat(slash, &st) < 0) {
	    fprintf(stderr, "Couldn't stat %s/%s\n", pwd, slash);
	    continue;
	}
	if (!S_ISREG(st.st_mode)) {
	    fprintf(stderr, "%s/%s: Not an ordinary file.\n", pwd, slash);
	    continue;
	}
	send_file(pwd, slash, &st);
    }
    exit(0);
}

static void send_file(char *pwd, char *f_name, struct stat *st)
{
    static char *buf = 0;
    static int buflen = 0;
    char *xref, *hold, *nl, *msgid, *eoh;
    int fd;
    int got, done;
    int i;

    if (!st->st_size)
	return;
    if (buflen < st->st_size) {
	if (buf)
	    free((void*) buf);
	buf = (char*) malloc(st->st_size);
	buflen = st->st_size;
    }
    fd = open(f_name, O_RDONLY, 0644);
    done = 0;
    while (done < st->st_size) {
	got = read(fd, buf, st->st_size - done);
	done += got;
    }
    close(fd);
    nl = buf;
    msgid = 0;
    xref = 0;
    while (1) {
	if (!strncasecmp(nl, "Message-ID: ", 12))
	    msgid = nl+12;
	else if (!strncasecmp(nl, "Xref: ", 6))
	    xref = nl+6;

	/* skip to the next newline in the buffer */
	nl = strchr(nl+1, '\n');
	if (!nl)
	    return;
	nl++;
	if (nl[0] == '\r')
	    nl++;
	if (nl[0] == '\n')
	    break;
    }
    eoh = nl;
    eoh--;
    if (*eoh == '\r')
	eoh--;

    if (!msgid) {
	fprintf(stderr, "%s/%s No message-ID.\n", pwd, f_name);
	return;
    }
    hold = msgid;
    nl = strchr(msgid, '\n');
    msgid = (char*) malloc(nl-hold+2);
    memcpy(msgid, hold, nl-hold+1);
    if (xref) {
	/* check that the current newsgroup is the first newsgroup in the
	 * xref */
	hold = xref;
	xref = strchr(xref, ' ')+1;
	for (i = 0; xref[i] != ':' &&
	  (xref[i] == pwd[dirskip+i] ||
	   (xref[i] == '.' && pwd[dirskip+i] == '/')); i++)
	    ;
	if (xref[i] != ':' || i+dirskip != strlen(pwd)) {
	    free((void*) msgid);
	    return;
	}
	xref = hold;
    } else {
	/* there's no xref header, so we have to rebuild the article file
	 * with one. */
	fprintf(stderr, "%s/%s Rebuilding xref.\n", pwd, f_name);
	fd = open("feedtree.tmp", O_CREAT|O_RDWR, 0644);
	fflush(stdout);
	done = eoh - buf+1;
	i = 0;
	while (i < done) {
	    got = write(fd, buf+i, done-i);
	    i += got;
	}
	sprintf(buf, "Xref: %s ", pathhost);
	xref = buf+strlen(buf);
	sprintf(xref, "%s:%s\n", pwd+dirskip, f_name);
	for (i = 0; xref[i] != ':'; i++)
	    if (xref[i] == '/')
		xref[i] = '.';
	done = strlen(buf);
	i = 0;
	while (i < done) {
	    got = write(fd, buf+i, done-i);
	    i += got;
	}
	done = st->st_size - (eoh - buf + 1);
	i = 0;
	while (i < done) {
	    got = write(fd, eoh+1+i, done-i);
	    i += got;
	}
	close(fd);
	rename("feedtree.tmp", f_name);
    }
    nl = strchr(msgid, '\n');
    *nl = '\0';
    if (*(nl-1) == '\r') {
	nl--;
	*nl = '\0';
    }
    printf("%s/%s %s\n", pwd, f_name, msgid);
    free((void*) msgid);
}
