/* filter.c -- track posts made by users */

/* revision history
 * 8/18/97 -
 *     eliminated race condition
 * 8/19/97 -
 *     fixed bug where FILTsetuser wouldn't properly load the user's limits.
 * 8/26/97 -
 *     externalized limit file loading/parseing
 * 8/27/97 -
 *     logs errors using syslog instead of stderr
 * 9/4/97 -
 *     now can use identd (encrypting or not) to get resolve a username
 * 9/15/97 -
 *     FILTresolvers move into passwd authentication from nnrp.access
 * 10/6/97 -
 *     Now connection oriented.
 * 10/20/97 - Aidan Cully
 *     Inherits data elements
 */
#include <stdio.h>
#include <string.h>
#include <syslog.h>
#include <ctype.h>
#include <stdlib.h>

#include "filter.h"
#include "filtpaths.h"
#include "config.h"
#include "configdata.h"
#include "misc.h"
#include "subst.h"

CONF *FILTkeylist;

typedef struct _FILTdatael {
    int type;
    char *name;
    void *data;
    size_t len;
    struct _FILTdatael *next;
} FILTdatael;

typedef struct _FILTpriv {
    FILTdatael *data;
    FILT *parent;
} FILTpriv;

typedef struct _FILTdellist {
    int type;
    FILTdelfunc func;
    struct _FILTdellist *next;
} FILTdellist;

typedef struct _FILTfiledata {
    char *fname;
    FILE *file;
    int isopen;
} FILTfiledata;

static FILTdellist *dellist = 0;

static void FILTdelfiledata(int, void*, size_t);
static void FILTdelfiltdata(int, void*, size_t);
static int FILTsvar(char*, char*, char*, CONF*);

static CONFkeyload FILTloaders[] =
 { {"VAR", FILTsvar} };

/*
** The remote user can be set by calling FILTsetuser(username)
*/
char *
FILTgetuser(filt)
    FILT *filt;
{
    return(FILTgetstring(filt, "user", 1));
}

/*
 * Set the user and load the new posting limits.
 */
void
FILTsetuser(filt, newuser)
    FILT *filt;
    char *newuser;
{
    FILTaddstring(filt, "user", newuser, 1);
    LIMloaduserlimits(filt);
}

/*
 * log the user/message id so that we can cancel the message
 */
void
FILTlogmessage(filt, id, from, newsgroups)
    FILT *filt;
    char *id, *from, *newsgroups;
{
    FILE *msgidlog;
    char *user;
    char *iter;

    if (!(user = FILTgetuser(filt)))
        return;
    if (!(msgidlog = fopen(L_USRMSGLOG, "a+")))
        return;

    /* log only the email part of the username.  Do this by searching for
     * an '@' and isolating that word. */
    iter = &from[strlen(from)-1];
    while (isblank(*iter))
        iter--;
    if (*iter == '>') {
        iter[1] = 0;
        while (iter > from && !isblank(*iter))
            iter--;
        from = iter;
    } else {
        for (from; *from && isblank(*from); from++)
            ;
        for (iter = from; *iter && !isblank(*iter); iter++)
            ;
        *iter = 0;
    }

    fprintf(msgidlog, "%s:%s\t%s\t%s\n", user, id, from, newsgroups);
    fclose(msgidlog);
}

/*
static void
die(fmt, args)
    char *fmt;
    va_alist args;
{
    fprintf(stderr, "%s:%d ", filename, lineno);
    fprintf(stderr, fmt, args);
    exit(1);
}
*/

/*
** Don't barf on system defined keywords.
*/
void
FILTglobalkeywordinit()
{
    CONFkeywordused(FILTkeylist, "mail");
    CONFkeywordused(FILTkeylist, "mail.text");
    CONFkeywordused(FILTkeylist, "mail.block");
    CONFkeywordused(FILTkeylist, "mail.block.text");
    CONFkeywordused(FILTkeylist, "hold.dir");
    CONFkeywordused(FILTkeylist, "hold.deldir");
    if (!CONFkeywordvalue(FILTkeylist, "limitfile")) {
        syslog(LOG_ERR, "You did not specify a filter limit file.  Aborting.\n");
        exit(1);
    }
    CONFkeywordused(FILTkeylist, "mail.error");
    CONFkeywordused(FILTkeylist, "mail.error.text");
    CONFkeywordused(FILTkeylist, "rnews.domain");
}

/*
** This procedure loads the keywords from the FILT_CONF_FILE into memory and
** then initializes the global keywords from the file.
*/
int
FILTinit()
{
    CONFsetload(FILTloaders, sizeof(FILTloaders)/sizeof(CONFkeyload));
    FILTkeylist = CONFload(_PATH_LIMITCONF);
    if (!FILTkeylist)
        return(1);
    FILTglobalkeywordinit();
    FILTadddel(kFILTfiletype, FILTdelfiledata);
    FILTadddel(kFILTfilttype, FILTdelfiltdata);
    return(0);
}

/*
** Initialize a FILT for holding this connection's data
*/
FILT *
FILTconnect(parent)
    FILT *parent;
{
    FILTpriv *real;

    real = (FILTpriv*)malloc(sizeof(FILTpriv));
    if (!real)
        return(0);
    real->data = 0;
    real->parent = parent;
    return((FILT*)real);
}

/*
** Destroy a FILT and all data associated with it.
*/
void
FILTdestroy(filt)
    FILT *filt;
{
    FILTpriv *real;
    FILTdatael *iter;

    real = (FILTpriv*)filt;
    real->parent = 0;
    iter = real->data;
    while (real->data)
        FILTdeldata(filt, real->data->type, real->data->name, 0);
    free((void*)real);
}

/*
** The FILT accessor functions.
*/
void *
FILTgetdata(filt, type, name, size, recurse)
    FILT *filt;
    int type;
    char *name;
    size_t *size;
    int recurse;
{
    FILTdatael *iter = ((FILTpriv*)filt)->data;

    while (iter) {
        if ((type == iter->type) && !strcmp(name, iter->name)) {
            *size = iter->len;
            return(iter->data);
        }
        iter = iter->next;
    }
    /* if we've got a parent, inherit their storage */
    if (recurse && ((FILTpriv*)filt)->parent)
        return(FILTgetdata(((FILTpriv*)filt)->parent, type, name, size, 1));
    *size = 0;
    return(0);
}

/* find the parent associated with this guy */
FILT *
FILTgetparent(filt)
    FILT *filt;
{
    return ((FILTpriv*)filt)->parent;
}

char *
FILTgetstring(filt, name, recurse)
    FILT *filt;
    char *name;
    int recurse;
{
    size_t size;

    return (char*)FILTgetdata(filt, kFILTstrtype, name, &size, recurse);
}

int
FILTgetint(filt, name, recurse)
    FILT *filt;
    char *name;
    int recurse;
{
    void *ret;
    size_t size;

    ret = FILTgetdata(filt, kFILTinttype, name, &size, recurse);
    if (!ret)
        return(0);
    return *((int*)ret);
}

int
FILTadddata(filt, type, name, data, size, recurse)
    FILT *filt;
    int type;
    char *name;
    void *data;
    size_t size;
    int recurse;
{
    FILTdatael *iter, *prev;
    FILT *parent;
    int found;
    char *from, *to;

    if (!data) {
        FILTdeldata(filt, type, name, recurse);
        return(1);
    }
    if (recurse) {
        while ((parent = FILTgetparent(filt)) != 0)
            filt = parent;
    }
    do {
        /* there can be only one. */
        iter = ((FILTpriv*)filt)->data;
        found = 0;
        prev = 0;
        while (!found && iter) {
            if ((type == iter->type) && !strcmp(name, iter->name))
                found = 1;
            prev = iter;
            iter = iter->next;
        }
        if (found)
            FILTdeldata(filt, type, name, 0);
    } while (found);

    /* fill in the data with error checking */
    iter = (FILTdatael*)malloc(sizeof(FILTdatael));
    if (!iter)
        return(0);
    iter->data = malloc(size);
    if (!iter->data) {
        free((void*)iter);
        return(0);
    }
    from = (char*)data;
    to = (char*)iter->data;
    while (size--)
        *to++ = *from++;
    iter->len = size;

    iter->name = (char*)malloc(strlen(name)+1);
    if (!iter->name) {
        free((void*)iter->data);
        free((void*)iter);
        return(0);
    }
    strcpy(iter->name, name);
    iter->type = type;
    iter->next = 0;

    /* If we got here, everything's fine.  Insert the new guy into the
     * list */
    if (prev)
        prev->next = iter;
    else
        ((FILTpriv*)filt)->data = iter;
    return(1);
}

int
FILTaddstring(filt, name, str, recurse)
    FILT *filt;
    char *name;
    char *str;
    int recurse;
{
    if (!str)
        FILTdeldata(filt, kFILTstrtype, name, recurse);
    else
        return(FILTadddata(filt, kFILTstrtype, name, (void*)str,
          strlen(str)+1, recurse));
}

int
FILTaddint(filt, name, add, recurse)
    FILT *filt;
    char *name;
    int add;
    int recurse;
{
    return(FILTadddata(filt, kFILTinttype, name, (void*)&add, sizeof(add), recurse));
}

int
FILThasdata(filt, type, name, recurse)
    FILT *filt;
    int type;
    char *name;
    int recurse;
{
    size_t s;

    return (FILTgetdata(filt, type, name, &s, recurse) != 0);
}

int
FILThasstring(filt, name, recurse)
    FILT *filt;
    char *name;
    int recurse;
{
    return FILThasdata(filt, kFILTstrtype, name, recurse);
}

int
FILThasint(filt, name, recurse)
    FILT *filt;
    char *name;
    int recurse;
{
    return FILThasdata(filt, kFILTinttype, name, recurse);
}

int
FILThasfile(filt, name, recurse)
    FILT *filt;
    char *name;
    int recurse;
{
    return FILThasdata(filt, kFILTfiletype, name, recurse);
}

/* delete data associated with a type and a name.  Don't delete it from the
 * parent if we don't have it. */
void
FILTdeldata(filt, type, name, recurse)
    FILT *filt;
    int type;
    char *name;
    int recurse;
{
    FILTdatael *prev = 0;
    FILTdatael *r = ((FILTpriv*)filt)->data;
    FILTdellist *iter = dellist;
    int del, found;

    found = 0;
    while (r && !found) {
        if ((r->type == type) && !strcmp(r->name, name))
            found = 1;
        else {
            prev = r;
            r = r->next;
        }
    }
    if (!r) {
        if (recurse && ((FILTpriv*)filt)->parent)
            FILTdeldata(((FILTpriv*)filt)->parent, type, name, recurse);
        return;
    }
    while (iter && !del) {
        if (iter->type == type) {
            iter->func(type, r->data, r->len);
            del = 1;
        }
        iter = iter->next;
    }
    if (!del)
        free((void*)r->data);
    free((void*)r->name);
    if (prev)
        prev->next = r->next;
    else
        ((FILTpriv*)filt)->data = r->next;
    free((void*)r);
}

void
FILTdelstring(filt, name, recurse)
    FILT *filt;
    char *name;
    int recurse;
{
    FILTdeldata(filt, kFILTstrtype, name, recurse);
}

void
FILTdelint(filt, name, recurse)
    FILT *filt;
    char *name;
    int recurse;
{
    FILTdeldata(filt, kFILTinttype, name, recurse);
}

void
FILTdelfile(filt, name, recurse)
    FILT *filt;
    char *name;
    int recurse;
{
    FILTdeldata(filt, kFILTfiletype, name, recurse);
}

int
FILTaddfilt(filt, name, add, recurse)
    FILT *filt;
    char *name;
    FILT *add;
    int recurse;
{
    return(FILTadddata(filt, kFILTfilttype, name, (void*)&add, sizeof(FILT*), recurse));
}

FILT *
FILTgetfilt(filt, name, recurse)
    FILT *filt;
    char *name;
    int recurse;
{
    size_t s;
    FILT **nflt = (FILT**)FILTgetdata(filt, kFILTfilttype, name, &s, recurse);
    if (!nflt)
        return(0);
    return(*nflt);
}

int
FILThasfilt(filt, name, recurse)
    FILT *filt;
    char *name;
    int recurse;
{
    return(FILTgetfilt(filt, name, recurse) != 0);
}

void
FILTdelfilt(filt, name, recurse)
    FILT *filt;
    char *name;
    int recurse;
{
    FILTdeldata(filt, kFILTfilttype, name, recurse);
}

/* File functions */
int
FILTaddfile(filt, name, fname, f, recurse)
    FILT *filt;
    char *name;
    char *fname;
    FILE *f;
    int recurse;
{
    FILTfiledata d;
    int ret;

    d.fname = (char*)malloc(strlen(fname)+1);
    if (!d.fname)
        return(0);
    strcpy(d.fname, fname);
    d.file = f;
    d.isopen = 1;
    ret = FILTadddata(filt, kFILTfiletype, name, (void*)&d, sizeof(d), recurse);
    if (!ret)
        free((void*)d.fname);
    return(ret);
}

FILE *
FILTgetfile(filt, name, recurse)
    FILT *filt;
    char *name;
    int recurse;
{
    FILTfiledata *d;
    size_t s;

    d = (FILTfiledata*)FILTgetdata(filt, kFILTfiletype, name, &s, recurse);
    if (!d || !d->isopen)
        return(0);
    return(d->file);
}

char *
FILTgetfname(filt, name, recurse)
    FILT *filt;
    char *name;
    int recurse;
{
    FILTfiledata *d;
    size_t s;

    d = (FILTfiledata*)FILTgetdata(filt, kFILTfiletype, name, &s, recurse);
    if (!d)
        return(0);
    return(d->fname);
}

int
FILTisopenfile(filt, name, recurse)
    FILT *filt;
    char *name;
    int recurse;
{
    FILTfiledata *d;
    size_t s;

    d = (FILTfiledata*)FILTgetdata(filt, kFILTfiletype, name, &s, recurse);
    if (!d)
        return(0);
    return(d->isopen);
}

void
FILTclosefile(filt, name, recurse)
    FILT *filt;
    char *name;
    int recurse;
{
    FILTfiledata *d;
    size_t s;

    d = (FILTfiledata*)FILTgetdata(filt, kFILTfiletype, name, &s, recurse);
    if (!d || !d->isopen)
        return;
    fclose(d->file);
    d->isopen = 0;
    d->file = 0;
}

/* standard deletion functions */
void
FILTdelfiledata(type, data, s)
    int type;
    void *data;
    size_t s;
{
    FILTfiledata *d = (FILTfiledata*)data;
    free(d->fname);
    if (d->isopen)
        fclose(d->file);
    free((void*)d);
}

void
FILTdelfiltdata(type, data, s)
    int type;
    void *data;
    size_t s;
{
    FILT *foo;
    foo = *((FILT**)data);
    FILTdestroy(foo);
    free((void*)data);
}

/* add a deletion function */
void
FILTadddel(type, del)
    int type;
    FILTdelfunc del;
{
    FILTdellist *next = dellist;
    dellist = (FILTdellist*)malloc(sizeof(FILTdellist));
    dellist->next = next;
    dellist->type = type;
    dellist->func = del;
}

/* our config loading functions */
int
FILTsvar(name, arg, value, conf)
    char *name, *arg, *value;
    CONF *conf;
{
    if (!SUBSTglobvar)
        SUBSTglobvar = FILTconnect(0);
    FILTaddstring(SUBSTglobvar, arg, value, 0);
    return(1);
}
