/* Postincr V0.2 - 7/9/97 - Alexis Rosen/Aidan Cully
 * This procedure will increment the posting count database entry for the
 * user who invokes it by one.
 *
 * This code is extremely heavily commented for the sake of those whose C
 * skills are a bit rusty, but who want to investigate the code before
 * running it on their system.
 *
 * Sites running a shared spamcount database across multiple machines will
 * have to use a different locking scheme that the one used here (flock).
 * The only truly reliable technique I know is to use nfs link-to-file
 * type locks.
 *
 * SECURITY ANALYSIS:
 * This program runs as news (or whatever uid nnrpd runs as) so that the
 * database can be protected from random alteration by users, and yet be
 * modifiable by this program.
 * This procedure will only open an already-existing dbm file. It does not
 * create the dbm if it doesn't exist. Therefore, it is not subject to the
 * class of security holes based on race conditions between file creation
 * and modification. (Furthermore, the database normally resides in a
 * directory which users have no write access to, so it's doubly safe.)
 * This procedure will only modify a file which is in a format it recognizes.
 *
 * Because of this, it is safe to include this procedure in nrrpd. This
 * simplifies setup a bit.
 *
 * MODIFICATION HISTORY:
 * 7/9/97 -
 *     modified to become part of nnrpd
 * 8/18/97 -
 *     modified to eliminate race condition
 * 8/26/97 -
 *     modified to abstract away actual in/decrimenting procedures
 * 10/2/97 -
 *     modified to allow static data to exist in the database
 * 10/27/97 - Aidan Cully
 *     logs db open failures
 *
 */

#include <ndbm.h>
#include <sys/file.h>
#include <sys/types.h>
#include <sys/time.h>
#include <time.h>
#include <sysexits.h>
#include <pwd.h>
#include <stdio.h>

#include "filtpaths.h"
#include "logging.h"

/* Routine to lock the database. If you're not locking on an nfs-mounted
 * volume, or if your nfs locking works (it doesn't), the flock() call
 * provided is sufficient. Otherwise you'll need some sort of nfs-safe
 * (link-based) lock routine.
 */

DBM *D=0;
int db=-1;

void
lockdb(db)
  int db;
{
  flock(db, LOCK_EX);
}

/* Unlock the database. See comments above on lockdb(). */
void
unlockdb(db)
  int db;
{
  flock(db, LOCK_UN);
}

int
opendb()
{
  /* Setup: Open database as a regular file. Fail silently if the file
   * doesn't exist. Otherwise, try to lock it. Once we get the lock, open
   * it as a database.
   */
  if ((db = open(_PATH_POSTDB, O_RDWR)) == -1) {
    syslog(L_ERROR, "could not open post database lock file: %s %m", _PATH_POSTDB);
    return(0);
  }
  lockdb(db);
  if (! (D = dbm_open(_PATH_POSTDB, O_RDWR, 0600))) {
    syslog(L_ERROR, "could not open post database: %s %m", _PATH_POSTDB);
    unlockdb(db);
    close(db);
    return(0);
  }
  return(1);
}

void
closedb()
{
   if (!D)
      return;
   dbm_close(D);
   unlockdb(db);
   close(db);
   D = 0;
   db = -1;
}

/* granularity of the post database */
int
numperiods()
{
  return(2);
}

/* the value associated with a username */
void
numposts(user, ret)
  char *user, **ret;
{
  struct tm *t;
  time_t curtime;
  register char *i;
  datum key, data;	/* The key to look up, and space for the result */
  char keystr[255];	/* Storage for key.dptr */

  /* Database hasn't been opened. */
  if (!D) {
    ret[0] = 0;
    return;
  }
  /* First step: Figure out what record in the database we need to create
   * or modify. The key is built out of "A" or "P" (based on time of day),
   * then a "-", then the username.
   */
  key.dptr = keystr;
  (void)time(&curtime);		/* Get current time into "curtime" */
  t = localtime(&curtime);	/* Break time into various pieces */
  if (t->tm_hour < 12)	/* it's before noon */
    keystr[0] = 'A';
  else
    keystr[0] = 'P';
  keystr[1] = '-';
  key.dsize = 1;
  i = user;
  while (keystr[++key.dsize] = *i++);	/* copy name into key */

  /* Second step: Look up key in database */
  data = dbm_fetch(D, key);
  if (data.dptr) {
    ret[0] = (char*)malloc(data.dsize+1);
    strncpy(ret[0], data.dptr, data.dsize);
    ret[0][data.dsize] = 0;
  } else {
    ret[0] = (char*)malloc(1);
    ret[0][0] = 0;
  }

  /* Third step: Get the number of posts that happened last time period */
  if (keystr[0] == 'A') /* change to the other time period */
    keystr[0] = 'P';
  else
    keystr[0] = 'A';

  data = dbm_fetch(D, key);
  if (data.dptr) {
    ret[1] = (char*)malloc(data.dsize+1);
    strncpy(ret[1], data.dptr, data.dsize);
    ret[1][data.dsize] = 0;
  } else {
    ret[1] = (char*)malloc(1);
    ret[1][0] = 0;
  }
  ret[2] = 0;
}

void
setposts(user, val)
  char *user;
  char *val;
{
  struct tm *t;
  time_t curtime;
  register char *i;
  datum key, data;	/* The key to look up, and space for the result */
  char keystr[255];	/* Storage for key.dptr */

  /* First step: Figure out what record in the database we need to create
   * or modify. The key is built out of "A" or "P" (based on time of day),
   * then a "-", then the username.
   */
  key.dptr = keystr;
  (void)time(&curtime);		/* Get current time into "curtime" */
  t = localtime(&curtime);	/* Break time into various pieces */
  if (t->tm_hour < 12)	/* it's before noon */
    keystr[0] = 'A';
  else
    keystr[0] = 'P';
  keystr[1] = '-';
  key.dsize = 1;
  i = user;
  while (keystr[++key.dsize] = *i++);	/* copy name into key */

  /* Second step: Look up key in database. If it's there, increment the value
   * by one. If it's not, create it and set it to one.
   */
  if (!*val)
    dbm_delete(D, key);
  else {
    data.dptr = val;
    data.dsize = strlen(val);
    dbm_store(D, key, data, DBM_REPLACE);
  }
}

/* get static data for a user */
void
getstatic(user, str, len)
  char *user, *str;
  size_t len;
{
  char keystr[255];
  datum key, data;

  strcpy(keystr, "S-");
  strcat(keystr, user);
  key.dptr = keystr;
  key.dsize = strlen(keystr);

  data = dbm_fetch(D, key);
  if (data.dptr) {
    strncpy(str, data.dptr, data.dsize);
    str[data.dsize] = 0;
  } else
    str[0] = 0;
}

/* replace a user's static data */
void
setstatic(user, str)
  char *user, *str;
{
  char keystr[255];
  datum key, data;

  strcpy(keystr, "S-");
  strcat(keystr, user);
  key.dptr = keystr;
  key.dsize = strlen(keystr);

  if (!*str)
    dbm_delete(D, key);
  else {
    data.dptr = str;
    data.dsize = strlen(str);
    dbm_store(D, key, data, DBM_REPLACE);
  }
}
