/*
** catchup -- scans a directory for articles and eventually sends them off
** to the server
**
** every 30 seconds, goes through a directory associated with the current
** date.  When it encounters a file it hasn't met before, allocates an
** entry in a linked list for it saying how much time is left before the
** file gets sent to the news server.  When that time has come, sends the
** file to the server, deletes the physical file, then deletes the entry
** in the linked list.
**
** People can delete files in this directory to prevent them going out over
** the 'net, so this is built with the ground falling out from it in mind.
**
** Change log
**   19970811 Aidan Cully
**       Initial version
**   19970815 Aidan Cully
**       Working version ;-)
**   19970821 Aidan Cully
**       Uses syslog rather than fprintf to report errors
**   19970923 Dan Foster
**       AIX hacks
**   19970925 Aidan Cully
**       Now keeps logged files for an additional 3 days after sending them
**       to the server.
*/

#include <signal.h>
#include <fcntl.h>
#include <termios.h>

#ifdef _AIX
#include <sys/ioctl.h>
#endif

#include <stdio.h>
#include <time.h>
#include <string.h>
#include <dirent.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <syslog.h>
#include <errno.h>

#include "filtpaths.h"
#include "filter.h"

#include "catchup.h"
#include "configdata.h"

typedef int (*CUPclearfunc)(CUPent*);

void detach_terminal();
void load_config();
void init();
void die();
int check_dir();
void add_rmlist(CUPlist*);
int check_rment(CUPent*);
void checklist(CUPlist**, CUPclearfunc, int);

int reload; /* use this global to coordinate free/reload of config file */
char holddir[255];

CUPlist *dirlist;
CUPlist *rmlist;

FILT *curconn;

int
check_dir(dir, list)
    char *dir;
    CUPlist **list;
{
    char fname[255];
    char user[255];
    int delay;
    time_t t;
    DIR *d;
    struct dirent *ent;
    struct stat st;
    int fd, found, minnext = 30;
    CUPlist *iter;

    time(&t);

    d = opendir(dir);
    if (!d) {
        syslog(LOG_ERR, "Couldn't open directory %s.", dir);
        die();
    }
    while (ent = readdir(d)) {
        strcpy(fname, dir);
        strncat(fname, ent->d_name, ent->d_namlen);
        if (CUPparsefname(fname, user, sizeof(user), &delay) != 0)
            continue;
        found = 0;
        if (!*list) {
            /* first entry */
            *list = (CUPlist*)malloc(sizeof(CUPlist));
            (*list)->next = 0;
        } else {
            /* have we run into this file before? */
            iter = *list;
            while (!found&&iter) {
                if (!strcmp(iter->ent.fname, fname))
                    found = 1;
                /* since we're already going through the entire array, */
                /* figure out how much time we have before the next */
                /* check */
                if (iter->ent.timeleft < minnext)
                    minnext = iter->ent.timeleft;
                iter = iter->next;
            }
            if (!found) {
                /* no, so allocate space in our list for it */
                iter = (CUPlist*)malloc(sizeof(CUPlist));
                iter->next = *list;
                *list = iter;
            }
        }
        if (!found) {
            /* make sure we can open the file, fstat it later */
            if ((fd = open(fname, O_RDWR)) < 0) {
                iter = (*list)->next;
                free((void*)*list);
                *list = iter;
                continue;
            }
            if ((fstat(fd, &st) < 0) || !S_ISREG(st.st_mode)) {
                /* I don't think this should ever happen, but handle the */
                /* case if it does */
                iter = (*list)->next;
                free((void*)*list);
                *list = iter;
                close(fd);
                continue;
            }
            close(fd);
            (*list)->ent.delay = delay;
            (*list)->ent.user = (char*)malloc(strlen(user)+1);
            strcpy((*list)->ent.user, user);
            (*list)->ent.timeleft = delay-(t-st.st_ctime);
            (*list)->ent.fname = (char*)malloc(strlen(fname)+1);
            strcpy((*list)->ent.fname, fname);
            if ((*list)->ent.timeleft < minnext)
                minnext = (*list)->ent.timeleft;
            (*list)->ent.tries = 0;
        }
    }
    closedir(d);
    return (minnext);
}

int
main()
{
    int delay, delay2;
    CUPlist *iter, *store;

    detach_terminal();
    init();
    delay = check_dir(CONFkeywordvalue(FILTkeylist, "hold.dir"), &dirlist);
    while (1) {
        if (delay>0)
            sleep(delay);
        if (reload) {
            /* someone kill -HUP'ed us */
            if (FILTkeylist)
                CONFfree(FILTkeylist);
            FILTinit(0);
            if (!CONFkeywordvalue(FILTkeylist, "hold.dir")) {
                syslog(LOG_ERR, "No holding directory specified in %s.", _PATH_LIMITCONF);
                die();
            }
            strcpy(holddir, CONFkeywordvalue(FILTkeylist, "hold.dir"));
            reload = 0;
        }
        checklist(&dirlist, CUPsendfile, delay);
        checklist(&rmlist, check_rment, delay);
        delay = check_dir(CONFkeywordvalue(FILTkeylist, "hold.dir"), &dirlist);
        delay2 = check_dir(CONFkeywordvalue(FILTkeylist, "hold.deldir"), &rmlist);
        if (delay2 < delay)
            delay = delay2;
    }
}

void
init()
{
    FILE *pidf;

    reload = 0;
    dirlist = 0;
    rmlist = 0;
    FILTinit();
    pidf = fopen(_PATH_CATCHUPPID, "r");
    if (pidf) {
        fprintf(stderr, "Catchup already running!  If not, delete the file %s.\n", _PATH_CATCHUPPID);
        fclose(pidf);
        exit(1);
    }
    signal(SIGHUP, load_config);
    pidf = fopen(_PATH_CATCHUPPID, "w");
    fprintf(pidf, "%d\n", getpid());
    fclose(pidf);
    signal(SIGQUIT, die);
    signal(SIGTERM, die);
    signal(SIGINT, die);
#ifdef _AIX
    openlog("catchup", LOG_CONS|LOG_PID, LOG_DAEMON);
#else
    openlog("catchup", LOG_PERROR, LOG_DAEMON);
#endif
    curconn = FILTconnect(0);
}

void
load_config()
{
    reload = 1; /* let the main thread reload the configuration */
}

void
die()
{
    FILTdestroy(curconn);
    unlink(_PATH_CATCHUPPID);
    exit(0);
}

void
detach_terminal()
{
    int i;

    i = fork();
    if (i < 0) {
        fprintf(stderr, "Out of processes.\n");
        exit(1);
    }
    if (i > 0)
        exit(0);

    /* we're the child.  detach from tty */
    if ((i = open("/dev/tty", O_RDWR)) >= 0) {
        if (ioctl(i, TIOCNOTTY, (char*)NULL)) {
            /* TODO: log errors */
        }
        if (close(i) < 0) {
            /* TODO: log errors */
        }
        if (setsid() < 0) {
            /* TODO: log errors */
        }
    }
}

void
add_rmlist(iter)
    CUPlist *iter;
{
    char temp[255];
    char *newf, *slash;

    if (!rmlist) {
        iter->next = 0;
        rmlist = iter;
    } else {
        iter->next = rmlist;
        rmlist = iter;
    }
    rmlist->ent.timeleft = 60*60*24*3; /* 3 days */

    /* change directories */
    newf = 0;
    while (!newf) {
        CUPfilename(rmlist->ent.user, rmlist->ent.timeleft,
          temp, sizeof(temp),
          CONFkeywordvalue(FILTkeylist, "hold.deldir"));
        if (link(rmlist->ent.fname, temp) >= 0) {
            newf = (char*)malloc(strlen(temp)+1);
            strcpy(newf, temp);
        } else if (errno != EEXIST)
            break;
    }
    if (!newf) {
        rmlist = rmlist->next;
        free((void*)iter->ent.fname);
        free((void*)iter->ent.user);
        free((void*)iter);
        return;
    }
    unlink(rmlist->ent.fname);
    free((void*)rmlist->ent.fname);
    rmlist->ent.fname = newf;
}

int
check_rment(ent)
    CUPent *ent;
{
    unlink(ent->fname);
    free(ent->fname);
    free(ent->user);
    return(2);
}

void
checklist(list, func, dec)
    CUPlist **list;
    CUPclearfunc func;
    int dec;
{
    CUPlist *iter, *store;
    int ret;

    iter = *list;
    store = 0;
    /* let our file list know we've just been decremented. */
    while (iter) {
        if (dec>0)
            iter->ent.timeleft -= dec;
        if (iter->ent.timeleft <= 0) {
            /* send the file on */
            ret = func(&iter->ent);
            if (ret == 1) {
                if (iter == *list) {
                    *list = (*list)->next;
                    add_rmlist(iter);
                    iter = *list;
                    continue;
                } else {
                    store->next = iter->next;
                    add_rmlist(iter);
                    iter = store->next;
                    continue;
                }
            } else if (ret == 2) {
                if (iter == *list) {
                    *list = (*list)->next;
                    free((void*)iter);
                    iter = *list;
                    continue;
                } else {
                    store->next = iter->next;
                    free((void*)iter);
                    iter = store->next;
                    continue;
                }
            } else
                /* encountered an error sending the file.  Try again
                 * later. */
                iter->ent.timeleft = POST_LIMITDELAY;
        }
        if (!store)
            store = iter;
        else
            store = store->next;
        iter = iter->next;
    }
}
