/* limitfile.c -- Load and use extendable posting limits */
/* Revision history:
 * 8/19/97 -
 *     Started work
 * 8/26/97 -
 *     Now allows several kinds of info to be stored in the database
 * 9/3/97 -
 *     Now has "user groups" in the file
 * 9/12/97 -
 *     You can now specify which parameters get loaded.
 * 10/2/97 -
 *     Now will keep static data if anyone wants
 * 10/6/97 -
 *     Connection oriented
 * 10/13/97 -
 *     Synchronization functions supported
 * 10/20/97 - Aidan Cully
 *     Started work on supporting conditional limits
 */

#include <stdio.h>
#include <string.h>

#include "limitfile.h"
#include "configdata.h"
#include "misc.h"
#include "filter.h"
#include "proto.h"

#define kFILTlimtype	1

/* behavior constants and masks */
#define kLIMnothing	0x0
#define kLIMsetstat	0x1
#define kLIMclearstat	0x2
#define kLIMstatmask	0x3
#define kLIMsetdyn	0x4
#define kLIMcleardyn	0x8
#define kLIMdynmask	0xC
#define kLIMblock	0x10

typedef struct _LIMparamdata {
    char *name;
    char *statdata;
    char **timedata;
    char *newtimedata;
    char *newstatdata;
    LIMfinishfunc finish;
    LIMsyncfunc sf;
    int behavior;
} LIMparamdata;

typedef struct _LIMpriv {
    LIMparamdata *data;
    int shouldblock;
    int nument;
} LIMpriv;

typedef struct _LIMentlist {
    LIMent *entries;
    int nument;
} LIMentlist;

typedef struct _LIMtestlist {
    LIMtest *entries;
    int nument;
} LIMtestlist;

typedef struct _LIMfiltlist {
    char *name;
    char *param;
    LIMtest *test;
    FILT *filt;
    struct _LIMfiltlist *parent, *next;
} LIMfiltlist;

LIMentlist LIMparams = {0,0};
LIMtestlist LIMtests = {0,0};

typedef struct _LIMgrouplist {
    char *groupname;
    struct _LIMgrouplist *next;
} LIMgrouplist;

/* figure out if the limits have been loaded for this user */
int
LIMloaded(filt, name)
    FILT *filt;
    char *name;
{
    char scratch[255];

    sprintf(scratch, sLIMload, name);
    return FILThasint(filt, scratch, 0);
}

void
LIMload(filt, name, load)
    FILT *filt;
    char *name;
    int load;
{
    char scratch[255];

    if (!load)
        return;
    sprintf(scratch, sLIMload, name);
    FILTaddint(filt, scratch, 0, 0);
}

/* go through a list of groups to which a user belongs looking for the
 * group passed in */
int
LIMuseringroup(list, group)
    LIMgrouplist *list;
    char *group;
{
    while (list) {
        if (!strcmp(list->groupname, group))
            return(1);
        list = list->next;
    }
    return(0);
}

/* parse a group declaration line and determine if the user is in the
 * group it belongs to */
int
LIMuserindecl(list, user, decl)
    LIMgrouplist *list;
    char *user;
    char *decl;
{
    char *iter;
    char operator, save;
    int in, inpres, not, nest;

    /* we don't want to be overly complicated, so we won't define precedence
     * at all and simply move left to right updating the 'in' variable when
     * an operator changes it */
    operator = '+';
    in = 0;
    not = 0;
    if (!decl)
        return(0);
    while (*decl) {
        /* skip past whitespace */
        while (*decl && isspace(*decl))
            decl++;
        /* bang means invert the following */
        if (*decl == '!') {
            not = !not;
            decl++;
            continue;
        }

        if (*decl == '(') {
            /* recursively evaluate what's between parens */
            nest = 1;
            iter = decl+1;
            while (nest && *iter) {
                if (*iter == '(')
                    nest++;
                else if (*iter == ')')
                    nest--;
                if (nest)
                    iter++;
            }
            if (nest)
                return(0);
            *iter = 0;
            inpres = LIMuseringroup(decl+1);
            save = ')';
        } else {
            iter = decl;
            while (*iter && (!isspace(*iter)))
                iter++;
            save = *iter;
            *iter = 0;
            if (*decl == '$')
                inpres = LIMuseringroup(list, decl+1);
            else
                inpres = !strcmp(decl, user);
        }

        if (not)
            inpres = !inpres;
        *iter = save;

        /* inpres modifies in depending on the operator */
        switch (operator) {
        case '+': /* union */
            in = in || inpres;
            break;
        case '-': /* difference */
            in = in && (!inpres);
            break;
        case '|': /* intersection */
            in = in && inpres;
            break;
        default: /* unknown operator.  Barf */
            return(0);
        }
        not = 0;
        /* find the next operator (if one exists) and go through the loop
         * again */
        while (*iter && isspace(*iter))
            iter++;
        if (*iter) {
            operator = *iter;
            decl = iter+1;
        } else
            decl = iter;
    }
    return (in);
}

void
LIMaddgroup(list, group)
    LIMgrouplist *list;
    char *group;
{
    while (list->next)
        list = list->next;
    list->next = (LIMgrouplist*)malloc(sizeof(LIMgrouplist));
    list = list->next;
    list->next = 0;
    list->groupname = (char*)malloc(strlen(group)+1);
    strcpy(list->groupname, group);
}

int
LIMfindparam(str)
    char *str;
{
    int index, iter;

    for (index = 0; index < LIMparams.nument; index++)
        for (iter = 0; LIMparams.entries[index].param[iter]; iter++)
            if (!strcasecmp(LIMparams.entries[index].param[iter], str))
                return(index);
    return(-1);
}

int
LIMdouserline(filt, line, inuser, inbrace)
    FILT *filt;
    char *line;
    int inuser;
    int *inbrace;
{
    int index, done;
    int numparams = LIMparams.nument;
    LIMent *params = LIMparams.entries;
    char param[256], *iter, *close;

    iter = line;
    while (iter = PARSEgettoken(iter, param)) {
        if (!strcmp(param, "{"))
            *inbrace = 1;
        else if (!strcmp(param, "}"))
            *inbrace = 0;
        else if (!strcmp(param, ","))
            continue;
        else if (inuser) {
            index = LIMfindparam(param);
            if (index >= 0) {
                if (*iter == '(') {
                    if (!(close = STRfindrparen(iter+1)))
                        return(0);
                    *close = 0;
                    strncpy(param, iter+1, sizeof(param));
                    *close = ')';
                    iter = close+1;
                } else
                    *param = 0;
                if (!LIMloaded(filt, params[index].param[0]))
                    LIMload(filt, params[index].param[0],
                      params[index].pf(filt, params[index].param[0], param));
            }
        }
    }
    done = 1;
    for (index = 0; index<numparams; index++) {
        if (!params[index].pf)
            continue;
        if (!LIMloaded(filt, params[index].param[0]))
            done = 0;
    }
    return(done);
}

char *
LIMgrowdecl(decl, iter, inbrace)
    char *decl;
    char *iter;
    int *inbrace;
{
    char delim[256];

    while (iter = PARSEgettoken(iter, delim)) {
        if (!strcmp(delim, "{"))
            *inbrace = 1;
        else if (!strcmp(delim, "}"))
            *inbrace = 0;
        else {
            if (decl)
                decl = (char*)realloc(decl, strlen(decl)+strlen(delim)+2);
            else {
                decl = (char*)malloc(strlen(delim)+2);
                *decl = 0;
            }
            strcat(decl, delim);
            /* make sure we parse individual items as individual items */
            strcat(decl, " ");
        }
    }
    return(decl);
}

void
LIMloaduserlimits(filt)
    FILT *filt;
{
    char *user;
    FILE *limitfile;
    char buf[256], param[256], delim[256], grpname[256];
    char *iter, *grpdecl;
    int inbrace, inuser, ingroup, done, index;
    LIMgrouplist *groups, *del;

    if (!LIMparams.nument)
        return;
    user = FILTgetuser(filt);
    limitfile = fopen(CONFkeywordvalue(FILTkeylist, "limitfile"), "r");
    if (!limitfile)
        return;
    done = 0;
    inbrace = 0;
    inuser = 0;
    ingroup = 0;

    /* initialize groups user belongs to with '*' */
    groups = (LIMgrouplist*) malloc(sizeof(LIMgrouplist));
    groups->next = 0;
    groups->groupname = (char*)malloc(2);
    strcpy(groups->groupname, "*");

    while (!done && fgets(buf, sizeof(buf), limitfile)) {
        if (!(iter = MISCparseline(buf)))
            continue;
        buf[strlen(buf)-1] = 0;
        if (inbrace) {
            if (ingroup)
                grpdecl = LIMgrowdecl(grpdecl, iter, &inbrace);
            else {
                /* we're in the user declaration */
                done = LIMdouserline(filt, iter, inuser, &inbrace);
                continue;
            }
        } else if (ingroup) {
            if (LIMuserindecl(groups, user, grpdecl))
                LIMaddgroup(groups, grpname);
            free((void*)grpdecl);
            ingroup = 0;
        }
        /* always has at least two tokens on a line */
        if (!(iter = PARSEgettoken(iter, param)))
            continue;
        if (!(iter = PARSEgettoken(iter, delim)))
            continue;
        if (!strcmp(delim, ":")) {
            /* user declaration */
            if (*param == '$')
                inuser = LIMuseringroup(groups, param+1);
            else
                inuser = !strcmp(param, user);
            done = LIMdouserline(filt, iter, inuser, &inbrace);
            continue;
        } else if (!strcmp(delim, "=")) {
            strcpy(grpname, param);
            grpdecl = 0;
            grpdecl = LIMgrowdecl(grpdecl, iter, &inbrace);
            ingroup = 1;
        }
    }

    /* free up space associated with our user's groups */
    while (groups) {
        del = groups;
        groups = groups->next;
        free((void*)del->groupname);
        free((void*)del);
    }
}

/* add a new data el to the database */
int
LIMadddata(lim, name)
    LIMpriv *lim;
    char *name;
{
    int index;

    index = lim->nument++;
    lim->data = (LIMparamdata*)realloc(lim->data, sizeof(LIMparamdata)*lim->nument);
    lim->data[index].timedata = 0;
    lim->data[index].statdata = 0;
    lim->data[index].newstatdata = 0;
    lim->data[index].newtimedata = 0;
    lim->data[index].sf = 0;
    lim->data[index].finish = 0;
    lim->data[index].behavior = kLIMnothing;
    lim->data[index].name = (char*)malloc(strlen(name)+1);
    strcpy(lim->data[index].name, name);
    return(index);
}

/* delete a data el from the database */
void
LIMdeldata(lim, name)
    LIMpriv *lim;
    char *name;
{
    /* TODO: this function */
}

/* fill in data for every parameter at the outset, in case someone needs
 * the data for someone else's parameter */
void
LIMfilldata(lim, timedata, statdata)
    LIMpriv *lim;
    char **timedata;
    char *statdata;
{
    int index, index2, index3;
    int found;
    char *iter, *lparen;
    char scratch[255];

    lim->shouldblock = 0;
    lim->data = 0;
    lim->nument = 0;
    for (index = 0; index < numperiods()+1; index++) {
        if (index < numperiods())
            /* stick in the time data for everyone */
            iter = timedata[index];
        else
            iter = statdata;
        while (iter = PARSEgettoken(iter, scratch)) {
            found = 0;
            index2 = 0;
            /* find the name of this jobby in our data, if possible */
            while (!found && (index2 < lim->nument)) {
                if (!strcmp(lim->data[index2].name, scratch))
                    found = 1;
                else
                    index2++;
            }
            if (!found) {
                /* initialize a new parameter */
                index2 = LIMadddata(lim, scratch);
                if (index < numperiods()) {
                    /* we'll need space for the time data */
                    lim->data[index2].timedata =
                      (char**)malloc(numperiods()*sizeof(char*));
                    for (index3 = 0; index3 < numperiods(); index3++)
                        lim->data[index2].timedata[index3] = 0;
                }
            }
            /* index2 now points at the current parameter */
            if (*iter == '(') {
                lparen = iter;
                if (!(iter = STRfindrparen(iter+1)))
                    break;
                *iter = 0;
                if (index < numperiods()) {
                    lim->data[index2].timedata[index] =
                      (char*)malloc(strlen(lparen+1)+1);
                    strcpy(lim->data[index2].timedata[index], lparen+1);
                } else {
                    lim->data[index2].statdata =
                      (char*)malloc(strlen(lparen+1)+1);
                    strcpy(lim->data[index2].statdata, lparen+1);
                }
                *iter = ')';
            } else {
                if (index < numperiods()) {
                    lim->data[index2].timedata[index] = (char*)malloc(1);
                    *lim->data[index2].timedata[index] = 0;
                } else {
                    lim->data[index2].statdata = (char*)malloc(1);
                    *lim->data[index2].statdata = 0;
                }
            }
            iter++;
        }
    }
}

/* write to the database */
void
LIMwritedb(filt, lim)
    FILT *filt;
    LIMpriv *lim;
{
    int index, index2;
    int tmp;
    char tmnew[255], statnew[255];

    *statnew = 0;
    *tmnew = 0;
    tmp = numperiods()-1;
    for (index = 0; index<lim->nument; index++) {
        switch(lim->data[index].behavior & kLIMstatmask) {
        case kLIMsetstat:
            strcat(statnew, lim->data[index].name);
            if (*lim->data[index].newstatdata) {
                strcat(statnew, "(");
                strcat(statnew, lim->data[index].newstatdata);
                strcat(statnew, ")");
            } else
                strcat(statnew, ",");
            break;
        case kLIMnothing:
            if (lim->data[index].statdata) {
                strcat(statnew, lim->data[index].name);
                if (*lim->data[index].statdata) {
                    strcat(statnew, "(");
                    strcat(statnew, lim->data[index].statdata);
                    strcat(statnew, ")");
                } else
                    strcat(statnew, ",");
            }
        }
        switch(lim->data[index].behavior & kLIMdynmask) {
        case kLIMsetdyn:
            strcat(tmnew, lim->data[index].name);
            if (*lim->data[index].newtimedata) {
                strcat(tmnew, "(");
                strcat(tmnew, lim->data[index].newtimedata);
                strcat(tmnew, ")");
            } else
                strcat(tmnew, ",");
            break;
        case kLIMnothing:
            if (lim->data[index].timedata && lim->data[index].timedata[tmp]) {
                strcat(tmnew, lim->data[index].name);
                if (*lim->data[index].timedata[tmp]) {
                    strcat(tmnew, "(");
                    strcat(tmnew, lim->data[index].timedata[tmp]);
                    strcat(tmnew, ")");
                } else
                    strcat(tmnew, ",");
            }
        }
    }
    setposts(FILTgetuser(filt), tmnew);
    setstatic(FILTgetuser(filt), statnew);
}

void
LIMfree(lim)
    LIMpriv *lim;
{
    int index, index2;

    if (!lim)
        return;
    if (!lim->data) {
        free((void*)lim);
        return;
    }
    for (index = 0; index < lim->nument; index++) {
        if (lim->data[index].timedata) {
            for (index2 = 0; index2 < numperiods(); index2++)
                if (lim->data[index].timedata[index2])
                    free((void*)lim->data[index].timedata[index2]);
            free((void*)lim->data[index].timedata);
        }
        if (lim->data[index].newtimedata)
            free((void*)lim->data[index].newtimedata);
        if (lim->data[index].statdata)
            free((void*)lim->data[index].statdata);
        if (lim->data[index].newstatdata)
            free((void*)lim->data[index].newstatdata);
    }
    free((void*)lim->data);
    free((void*)lim);
}

int
LIMshouldsend(filt)
    FILT *filt;
{
    size_t s;

    LIMpriv *lim = (LIMpriv*)FILTgetdata(filt, kFILTlimtype, sLIM, &s, 1);
    if (!lim)
        return(1);
    return !((LIMpriv*)lim)->shouldblock;
}

/* hook into the detection functions people registered with us before */
void
LIMtrackpost(filt)
    FILT *filt;
{
    int index;
    char **dbold, statdata[255];
    char *user;
    LIMent *params;
    LIMpriv lim;

    if (!(LIMparams.nument && (user = FILTgetuser(filt))))
        return;
    if (!opendb())
        return;

    lim.data = 0;
    /* we're expecting these guys to drop their error messages into the
     * error field */
    ERRORclearerror(filt);

    /* allocate space for old data */
    dbold = (char**)malloc((numperiods()+1)*sizeof(char*));

    /* numposts will fill allocate and fill in data for every other period
     * we've got */
    numposts(user, dbold);
    getstatic(user, statdata, sizeof(statdata));

    LIMfilldata(&lim, dbold, statdata);

    /* clean up */
    for (index = 0; index<numperiods(); index++)
        free((void*)dbold[index]);
    free((void*)dbold);

    params = LIMparams.entries;
    for (index = 0; index<LIMparams.nument; index++) {
        if (LIMloaded(filt, params[index].param[0]) && params[index].df)
            params[index].df(filt, (LIM*)&lim, params[index].param[0]);
    }

    /* give these guys a chance to resolve any requests for the same
     * resource */
    for (index = 0; index<lim.nument; index++) {
        if (lim.data[index].sf)
            lim.data[index].sf(filt, (LIM*)&lim, lim.data[index].name);
    }

    /* the structure will be freed when LIMfinish is called, but write it
     * now so that we release the lock as soon as possible */
    LIMwritedb(filt, &lim);
    closedb();
    FILTadddata(filt, kFILTlimtype, sLIM, (void*)&lim, sizeof(lim), 0);
}

/* call the finisher functions and free the LIM* */
void
LIMfinish(filt)
    FILT *filt;
{
    LIMpriv *lim;
    int index;
    size_t s;

    lim = (LIMpriv*)FILTgetdata(filt, kFILTlimtype, sLIM, &s, 1);
    if (!lim)
        return;
    for (index = 0; index < lim->nument; index++)
        if (lim->data[index].finish)
            lim->data[index].finish(filt, (LIM*)lim,
              lim->data[index].name);
    LIMfree(lim);
}

void
LIMsetparams(ent, nument)
    LIMent *ent;
    int nument;
{
    LIMparams.entries = ent;
    LIMparams.nument = nument;
}

void
LIMsettests(ent, nument)
    LIMtest *ent;
    int nument;
{
    LIMtests.entries = ent;
    LIMtests.nument = nument;
}

/* database accessor functions.  Control what goes into the persistant
 * database */
int
LIMfindname(lim, name)
    LIMpriv *lim;
    char *name;
{
    int index;

    for (index = 0; index<lim->nument; index++)
        if (!strcmp(lim->data[index].name, name))
            return(index);
    return(-1);
}

char **
LIMdata(lim, name)
    LIM *lim;
    char *name;
{
    int index;

    if ((index = LIMfindname((LIMpriv*)lim, name)) < 0)
        return(0);
    return ((LIMpriv*)lim)->data[index].timedata;
}

char *
LIMnewdata(lim, name)
    LIM *lim;
    char *name;
{
    int index;

    if ((index = LIMfindname((LIMpriv*)lim, name)) < 0)
        return(0);
    return ((LIMpriv*)lim)->data[index].newtimedata;
}

void
LIMsetdata(lim, name, data)
    LIM *lim;
    char *name;
    char *data;
{
    int index;
    LIMpriv *rlim;

    rlim = (LIMpriv*)lim;
    if ((index = LIMfindname(rlim, name)) < 0)
        index = LIMadddata(rlim, name);
    rlim->data[index].behavior |= kLIMsetdyn;
    rlim->data[index].behavior &= ~kLIMcleardyn;
    if (rlim->data[index].newtimedata)
        free((void*)rlim->data[index].newtimedata);
    rlim->data[index].newtimedata = (char*)malloc(strlen(data)+1);
    strcpy(rlim->data[index].newtimedata, data);
}

void
LIMcleardata(lim, name)
    LIM *lim;
    char *name;
{
    int index;
    LIMpriv *rlim;

    rlim = (LIMpriv*)lim;
    if ((index = LIMfindname(rlim, name)) < 0)
        return;
    rlim->data[index].behavior |= kLIMcleardyn;
    rlim->data[index].behavior &= ~kLIMsetdyn;
    if (rlim->data[index].newtimedata)
        free((void*)rlim->data[index].newtimedata);
    rlim->data[index].newtimedata = 0;
}

void
LIMresetdata(lim, name)
    LIM *lim;
    char *name;
{
    int index;
    LIMpriv *rlim;

    rlim = (LIMpriv*)lim;
    if ((index = LIMfindname(rlim, name)) < 0)
        return;
    rlim->data[index].behavior &= ~kLIMdynmask;
    if (rlim->data[index].newtimedata)
        free((void*)rlim->data[index].newtimedata);
    rlim->data[index].newtimedata = 0;
}

char *
LIMstatdata(lim, name)
    LIM *lim;
    char *name;
{
    int index;

    if ((index = LIMfindname((LIMpriv*)lim, name)) < 0)
        return(0);
    return ((LIMpriv*)lim)->data[index].statdata;
}

char *
LIMnewstatdata(lim, name)
    LIM *lim;
    char *name;
{
    int index;

    if ((index = LIMfindname((LIMpriv*)lim, name)) < 0)
        return(0);
    return ((LIMpriv*)lim)->data[index].newstatdata;
}

void
LIMsetstatdata(lim, name, data)
    LIM *lim;
    char *name;
    char *data;
{
    int index;
    LIMpriv *rlim;

    rlim = (LIMpriv*)lim;
    if ((index = LIMfindname(rlim, name)) < 0)
        index = LIMadddata(rlim, name);
    rlim->data[index].behavior |= kLIMsetstat;
    rlim->data[index].behavior &= ~kLIMclearstat;
    if (rlim->data[index].newstatdata)
        free((void*)rlim->data[index].newstatdata);
    rlim->data[index].newstatdata = (char*)malloc(strlen(data)+1);
    strcpy(rlim->data[index].newstatdata, data);
}

void
LIMclearstatdata(lim, name)
    LIM *lim;
    char *name;
{
    int index;
    LIMpriv *rlim;

    rlim = (LIMpriv*)lim;
    if ((index = LIMfindname(rlim, name)) < 0)
        return;
    rlim->data[index].behavior |= kLIMclearstat;
    rlim->data[index].behavior &= ~kLIMsetstat;
    if (rlim->data[index].newstatdata)
        free((void*)rlim->data[index].newstatdata);
    rlim->data[index].newstatdata = 0;
}

void
LIMresetstatdata(lim, name)
    LIM *lim;
    char *name;
{
    int index;
    LIMpriv *rlim;

    rlim = (LIMpriv*)lim;
    if ((index = LIMfindname(rlim, name)) < 0)
        return;
    rlim->data[index].behavior &= ~kLIMstatmask;
    if (rlim->data[index].newstatdata)
        free((void*)rlim->data[index].newstatdata);
    rlim->data[index].newstatdata = 0;
}

void
LIMblock(lim, name)
    LIM *lim;
    char *name;
{
    int index;
    LIMpriv *rlim;

    rlim = (LIMpriv*)lim;
    if ((index = LIMfindname(rlim, name)) < 0)
        index = LIMadddata(rlim, name);
    rlim->data[index].behavior |= kLIMblock;
    rlim->shouldblock = 1;
}

void
LIMclearblock(lim, name)
    LIM *lim;
    char *name;
{
    int index, rblock;
    LIMpriv *rlim;

    rlim = (LIMpriv*)lim;
    if ((index = LIMfindname(rlim, name)) < 0)
        return;
    rlim->data[index].behavior &= ~kLIMblock;
    rblock = 0;
    for (index = 0; !rblock && (index < rlim->nument); index++) {
        if (rlim->data[index].behavior & kLIMblock)
            rblock = 1;
    }
    if (!rblock)
        rlim->shouldblock = 1;
}

int
LIMisblock(lim, name)
    LIM *lim;
    char *name;
{
    int index, blockthis, iter;
    LIMpriv *rlim;

    rlim = (LIMpriv*)lim;
    index = LIMfindname(rlim, name);
    blockthis = 0;
    for (iter = 0; iter < rlim->nument; iter++)
        if (rlim->data[iter].behavior & kLIMblock) {
            if (iter == index)
                blockthis |= kLIMblockthis;
            else
                blockthis |= kLIMblockelse;
        }
    return blockthis;
}

/* finisher functions */
void
LIMaddfinish(lim, name, finish)
    LIM *lim;
    char *name;
    LIMfinishfunc finish;
{
    LIMpriv *rlim;
    int index;

    rlim = (LIMpriv*)lim;
    if ((index = LIMfindname(rlim, name)) < 0)
        index = LIMadddata(rlim, name);
    rlim->data[index].finish = finish;
}

void
LIMresetfinish(lim, name)
    LIM *lim;
    char *name;
{
    LIMaddfinish(lim, name, 0);
}

void
LIMaddsync(lim, name, sync)
    LIM *lim;
    char *name;
    LIMsyncfunc sync;
{
    LIMpriv *rlim;
    int index;

    rlim = (LIMpriv*)lim;
    if ((index = LIMfindname(rlim, name)) < 0)
        index = LIMadddata(rlim, name);
    rlim->data[index].sf = sync;
}

void
LIMresetsync(lim, name)
    LIM *lim;
    char *name;
{
    LIMaddsync(lim, name, 0);
}
