#include <stdio.h>
#include <string.h>

#include "config.h"
#include "configdata.h"
#include "misc.h"

typedef struct _CONFkeywordlist {
    CONFkeyword key;
    struct _CONFkeywordlist *next;
} CONFkeywordlist;

typedef struct _CONFfile {
    CONFkeywordlist *keywords;
    char *filename;
} CONFfile;

typedef struct _CONFloadarr {
    CONFkeyload *list;
    int size;
} CONFloadarr;

CONFloadarr loaders = {0,0};

/*
** The configuration file is stored as a bunch of keyword:value pairs..  These
** keywords are not strictly defined by the main filter engine so that
** filters people write later can define new keywords or make use of keywords
** already defined.  This function inserts a keyword and a value into the
** global keyword list and returns a pointer to the keyword structure just
** created.
*/
void
AddKeyword(conf,keyword, value, lineno)
    CONFfile *conf;
    char *keyword, *value;
    int lineno;
{
    CONFkeywordlist *iter;
    char *lparen, *rparen;
    int index;

    if ((!keyword) || (!conf))
        return;

    lparen = strchr(keyword, '(');
    if (lparen) {
        *lparen = 0;
        rparen = STRfindrparen(lparen+1);
        *rparen = 0;
        for (index = 0; index < loaders.size; index++)
            if (!strcmp(loaders.list[index].name, keyword)) {
                loaders.list[index].func(loaders.list[index].name,
                  lparen+1, value, (CONF*)conf);
                return;
            }
        return;
    }

    iter = conf->keywords;

    if (!iter) {
        /* allocate space for the first keyword */
        iter = (CONFkeywordlist*)malloc(sizeof(CONFkeywordlist));
        iter->key.keyword = (char*)malloc(strlen(keyword)+1);
        strcpy(iter->key.keyword, keyword);
        if (value) {
            iter->key.value = (char*)malloc(strlen(value)+1);
            strcpy(iter->key.value, value);
        } else
            iter->key.value = 0;
        iter->key.used = 0;
        iter->key.lineno = lineno;
        iter->next = 0;
        conf->keywords = iter;
        return;
    }

    /* Whoops.. We've added this keyword before.  Return NULL for error */
    if (!strcmp(iter->key.keyword, keyword))
        return;
    while (iter->next) {
       iter = iter->next;
       if (!strcmp(iter->key.keyword, keyword))
           return;
    }
    iter->next = (CONFkeywordlist*)malloc(sizeof(CONFkeywordlist));
    iter = iter->next;

    if (!iter)
        return;

    /* store the values we've just generated here */
    iter->key.keyword = (char*)malloc(strlen(keyword)+1);
    strcpy(iter->key.keyword, keyword);
    if (value) {
        iter->key.value = (char*)malloc(strlen(value)+1);
        strcpy(iter->key.value, value);
    } else
        iter->key.value = 0;

    /* So far, nothing uses this keyword */
    iter->key.used = 0;
    iter->key.lineno = lineno;
    iter->next = 0;
    return;
}

/*
** In order to present the best interface we can to the user of this program,
** we make sure that all keywords in the configuration file are used by
** something.  This function marks a keyword as used so that later, if we
** find an unused keyword we can abort with an appropriate error message.
** This procedure does not kill the program if the keyword doesn't exist
** to allow for optional keywords.  If you require a keyword, it is your
** responsibility to check for its presence and die if it doesn't exist.
*/
void
CONFkeywordused(conf, keyword)
    CONF *conf;
    char *keyword;
{
    CONFkeywordlist *iter = ((CONFfile*)conf)->keywords;

    if (!iter)
        return;
    while (iter) {
        if (!strcmp (iter->key.keyword, keyword)) {
            iter->key.used = 1;
            break;
        }
        iter = iter->next;
    }
}

/*
** Get the value associated with a keyword.  Marks the keyword as used.
*/
char *
CONFkeywordvalue(conf, keyword)
    CONF *conf;
    char *keyword;
{
    CONFkeywordlist *iter = ((CONFfile*)conf)->keywords;

    if (!iter)
        return(0);
    while (iter) {
        if (!strcmp(iter->key.keyword, keyword)) {
            iter->key.used = 1;
            return(iter->key.value);
        }
        iter = iter->next;
    }
    return(0);
}

/*
** Make sure all our keywords are used.
*/
CONF *
CONFcheckkeywords(conf)
    CONF *conf;
{
    CONFkeywordlist *inst = ((CONFfile*)conf)->keywords;
    CONFfile *dead = 0;

    while (inst) {
        if (!inst->key.used) {
            if (!dead) {
                dead = (CONFfile*)malloc(sizeof(CONFfile));
                dead->keywords = 0;
                dead->filename = (char*)malloc(strlen(CONFfilename(conf))+1);
                strcpy(dead->filename, CONFfilename(conf));
            }
            AddKeyword(dead, inst->key.keyword, inst->key.value, inst->key.lineno);
        }
        inst = inst->next;
    }
    return((CONF*)dead);
}

/*
*/
CONF *
CONFload(fname)
    char *fname;
{
    FILE *f;
    char buf[256];
    char title[256];
    char *iter, *colon;
    char *last;
    CONFfile *head;
    int slashed;
    int lineno, llineno;

    /* open and parse the config file */
    f=fopen(fname, "r");
    if (!f)
        return(0);
    slashed = 0;
    last = 0;
    head = (CONFfile*)malloc(sizeof(CONFfile));
    /* let the AddKeyword function know this is a new keyword list */
    head->keywords = 0;
    head->filename = (char*)malloc(strlen(fname)+1);
    strcpy(head->filename, fname);
    lineno = 0;
    while (iter = fgets(buf, sizeof(buf), f)) {
        lineno++;
        iter[strlen(iter)-1] = 0; /* remove the newline */

        /* Are we continuing an old value? */
        if (slashed) {

            /* Grow the last value we entered */
            last=(char*)realloc(last, strlen(last)+3+strlen(iter));
            if (!last)
                return(0);
            strcat(last, "\n");
            slashed = iter[strlen(iter)-1]=='\\';
            if (slashed)
                iter[strlen(iter)-1] = 0;
            strcat(last, iter);
        } else {

            /* Skip past whitespace and empty lines */
            if (!(iter = MISCparseline(iter)))
                continue;

            /* We interpret a slash on the end of the line as a request to
             * continue the value input */
            llineno = lineno;
            slashed = iter[strlen(iter)-1]=='\\';
            if (slashed)
                iter[strlen(iter)-1] = 0; /* remove the slash */
            colon = strchr(iter, ':');
            if (colon) {
                *colon = 0;
                strcpy(title, iter);
                last = (char*)malloc(strlen(colon+1)+1);
                strcpy(last, colon+1);
            } else
                return(0);
        }
        if (!slashed) {
            AddKeyword(head, title, last, llineno);
            free((void*)last);
        }
    }
    fclose(f);
    return((CONF*)head);
}

void
CONFfree(conf)
    CONF *conf;
{
    CONFkeywordlist *head = ((CONFfile*)conf)->keywords;
    CONFkeywordlist *iter = head;

    while (iter) {
        head = iter;
        iter = iter->next;
        free(head->key.keyword);
        if (head->key.value)
            free(head->key.value);
    }
    free((void*)((CONFfile*)conf)->filename);
    free((void*)conf);
}

CONFkeyword *
CONFfirstkey(conf)
    CONF *conf;
{
    if ((!conf)||(!((CONFfile*)conf)->keywords))
        return(0);
    return(&((CONFfile*)conf)->keywords->key);
}

CONFkeyword *
CONFnextkey(conf, key)
    CONF *conf;
    CONFkeyword *key;
{
    CONFkeywordlist *list;
    if ((!conf) || (!((CONFfile*)conf)->keywords))
        return(0);
    list = ((CONFfile*)conf)->keywords;
    while (list->next && (&(list->key) != key))
        list = list->next;
    if (!list->next)
        return(0);
    return(&list->next->key);
}

char *
CONFfilename(conf)
    CONF *conf;
{
    return(((CONFfile*)conf)->filename);
}

/*
** loading functions
*/
void
CONFsetload(keys, size)
    CONFkeyload *keys;
    int size;
{
    loaders.list = keys;
    loaders.size = size;
}
