#include <sys/types.h>
#include <sys/select.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <netdb.h>
#include <syslog.h>
#include <signal.h>
#include <fcntl.h>
#include <errno.h>

#include <stdio.h>
#include <dirent.h>
#include <string.h>

#include "config.h"
#include "paths.h"

static void config(void);
static void hupself(void);
static void killself(void);
static void chldself(void);

static void handle_line(char*);
static void time_pass(struct timeval*);
static void scan_directory(char*);
static void get_delay(struct timeval *delay);

static char *hostname;
static char *port;
static int need_reconfig;

static int newssock;
#define STATE_NOCONN -1
#define STATE_STARTUP 0
#define STATE_NORM 1
#define STATE_STARTICLE 2
#define STATE_ARTICLE 3
#define STATE_SHUTDOWN 4
static int state;

static char readbuf[4096];
static int readpos;

typedef struct _wrbuf_s {
    char *writebuf;
    int todo, curpos;
    void *data;
    void (*abort_func)(struct _wrbuf_s*);
    void (*finish_func)(struct _wrbuf_s*, int, char*);
    struct _wrbuf_s *next;
} wrbuf_s;
wrbuf_s *write_head;

static void handle_line(char *line)
{
    int respcode;
    char *explbuf;
    wrbuf_s *lastbuf;

    for (explbuf = line; *explbuf && !isspace(*explbuf); explbuf++)
	if (!isdigit(*explbuf)) {
	    syslog(LOG_NOTICE, "%s:%s Got non-numeric response code, disconnecting: %s",
	      hostname, port, line);
	    close(newssock);
	    if (write_head) {
		write_head->todo += write_head->curpos;
		write_head->curpos = 0;
		if (write_head->abort_func)
		    write_head->abort_func(write_head);
	    }
	    newssock = -1;
	    state = STATE_NOCONN;
	    return;
	}
    if (*explbuf)
	explbuf++;
    respcode = atoi(line);
    switch (state) {
      case STATE_STARTUP:
	if (line[0] != '2') {
	    /* one of the various 'Permission Denied' errors */
	    syslog(LOG_WARNING, "%s:%s Permission denied opening connection: %s",
	      hostname, port, line);
	    close(newssock);
	    newssock = -1;
	    if (write_head) {
		write_head->todo += write_head->curpos;
		write_head->curpos = 0;
		if (write_head->abort_func)
		    write_head->abort_func(write_head);
	    }
	} else {
	    if (line[2] != '0') {
		syslog(LOG_NOTICE, "%s:%s Can read, but not post: %s",
		  hostname, port, line);
		close(newssock);
		newssock = -1;
		if (write_head) {
		    write_head->todo += write_head->curpos;
		    write_head->curpos = 0;
		    if (write_head->abort_func)
			write_head->abort_func(write_head);
		}
	    } else {
		if (write_head)
		    state = STATE_STARTICLE;
		else
		    state = STATE_NORM;
	    }
	}
	break;
      case STATE_STARTICLE:
      case STATE_ARTICLE:
	if (write_head) {
	    if (write_head->finish_func && !write_head->todo)
		write_head->finish_func(write_head, respcode, line);
	    free((void*) write_head->writebuf);
	    lastbuf = write_head;
	    write_head = write_head->next;
	    free((void*) lastbuf);
	} else {
	    syslog(LOG_WARNING, "%s:%s Funny condition: in article write state with nothing that put us there.",
	      hostname, port);
	}
	break;
      case STATE_NORM:
	/* the server sent us some unsolicited response */
	syslog(LOG_WARNING, "%s:%s Unexpected input from server: %s",
	  hostname, port, line);
	break;
      case STATE_SHUTDOWN:
	/* sent us data _after_ verifying a shutdown? */
	syslog(LOG_WARNING, "%s:%s Unexpected input from server after shutdown: %s",
	  hostname, port, line);
	break;
      default:
	syslog(LOG_NOTICE, "%s:%s Got data in unknown state: %s",
	  hostname, port, line);
	break;
    }
}

int main(int argc, char *argv[])
{
    extern int optind;
    extern char *optarg;
    int opt;
    fd_set rdfds, wrfds;
    struct timeval tmout, stsel, endsel;
    int got;
    char *nl, *cp;

    hostname = "127.0.0.1";
    port = "nntp";
    need_reconfig = 0;
    newssock = -1;

    while ((opt = getopt(argc, argv, "h:p:")) != -1) {
	switch (opt) {
	  case 'p':
	    port = optarg;
	    break;
	  case 'h':
	    hostname = optarg;
	    break;
	  default:
	    break;
	}
    }
    config();
    openlog("catchup", 0, LOG_NEWS);
    readpos = 0;
    write_head = 0;
    scan_directory(SPOOLDIR "/hold");
    (void) signal(SIGCHLD, chldself);
    while (1) {
	if (newssock >= 0) {
	    FD_ZERO(&rdfds);
	    FD_SET(newssock, &rdfds);
	    FD_ZERO(&wrfds);
	    if (write_head && write_head->todo)
		FD_SET(newssock, &wrfds);
	    get_delay(&tmout);
	} else {
	    /* wait 30 secs, then try to connect again. */
	    sleep(30);
	    config();
	    tmout.tv_sec = 30;
	    tmout.tv_usec = 0;
	    time_pass(&tmout);
	    continue;
	}
	gettimeofday(&stsel, 0);
	got = select(newssock+1, &rdfds, &wrfds, 0, &tmout);
	gettimeofday(&endsel, 0);
	tmout.tv_sec = endsel.tv_sec - stsel.tv_sec;
	if (endsel.tv_usec < stsel.tv_usec) {
	    tmout.tv_sec--;
	    tmout.tv_usec = endsel.tv_usec + 1000000 - stsel.tv_usec;
	} else
	    tmout.tv_usec = endsel.tv_usec - stsel.tv_usec;
	time_pass(&tmout);
	if (got < 0) {
	    if (errno == EINTR)
		continue;
	    break;
	} else if (got == 0) {
	    /* timer ran out */
	    scan_directory(SPOOLDIR "/hold");
	    continue;
	}
	/* read data */
	if (FD_ISSET(newssock, &rdfds)) {
	    got = read(newssock, readbuf+readpos, sizeof(readbuf)-readpos);
	    if (got <= 0) {
		/* encountered EOF */
		switch (state) {
		  case STATE_SHUTDOWN:
		    /* we expect an EOF in this situation */
		    break;
		  case STATE_STARTUP:
		    syslog(LOG_NOTICE, "%s:%s Got EOF starting connection",
		      hostname, port);
		    break;
		  case STATE_NORM:
		  case STATE_STARTICLE:
		  case STATE_ARTICLE:
		    if (write_head) {
			syslog(LOG_NOTICE, "%s:%s Got unexpected EOF while sending data.",
			  hostname, port);
			if (write_head->curpos) {
			    write_head->todo += write_head->curpos;
			    write_head->curpos = 0;
			}
			if (write_head->abort_func)
			    write_head->abort_func(write_head);
		    } else {
			syslog(LOG_NOTICE, "%s:%s Got unexpected EOF",
			  hostname, port);
		    }
		    break;
		  default:
		    syslog(LOG_WARNING, "%s:%s Got EOF from %s:%s in unknown state: %d",
		      hostname, port, state);
		    break;
		}
		close(newssock);
		readpos = 0;
		/* we'll open another connection when we have to. */
		state = STATE_NOCONN;
		newssock = -1;
	    } else {
		readpos += got;
		while (nl = memchr(readbuf, '\n', readpos)) {
		    *nl++ = '\0';
		    /* deal with this line */
		    handle_line(readbuf);
		    cp = readbuf;
		    while (nl < readbuf+readpos)
			*cp++ = *nl++;
		    readpos = cp-readbuf;
		}
	    }
	}
	/* write data.  if we got some exceptional condition above, newssock
	 * could be closed and -1 now. */
	if (newssock >= 0 && FD_ISSET(newssock, &wrfds)) {
	    got = write(newssock, write_head->writebuf+write_head->curpos,
	      write_head->todo);
	    if (got <= 0) {
		syslog(LOG_NOTICE, "%s:%s Got error on write: %m",
		  hostname, port);
		close(newssock);
		readpos = 0;
		/* we'll open another connection when we have to. */
		state = STATE_NOCONN;
		newssock = -1;
	    } else {
		write_head->todo -= got;
		write_head->curpos += got;
	    }
	}
    }
    return(0);
}

typedef struct _filelist_s {
    char *filename;
    char *user;
    struct timeval delay;
#define FILESTATE_HOLD 0
#define FILESTATE_SEND 1
#define FILESTATE_DELETE 2
#define FILESTATE_FLAG 128
    int state;
    struct _filelist_s *next;
} filelist_s;
filelist_s *filehead = 0;

static void send_article(filelist_s*);
static int parse_filename(char*, char**, int*);

static void time_pass(struct timeval *elapsed)
{
    filelist_s *iter, *last;

    last = 0;
    for (iter = filehead; iter; iter = iter->next) {
	if (iter->delay.tv_usec > elapsed->tv_usec)
	    iter->delay.tv_usec -= elapsed->tv_usec;
	else {
	    iter->delay.tv_usec += 1000000 - elapsed->tv_usec;
	    iter->delay.tv_sec--;
	}
	if (iter->delay.tv_sec <= elapsed->tv_sec) {
	    iter->delay.tv_sec = 0;
	    iter->delay.tv_usec = 0;
	} else
	    iter->delay.tv_sec -= elapsed->tv_sec;
	if (iter->delay.tv_sec == 0 && iter->delay.tv_usec == 0) {
	    if ((iter->state & FILESTATE_DELETE) != 0) {
		/* the file's to be deleted. */
		unlink(iter->filename);
		free((void*) iter->filename);
		free((void*) iter->user);
		if (last)
		    last->next = iter->next;
		else
		    filehead = filehead->next;
		free((void*) iter);
	    } else if ((iter->state & FILESTATE_SEND) == 0) {
		/* we're to send the article to the server. */
		send_article(iter);
	    }
	}
	last = iter;
    }
}

static void get_delay(struct timeval *delay)
{
    filelist_s *iter;

    /* rescan the directory every minute, at least. */
    delay->tv_sec = 60;
    delay->tv_usec = 0;
    for (iter = filehead; iter; iter = iter->next) {
	if (((iter->state & FILESTATE_SEND) == 0) &&
	  iter->delay.tv_sec <= delay->tv_sec) {
	    if (iter->delay.tv_sec == delay->tv_sec &&
	      iter->delay.tv_usec > delay->tv_usec)
		continue;
	    delay->tv_sec = iter->delay.tv_sec;
	    delay->tv_usec = iter->delay.tv_usec;
	}
    }
}

static void scan_directory(char *dname)
{
    DIR *d;
    struct dirent *ent;
    filelist_s *iter, *del;
    char *user;
    int delay;
    int seen;
    char *realfname;
    int realfnamelen;
    int baselen;
    struct stat st;
    time_t now;

    /* mark all entries in the list as unseen */
    for (iter = filehead; iter; iter = iter->next)
	iter->state &= ~FILESTATE_FLAG;
    d = opendir(dname);
    if (!d) {
	syslog(LOG_WARNING, "%s:%s Couldn't open holding directory %s: %m",
	  hostname, port, d);
	return;
    }
    baselen = strlen(dname);
    realfnamelen = baselen+15;
    realfname = (char*) malloc(realfnamelen);
    while (ent = readdir(d)) {
	if (parse_filename(ent->d_name, &user, &delay) != 0) {
	    if (user)
		free((void*) user);
	    continue;
	}
	if (strlen(ent->d_name) > realfnamelen-baselen-2) {
	    realfnamelen = baselen + strlen(ent->d_name)+2;
	    realfname = (char*) realloc((void*)realfname,
	      realfnamelen);
	}
	sprintf(realfname, "%s/%s", dname, ent->d_name);
	seen = 0;
	for (iter = filehead; !seen && iter; iter = iter->next) {
	    if (!(iter->state & FILESTATE_FLAG) &&
	      !strcmp(realfname, iter->filename)) {
		seen = 1;
		iter->state |= FILESTATE_FLAG;
	    }
	}
	if (!seen && (stat(realfname, &st) == 0)) {
	    /* new holding file */
	    iter = (filelist_s*) malloc(sizeof(filelist_s));
	    iter->user = user;
	    now = time(0);
	    iter->delay.tv_sec = delay - (now - st.st_mtime);
	    iter->delay.tv_usec = 0;
	    if (iter->delay.tv_sec < 0)
		iter->delay.tv_sec = 0;
	    iter->filename = strdup(realfname);
	    iter->next = filehead;
	    iter->state = FILESTATE_FLAG;
	    filehead = iter;
	} else
	    free((void*) user);
    }
    closedir(d);

    /* clean out entries from the list that have been deleted externally. */
    iter = filehead;
    for (del = iter; del; del = del->next) {
	if (! (del->state & FILESTATE_FLAG)) {
	    /* this file didn't exist this pass. */
	    if (del == iter) {
		del = del->next;
		free((void*) iter->filename);
		free((void*) iter->user);
		free((void*) iter);
	    } else {
		iter->next = del->next;
		free((void*) del->filename);
		free((void*) del->user);
		free((void*) del);
		del = iter->next;
	    }
	}
	iter = del;
    }
}

static char *get_ihave_string(char *text, filelist_s *art)
{
    char *endhdr;
    char *msgid;
    char *ret;

    endhdr = strstr(text, "\r\n\r\n");
    if (!endhdr) {
	syslog(LOG_WARNING, "%s:%s File %s had no end of headers.",
	  hostname, port, art->filename);
	free((void*) text);
	return;
    }
    for (msgid = text; msgid < endhdr; msgid = strchr(msgid, '\n')) {
	if (*msgid == '\n')
	    msgid++;
#define MSGIDSTRING "Message-ID: "
	if (!strncmp(msgid, MSGIDSTRING, strlen(MSGIDSTRING)))
	    break;
    }
    if (!msgid || msgid >= endhdr) {
	syslog(LOG_WARNING, "%s:%s File %s had no Message-ID header.",
	  hostname, port, art->filename);
	free((void*) text);
	return;
    }
    msgid += strlen(MSGIDSTRING);
    endhdr = strchr(msgid, '\n');
#define IHAVESTR "IHAVE "
    ret = (char*) malloc(strlen(IHAVESTR)+endhdr-msgid+1);
    strcpy(ret, IHAVESTR);
    strncpy(ret+strlen(IHAVESTR), msgid, endhdr-msgid+1);
    return(ret);
}

static void log_err(wrbuf_s *wrbuf, char *respline)
{
    FILE *errout;
    char *errfname;
    filelist_s *art;

    art = (filelist_s*) wrbuf->data;
    syslog(LOG_NOTICE, "%s:%s Couldnt send file %s: %s",
      hostname, port, art->filename, respline);
#define ERRSUFF ".err"
    errfname = (char*) malloc(strlen(art->filename)+strlen(ERRSUFF)+1);
    sprintf(errfname, "%s%s", art->filename, ERRSUFF);
    errout = fopen(errfname, "w");
    if (!errout) {
	syslog(LOG_WARNING, "%s:%s Couldnt create error output file %s: %m",
	  hostname, port, art->filename);
	free((void*) errfname);
	return;
    }
    fprintf(errout, "%s\n", respline);
    fprintf(errout, "%s\n", wrbuf->writebuf);
    fclose(errout);
    unlink(art->filename);
    free((void*) art->filename);
    art->filename = errfname;
}

static void ihave_finish(wrbuf_s *wrbuf, int code, char *respline)
{
    int num;
    wrbuf_s *del;
    filelist_s *art;

    if (respline[0] != '3') {
	/* If we couldn't send the article, tell someone... */
	log_err(wrbuf->next, respline);
	art = (filelist_s*) wrbuf->next->data;
	switch (fork()) {
	  case 0:
	    for (num = 2; num < 256; num++)
		close(num);
	    execl(BINDIR "/post_error.pl", "post_error.pl",
	      art->user, art->filename, 0);
	    exit(1);
	  case -1:
	    break;
	}
	/* ...and remove the next bit (the article) from the queue. */
	del = wrbuf->next;
	wrbuf->next = del->next;
	free((void*) wrbuf->writebuf);
	free((void*) del);
    } else
	state = STATE_ARTICLE;
}

static void ihave_abort(wrbuf_s *wrbuf)
{
    /* server dropped us in the middle of an ihave */
}

static void article_finish(wrbuf_s *wrbuf, int code, char *respline)
{
    char delfname[256];
    filelist_s *art;
    int num;

    num = 0;
    art = (filelist_s*) wrbuf->data;
    while (num < 500) {
	sprintf(delfname, "%s/%s.%d.%d", GIZMO_DEL_DIR, art->user,
	  60*60*24*3, num);
	if (link(art->filename, delfname) == 0) {
	    unlink(art->filename);
	    free((void*) art->filename);
	    art->filename = strdup(delfname);
	    art->state = FILESTATE_DELETE;
	    art->delay.tv_sec = 60*60*24*3;
	    art->delay.tv_usec = 0;
	    break;
	} else if (errno != EEXIST) {
	    /* couldn't link the file, for some reason.. */
	    syslog(LOG_WARNING, "%s:%s couldn't link %s to %s: %m",
	      hostname, port, art->filename, delfname);
	    break;
	}
    }
    if (wrbuf->next)
	state = STATE_STARTICLE;
    else
	state = STATE_NORM;
    if (respline[0] != '2') {
	/* If we couldn't send the article, tell someone. */
	log_err(wrbuf, respline);
	switch (fork()) {
	  case 0:
	    for (num = 0; num < 256; num++)
		close(num);
	    execl(BINDIR "/post_error.pl", "post_error.pl",
	      art->user, art->filename, 0);
	    exit(1);
	  case -1:
	    break;
	}
    }
}

static void article_abort(wrbuf_s *wrbuf)
{
    /* server dropped us in the middle of the article..  we should put the
     * wrbuf for the ihave back in. */
    /* XXX put the wrbuf for the ihave back in */
    wrbuf_s *ihave_buf;

    ihave_buf = (wrbuf_s*) malloc(sizeof(wrbuf_s));
    bzero(ihave_buf, sizeof(wrbuf_s));
    ihave_buf->writebuf = get_ihave_string(wrbuf->writebuf, wrbuf->data);
    ihave_buf->todo = strlen(ihave_buf->writebuf);
    ihave_buf->next = wrbuf;
    write_head = ihave_buf;
    state = STATE_STARTICLE;
}

static void send_article(filelist_s *art)
{
    char *endhdr;
    char *text;
    char *msgid;
    struct stat st;
    int fd, rgot, got, todo;
    wrbuf_s *wrbuf, *bufiter;
    int count;

    fd = open(art->filename, O_RDONLY, 0644);
    if (fd < 0) {
	/* XXX should delete entry */
	return;
    }
    if (fstat(fd, &st) < 0) {
	syslog(LOG_NOTICE, "%s:%s Couldn't stat article file %s for sending:%m",
	  hostname, port, art->filename);
	close(fd);
    }
#define TERMSTR ".\r\n"
    text = (char*) malloc(st.st_size+strlen(TERMSTR));
    todo = st.st_size;
    rgot = 0;
    while (todo) {
	got = read(fd, text+rgot, st.st_size);
	if (got > 0) {
	    todo -= got;
	    rgot += got;
	}
	else if (got == 0) {
	    syslog(LOG_WARNING, "%s:%s File %s shrunk while reading?",
	      hostname, port, art->filename);
	    break;
	} else if (got < 0 && errno != EINTR) {
	    syslog(LOG_WARNING, "%s:%s read file %s failed: %m",
	      hostname, port, art->filename);
	    break;
	}
    }
    close(fd);
    if (todo) {
	/* didn't read whole article */
	free((void*) text);
	return;
    }
    strncpy(text+rgot, TERMSTR, strlen(TERMSTR));
    wrbuf = (wrbuf_s*) malloc(sizeof(wrbuf_s));
    bzero(wrbuf, sizeof(wrbuf_s));
    wrbuf->writebuf = get_ihave_string(text, art);
    wrbuf->todo = strlen(wrbuf->writebuf);
    wrbuf->data = (void*) art;
    wrbuf->finish_func = ihave_finish;
    if (!write_head)
	write_head = wrbuf;
    else {
	count = 0;
	for (bufiter = write_head; bufiter->next; bufiter = bufiter->next)
	    count++;
	bufiter->next = wrbuf;
	if (count > 5)
	    /* we've got more than two articles in the queue? */
	    syslog(LOG_NOTICE, "%s:%s: %d write buffers in queue.",
	      hostname, port, count);
    }
    wrbuf->next = (wrbuf_s*) malloc(sizeof(wrbuf_s));
    wrbuf = wrbuf->next;
    bzero(wrbuf, sizeof(wrbuf_s));
    wrbuf->writebuf = text;
    wrbuf->data = (void*) art;
    wrbuf->todo = strlen(wrbuf->writebuf);
    wrbuf->finish_func = article_finish;
    wrbuf->abort_func = article_abort;
    art->state |= FILESTATE_SEND;
}

int parse_filename(char *fname, char **user, int *delay)
{
    char *basename, *dot;
    char buf[255];
    int index;

    *user = 0;
    basename = strrchr(fname, '/');
    if (basename)
        basename++;
    else
        basename = fname;
    strcpy(buf, basename);
    basename = buf;

    /* Do this backwards since username could be anything */
    if (!(dot = strrchr(basename, '.')))
        return (1);
    /* Make sure .<unique> is all number */
    *dot++ = 0;
    if (!*dot)
        return(1);
    for (index = 0; dot[index]; index++)
        if ((dot[index]<'0') || (dot[index]>'9'))
            return(1);

    /* Make sure .<delay> is all number */
    if (!(dot = strrchr(basename, '.')))
	return(1);
    *dot++ = 0;
    if (!*dot)
        return(1);
    for (index = 0; dot[index]; index++)
        if ((dot[index]<'0') || (dot[index]>'9'))
            return(1);
    *delay = atoi(dot);

    /* User gets the rest */
    if (!*basename)
        return(1);
    
    *user = strdup(basename);
    return(0);
}

static void config(void)
{
    struct hostent *hent;
    struct servent *sent;
    struct sockaddr_in sin;
    char *portiter;
    int addriter, stat;

    bzero(&sin, sizeof(sin));
    sin.sin_family = AF_INET;
    sin.sin_addr.s_addr = inet_addr(hostname);
    hent = 0;
    if (sin.sin_addr.s_addr == (u_int32_t) -1) {
	hent = gethostbyname(hostname);
	if (!hent) {
	    fprintf(stderr, "gethostbyname(%s) failed.\n", hostname);
	    exit(1);
	}
	bcopy(hent->h_addr_list[0], &sin.sin_addr.s_addr, sizeof(sin.sin_addr.s_addr));
    }
    for (portiter = port; (*portiter >= '0') && (*portiter <= '9'); portiter++)
	;
    if (!*portiter)
	/* completely numeric */
	sin.sin_port = htons(atoi(port));
    else {
	sent = getservbyname(port, "tcp");
	if (!sent) {
	    fprintf(stderr, "getservbyname(%s/tcp) failed.\n", port);
	    exit(1);
	}
	sin.sin_port = sent->s_port;
    }
    sin.sin_len = sizeof(sin);

    /* got the address, now attempt to connect to it. */
    newssock = socket(AF_INET, SOCK_STREAM, 0);
    if (newssock < 0) {
	fprintf(stderr, "Could not open socket: %s\n", strerror(errno));
	exit(1);
    }
    addriter = 1;
    while ((stat = connect(newssock, (struct sockaddr*) &sin, sizeof(sin))) < 0) {
	if (!hent || !hent->h_addr_list[addriter])
	    break;
	bcopy(hent->h_addr_list[addriter], &sin.sin_addr.s_addr, sizeof(sin.sin_addr.s_addr));
	addriter++;
    }
    if (stat < 0) {
	fprintf(stderr, "Couldn't connect to %s:%s: %s\n", hostname, port,
	  strerror(errno));
	close(newssock);
	exit(1);
    }
    state = STATE_STARTUP;
}

static void finquit(wrbuf_s *buf, int respcode, char *resp)
{
    state = STATE_SHUTDOWN;
    /* we're trying to quit. */
}

static void abortquit(wrbuf_s *buf)
{
    /* server dropped connection?  GREAT! */

    free((void*) buf->writebuf);
    write_head = write_head->next;
    free((void*) buf);
}

static void writequit()
{
    wrbuf_s *quitbuf;

    quitbuf = (wrbuf_s*) malloc(sizeof(wrbuf_s));
    bzero(quitbuf, sizeof(wrbuf_s));
    quitbuf->writebuf = strdup("QUIT\r\n");
    quitbuf->todo = strlen(quitbuf->writebuf);
    quitbuf->curpos = 0;
    quitbuf->finish_func = finquit;
    quitbuf->abort_func = abortquit;
}

static void hupself(void)
{
    /* temporarily ignore config changes. */
}

static void chldself(void)
{
    int status;

    /* when we get a SIGCHLD, just wait for the child process */
    wait(&status);
}
