#include "config.h"

#include <sys/types.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <netinet/in.h>

#include <dirent.h>
#include <errno.h>
#include <signal.h>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <syslog.h>
#include <time.h>
#include <unistd.h>

#include "libgucs.h"
#include "gucs_err.h"
#include "catchup_cfg.h"

#include "paths.h"

/* sorted (by action time) list of files for us to take action against. */
typedef struct _filelist_s
{
	char *filename;
	char *user;
	time_t exp;
	int seen;
	struct _filelist_s *next;
} filelist_s;

typedef struct _connection_s
{
	FILE *rd;
	FILE *wr;
} connection_s;

/* config */
static void catchup_log(int level, char *mesg, ...);
static void do_args(int argc, char *argv[]);
static void do_config(void);
static void check_config(void);
static int check_dir(char *, char *);
static int check_server(void);
static int check_pid(char *);

/* detach from the terminal */
static void detach(void);

/* signal handlers */
static void handle_hup(int);

static int do_rlimit(catchup_config_s *);
static void do_del(char *, time_t);
static time_t do_hold(filelist_s **, time_t);

static void scan_directory(filelist_s **, char *);
static int parse_filename(char*, int*, int*);

static int connect_server(catchup_config_s *, connection_s *);
static void disconnect_server(connection_s *);

static int server_rd_init(void);
static char *server_read(connection_s *);

static int send_article(filelist_s *, connection_s *);
static void del_article(filelist_s *);
static void err_article(filelist_s *, char *, FILE *);

/* global variables */
static char *config_file;
static catchup_config_s config;
static connection_s serv_conn;
static int connected;
static int detached;
static int reload_needed;

/* utility macros */
#define CLOG(log_args) \
	do \
	{ \
		if (detached) \
			syslog log_args ; \
		else \
			catchup_log log_args ; \
	} while (0)

int
main(int argc, char *argv[])
{
	struct sigaction pipe_handler;
	struct sigaction hup_handler;
	filelist_s *holdlist, *iter;
	time_t now, wait;

	detached = 0;
	do_args(argc, argv);
	do_config();

	if (server_rd_init() < 0)
	{
		CLOG((LOG_ERR, "could not initialize server read buffer: %s",
		      strerror(errno)));
		exit(1);
	}
	check_config();

	pipe_handler.sa_handler = SIG_IGN;
	sigemptyset(&pipe_handler.sa_mask);
	pipe_handler.sa_flags = 0;

	if (sigaction(SIGPIPE, &pipe_handler, 0) < 0)
	{
		CLOG((LOG_ERR, "couldn't ignore SIGPIPE: %s", strerror(errno)));
		exit(1);
	}

	hup_handler.sa_handler = handle_hup;
	sigemptyset(&hup_handler.sa_mask);
	hup_handler.sa_flags = 0;

	if (sigaction(SIGHUP, &hup_handler, 0) < 0)
	{
		CLOG((LOG_ERR, "couldn't install SIGHUP handler: %s",
		      strerror(errno)));
		exit(1);
	}

	if (do_rlimit(&config) < 0)
	{
		CLOG((LOG_ERR, "couldn't unlimit CPU usage: %s",
		      strerror(errno)));
		exit(1);
	}

	/* now that configuration has been loaded, detach from the terminal,
	 * write the pid, and start running proper. */
	if (config.detach)
		detach();

	holdlist = (filelist_s *) 0;
	scan_directory(&holdlist, config.hold.dname);

	connected = 0;
	while (1)
	{
		now = time(0);
		do_del(config.del.dname, now);
		wait = do_hold(&holdlist, now);
		sleep(wait);

		scan_directory(&holdlist, config.hold.dname);
	}

	exit(0);
}

/* syslog-like interface to fprintf. */
static void catchup_log(int level, char *mesg, ...)
{
	va_list ap;

	va_start(ap, mesg);
	vfprintf(stderr, mesg, ap);
	va_end(ap);
	fprintf(stderr, "\n");
}

/* parse command-line arguments */
static void do_args(int argc, char *argv[])
{
	int opt;

	config_file = CATCHUP_CONF_FILE;
	while ((opt = getopt(argc, argv, "f:h")) >= 0)
	{
		switch(opt)
		{
		  case 'f':
			config_file = optarg;
			break;
		  case 'h':
			break;
		}
	}
}

static void do_config(void)
{
	GUCS_cf *cf;

	reload_needed = 0;
	catchup_config_defaults(&config);
	cf = gucs_fopen(config_file);
	if (cf == (GUCS_cf *) 0)
	{
		CLOG((LOG_ERR, "could not open config file %s: %s",
		      CATCHUP_CONF_FILE, strerror(errno)));
		exit(1);
	}
	gucs_parse_catchup(cf, (void *) &config);
	gucs_fclose(cf);
}

static void handle_hup(int signum)
{
	reload_needed = 1;
}

/* check that we can read/write/modify a directory, on startup. */
static int check_dir(char *dname, char *desc)
{
	DIR *d;

	if (chdir(dname) != 0)
	{
		CLOG((LOG_ERR, "could not change to %s directory %s: %s",
		      desc, dname, strerror(errno)));
		return (1);
	} else if (access(dname, W_OK) != 0)
	{
		CLOG((LOG_ERR, "can't write to %s directory %s: %s",
		      desc, dname, strerror(errno)));
		return (1);
	} else
	{
		d = opendir(dname);
		if (d == (DIR *) 0)
		{
			CLOG((LOG_ERR, "couldn't opendir %s dir %s: %s",
			      desc, dname, strerror(errno)));
			return (1);
		} else
			closedir(d);
	}
	return (0);
}

/* check that the server from the config file will talk to us. */
static int check_server(void)
{
	connection_s conn;

	if (connect_server(&config, &conn) < 0)
	{
		CLOG((LOG_ERR, "could not connect to server: %s",
		      strerror(errno)));
		return(1);
	}
	disconnect_server(&conn);
	return (0);
}

/* check for active catchup process */
static int check_pid(char *pidfile)
{
	FILE *f;
	pid_t pid;
	char buf[256];

	f = fopen(pidfile, "r");
	if (f == (FILE *) 0)
		return (0);
	CLOG((LOG_WARNING, "catchup pid file (%s) found", pidfile));
	if (fgets(buf, sizeof(buf), f) != (char *) 0)
	{
		pid = atoi(buf);
		if (pid > 0)
		{
			if (kill(pid, 0) == 0)
			{
				CLOG((LOG_ERR, "pid %d still exists", pid));
				return (1);
			}
		}
	}
	CLOG((LOG_WARNING, "seems to be stale."));
	fclose(f);
	return (0);
}

/* check that we can operate, on startup. */
static void check_config(void)
{
	int err;

	err = 0;
	err += check_dir(config.hold.dname, "holding");
	err += check_dir(config.del.dname, "del");
	if (config.detach)
		err += check_pid(config.pidfile);
	err += check_server();

	if (err != 0)
		exit(1);
}

static void detach(void)
{
	FILE *pidf;

	switch (fork())
	{
	  case 0:
		break;
	  case -1:
		fprintf(stderr, "Couldn't fork detached process: %s\n",
		        strerror(errno));
		exit(1);
	  default:
		exit(0);
	}
	setsid();
	detached = 1;
	openlog("catchup", 0, LOG_NEWS);
	/* we're now detached. */

	/* update the pid file */
	pidf = fopen(config.pidfile, "w");
	if (pidf == (FILE *) 0)
	{
		CLOG((LOG_ERR, "Couldn't open pid file %s: %s",
		      config.pidfile, strerror(errno)));
		exit(1);
	}
	fprintf(pidf, "%d\n", getpid());
	fclose(pidf);
}

static void do_del(char *dname, time_t now)
{
	DIR *d;
	struct dirent *ent;
	int ulen;
	int delay;
	struct stat st;

	if (chdir(dname) != 0)
	{
		CLOG((LOG_ERR, "couldn't chdir() to deldir %s: %s", dname,
		      strerror(errno)));
		exit(1);
	}
	d = opendir(dname);
	if (d == (DIR *) 0)
	{
		CLOG((LOG_WARNING, "Couldn't open del directory %s: %s", dname,
		      strerror(errno)));
		return;
	}

	while ((ent = readdir(d)) != (struct dirent *) 0)
	{
		if (parse_filename(ent->d_name, &ulen, &delay) != 0)
			continue;

		if (stat(ent->d_name, &st) < 0)
			continue;
		if (st.st_mtime + delay < now)
			unlink(ent->d_name);
	}
	closedir(d);
}

static time_t do_hold(filelist_s **holdlist, time_t now)
{
	filelist_s *iter;
	char *msgid;
	char *newfname;

	if (! *holdlist)
		return (30);
	if (! connected)
	{
		if ((*holdlist)->exp < now + 30)
		{
			/* attempt to connect to the server */
			connect_server(&config, &serv_conn);
			connected = 1;
		} else
			return (30);
	}

	/* we've got a live connection to the server. */
	if (chdir(config.hold.dname) != 0)
	{
		CLOG((LOG_ERR, "can't chdir to hold directory %s: %s",
		      config.hold.dname, strerror(errno)));
		exit(1);
	}
	while (*holdlist != (filelist_s *) 0 && (*holdlist)->exp < now)
	{
		iter = *holdlist;
		*holdlist = iter->next;
		if (send_article(iter, &serv_conn) == 0)
			del_article(iter);
	}

	if (! *holdlist || (*holdlist)->exp - now > 30)
	{
		/* no usable files in the holding directory, or too long a
		 * wait before transmitting again. */
		if (connected)
		{
			/* don't need to keep the connection open */
			disconnect_server(&serv_conn);
			connected = 0;
		}
		return (30);
	}
}

static void del_article(filelist_s *finfo)
{
	char delbuf[256];
	int iter;
	struct stat st;

	iter = 0;
	do
	{
		sprintf(delbuf, "%s/%s.%d.%d", config.del.dname, finfo->user,
		        86400*3, iter);
		iter++;
	} while (stat(delbuf, &st) == 0);
	rename(finfo->filename, delbuf);

	free((void *) finfo->filename);
	free((void *) finfo->user);
	free((void *) finfo);
}

static char *get_msgid(filelist_s *finfo, FILE *file)
{
	static char buf[256];
	char *msgid;

	/* find the ihave string */
	buf[sizeof(buf)-1] = '\0';
	while (fgets(buf, sizeof(buf)-1, file) != (char *) 0)
	{
		if (strcmp(buf, "\r\n") == 0)
		{
			/* end of headers */
			CLOG((LOG_INFO, "no message id"));
			return (char *) 0;
		}
#define MSGIDSTRING "Message-ID: "
		if (strncasecmp(buf, MSGIDSTRING, strlen(MSGIDSTRING)) == 0)
		{
			/* we'll need to read the file again, for transmission.
			 */
			fseek(file, 0, SEEK_SET);

			msgid = strchr(buf, '<');
			if (msgid == (char *) 0)
				return ((char *) 0);
			while (buf[strlen(buf)-1] == '\r' ||
			  buf[strlen(buf)-1] == '\n')
			{
				buf[strlen(buf)-1] = '\0';
			}
			CLOG((LOG_INFO, "Found Message id %s", msgid));
			return (msgid);
		}
	}

	CLOG((LOG_INFO, "no message id"));
	return ((char *) 0);
}

static int send_article(filelist_s *finfo, connection_s *conn)
{
	FILE *f;
	static char buf[256];
	int got;
	char *msgid, *status;

	f = fopen(finfo->filename, "r");
	if (f == (FILE *) 0)
		return (-1);
	msgid = get_msgid(finfo, f);
	if (msgid == (char *) 0)
	{
		fclose(f);
		return (-1);
	}

	fprintf(conn->wr, "IHAVE %s\r\n", msgid);
	fflush(conn->wr);
	status = server_read(conn);
	if (status[0] != '3')
	{
		err_article(finfo, status, f);
		fclose(f);
		return (-1);
	}
	while ((got = fread(buf, 1, sizeof(buf), f)) > 0)
		fwrite(buf, 1, got, conn->wr);
	fwrite(".\r\n", 1, 3, conn->wr);
	fflush(conn->wr);

	/* get back status from the server. */
	status = server_read(conn);
	if (status[0] != '2')
	{
		fseek(f, 0, SEEK_SET);
		err_article(finfo, status, f);
		fclose(f);
		return(-1);
	}
	fclose(f);
	return (0);
}

static void err_article(filelist_s *finfo, char *err, FILE *f)
{
	FILE *out;
	pid_t child;
	int got, fd;
	int wait_status;
	char fname[256];
	char buf[256];

	sprintf(fname, "%s.err", finfo->filename);
	out = fopen(fname, "w");
	if (out == (FILE *) 0)
	{
		/* XXX: log error */
		return;
	}
	fprintf(out, "%s", err);
	while ((got = fread(buf, 1, sizeof(buf), f)) > 0)
		fwrite(buf, 1, got, out);
	fclose(out);
	unlink(finfo->filename);

	/* invoke post_error */
	CLOG((LOG_WARNING, "err to user %s", finfo->user));
	child = fork();
	switch (child)
	{
	  case 0:
		for (fd = 2; fd < 256; fd++)
			close(fd);
		execl(config.posterror, config.posterror,
		      finfo->user, fname, 0);
		exit(1);
	  case -1:
		break;
	}
	/* wait for the child process to finish. */
	waitpid(child, &wait_status, 0);
}

static int do_rlimit(catchup_config_s *config)
{
	struct rlimit rl;

	if (getrlimit(RLIMIT_CPU, &rl) < 0)
		return (-1);
	rl.rlim_cur = RLIM_INFINITY;
	if (setrlimit(RLIMIT_CPU, &rl) < 0)
		return (-1);
	return (0);
}

static int connect_server(catchup_config_s *config, connection_s *conn)
{
	int fd, fd2;
	char *init;

	fd = socket(AF_INET, SOCK_STREAM, 0);
	if (fd < 0)
	{
		CLOG((LOG_WARNING, "Could not open server socket: %s",
		      strerror(errno)));
		return (-1);
	}
	if (connect(fd, (struct sockaddr *) &config->server, sizeof(config->server)) < 0)
	{
		CLOG((LOG_WARNING, "Could not connect to server: %s",
		      strerror(errno)));
		close(fd);
		return (-1);
	}
	fd2 = dup(fd);
	if (fd2 < 0)
	{
		CLOG((LOG_WARNING, "Could not dup2 server socket: %s",
		      strerror(errno)));
		close(fd);
		return (-1);
	}

	conn->rd = fdopen(fd, "r");
	conn->wr = fdopen(fd2, "w");
	if (conn->rd == (FILE *) 0 || conn->wr == (FILE *) 0)
	{
		CLOG((LOG_WARNING,
		      "Could not fdopen server communication files: %s",
		      strerror(errno)));

		if (conn->rd != (FILE *) 0)
			fclose(conn->rd);
		else
			close(fd);

		if (conn->wr != (FILE *) 0)
			fclose(conn->wr);
		else
			close(fd2);
		return (-1);
	}

	/* the server connection is set up.  make sure it wants to talk to
	 * us. */
	init = server_read(conn);
	if (init[0] != '2' || init[2] != '0')
	{
		/* one of the various 'Permission Denied' errors */
		CLOG((LOG_WARNING, "Permission denied opening connection: %s",
		      init));
		disconnect_server(conn);
		exit(1);
	}
	return (0);
}

static void disconnect_server(connection_s *conn)
{
	fprintf(conn->wr, "QUIT\r\n");
	fflush(conn->wr);

	fclose(conn->rd);
	fclose(conn->wr);
}

static void filelist_insert(filelist_s **head, filelist_s *new)
{
	filelist_s *iter, *prev;

	prev = (filelist_s *) 0;
	iter = *head;
	while (iter && iter->exp <= new->exp)
	{
		prev = iter;
		iter = iter->next;
	}

	if (prev == (filelist_s *) 0)
	{
		/* insert before the first element. */
		new->next = *head;
		*head = new;
	} else
	{
		prev->next = new;
		new->next = iter;
	}
}

static void scan_directory(filelist_s **head, char *dname)
{
	DIR *d;
	struct dirent *ent;
	filelist_s *iter, *del;
	char *realfname;
	int delay;
	int ulen;
	int seen;
	int baselen;
	struct stat st;

	for (iter = *head; iter; iter = iter->next)
		iter->seen = 0;
	if (chdir(dname) != 0)
	{
		CLOG((LOG_ERR, "Couldn't chdir to holding directory %s: %s",
		      dname, strerror(errno)));
		exit(1);
	}
	d = opendir(dname);
	if (!d)
	{
		CLOG((LOG_WARNING, "Couldn't open holding directory %s: %s",
		      dname, strerror(errno)));
		return;
	}
	realfname = (char *) 0;

	while ((ent = readdir(d)) != (struct dirent *) 0)
	{
		if (parse_filename(ent->d_name, &ulen, &delay) != 0)
			continue;

		seen = 0;
		for (iter = *head; !seen && iter; iter = iter->next)
		{
			if (!iter->seen &&
			  strcmp(iter->filename, ent->d_name) == 0)
			{
				seen = 1;
				iter->seen = 1;
			}
		}

		if (! seen)
		{
			/* new holding file */
			if (stat(ent->d_name, &st) == 0)
			{
				realfname = (char *) malloc(strlen(ent->d_name) + 1);
				strcpy(realfname, ent->d_name);
				iter = (filelist_s *) malloc(sizeof(filelist_s));
				iter->user = (char *) malloc(ulen + 1);
				strncpy(iter->user, ent->d_name, ulen);
				iter->user[ulen] = '\0';
				iter->exp = st.st_mtime + delay;
				iter->filename = realfname;

				filelist_insert(head, iter);
			}
		}
	}
	closedir(d);

	/* clean out entries from the list that have been deleted externally. */
	del = iter = *head;
	while (del)
	{
		if ((strncmp(del->filename, dname, baselen) == 0) &&
		  ! del->seen)
		{
			/* this file didn't exist this pass. */
			if (del == iter)
			{
				del = del->next;
				free((void *) iter->filename);
				free((void *) iter->user);
				free((void *) iter);
			} else
			{
				iter->next = del->next;
				free((void *) del->filename);
				free((void *) del->user);
				free((void *) del);
				del = iter->next;
			}
		}
		iter = del;
		if (del)
			del = del->next;
	}
}

static int parse_filename(char *fname, int *ulen, int *delay)
{
	char *dot;
	char buf[255];
	int index;

	*ulen = 0;
	/* check for buffer overrun */
	if (strlen(fname) > sizeof(buf) - 1)
		return (1);
	strcpy(buf, fname);
	fname = buf;

	/* Do this backwards since username could be anything */
	if (!(dot = strrchr(fname, '.')))
		return (1);
	/* Make sure .<unique> is all number */
	*dot++ = 0;
	if (!*dot)
		return(1);
	for (index = 0; dot[index]; index++)
		if ((dot[index]<'0') || (dot[index]>'9'))
			return(1);

	/* Make sure .<delay> is all number */
	if (!(dot = strrchr(fname, '.')))
		return(1);
	*dot++ = 0;
	if (!*dot)
		return(1);
	for (index = 0; dot[index]; index++)
		if ((dot[index]<'0') || (dot[index]>'9'))
			return(1);
	*delay = atoi(dot);

	/* User gets the rest */
	if (!*fname)
		return(1);
    
	*ulen = strlen(buf);
	return(0);
}

static char *server_rd_buf;
static size_t server_buf_size;

static int server_rd_init(void)
{
	server_buf_size = 512;
	server_rd_buf = (char *) malloc(server_buf_size);
	if (server_rd_buf == (char *) 0)
		return (-1);
	server_rd_buf[server_buf_size - 1] = '\0';
	return (0);
}

static char *server_read(connection_s *conn)
{
	char *rval;

	rval = fgets(server_rd_buf, server_buf_size - 1, conn->rd);
	if (strchr(rval, '\n') == (char *) 0)
	{
		CLOG((LOG_WARNING, "no newline in response from server."));
		return ((char *) 0);
	}
	return (rval);
}
