/* limitfile.c - load the spam.limit config file */

/* revision history
 * 2/19/98 -
 *     Complete rewrite
 * 3/2/98 - Aidan Cully
 *     config file works, add database functionality.
 */

#include <string.h>

#include "limitfile.h"
#include "misc.h"
#include "dep.h"
#include "proto.h"

#define kLIMdeltmdata 1
#define kLIMdelstatdata 2

typedef struct _LIMdbdata {
    char *name;
    int action;
    char **tmdata;
    char *statdata;
    char *newtmdata;
    char *newstatdata;
    LIMfinishfunc finish;
    LIMsyncfunc sync;
} LIMdbdata;

typedef struct _LIMblockdata {
    char *name;
    /* if this exists, we're implicitely blocking the message */
} LIMblockdata;

struct _LIMpriv;

typedef struct _LIMchdata {
    char *name;
    struct _LIMpriv *data;
} LIMchdata;

typedef struct _LIMpriv {
    LIMdbdata **dbdata;
    LIMblockdata ***blockdata;
    LIMchdata **children;
    FILT *filt;
    int delfilt;
} LIMpriv;

/* turn a string into a bunch of arguments delimited by ':'s */
LIMconfpart *LIMargifyfile(char*, FILE*);
LIMconfpart *LIMargifypart(char*, LIMconfpart*, int*);
void LIMfreeconfpart(LIMconfpart*);

static int LIMparselim(char*, LIMconfpart*);
static char *LIMparselimname[] = { "limit", 0 };
static int LIMparsegroup(char*, LIMconfpart*);
static char *LIMparsegroupname[] = { "group", 0 };
static int LIMparsealias(char*, LIMconfpart*);
static char *LIMparsealiasname[] = { "alias", 0 };

static void LIMparsetest(depstruc*, void *);

LIMparse LIMparselist[] =
 { { LIMparselimname, LIMparselim },
   { LIMparsegroupname, LIMparsegroup },
   { LIMparsealiasname, LIMparsealias } };

static char *LIMopandname[] = { "and", "&", "&&", 0 };
static char *LIMoporname[] = { "or", "|", "||", 0 };
static char *LIMopnotname[] = { "not", "!", 0 };

LIMop LIMoplist[] =
 { { LIMopandname, kLIMopand },
   { LIMoporname, kLIMopor },
   { LIMopnotname, kLIMopnot } };

static void LIMuserdep(FILT*, LIMconfpart*, int*);
static char *LIMtestusername[] = { "user", 0 };
static void LIMtestuser(depstruc*, void*);
static void LIMgroupdep(FILT*, LIMconfpart*, int*);
static char *LIMtestgroupname[] = { "group", 0 };
static void LIMtestgroup(depstruc*, void*);
static void LIMdestdep(FILT*, LIMconfpart*, int*);
static char *LIMtestdestname[] = { "dest", 0 };
static void LIMtestdest(depstruc*, void*);

LIMdep LIMdeplist[] =
 { { LIMtestusername, LIMuserdep },
   { LIMtestgroupname, LIMgroupdep },
   { LIMtestdestname, LIMdestdep } };

typedef struct _LIMgroupdecl {
    char op;
    int isuser;
    int not;
    union {
        char *user;
        struct _LIMgroupdecl *group;
    } data;
    struct _LIMgroupdecl *next;
} LIMgroupdecl;

typedef struct _LIMgroup {
    char *name;
    char *domain;
    LIMgroupdecl *users;
} LIMgroup;

static LIMgroup **groups = 0, **aliases = 0;
static LIMgroupdecl *LIMparsegroupdecl(LIMconfpart*, char**, int*);
static LIMgroupdecl *LIMcopygroupdecl(LIMgroupdecl*, char*);
static void LIMfreegroup(LIMgroupdecl*);
static void LIMdelgroup(char*);
static LIMgroup *LIMfindgroup(char*);
static int LIMuseringroup(char*, LIMgroup*);
static int LIMuseringroupname(char*, char*);

static LIMlimdecl **limits = 0;

static void LIMaliasdep(depstruc*, void*);
static void LIMrunalias(depstruc*, void*);

LIMpriv *
LIMfindchildlim(lim, name)
    LIMpriv *lim;
    char *name;
{
    int i;

    if (!lim->children)
        return(0);
    for (i = 0; lim->children[i]; i++)
        if (!strcmp(lim->children[i]->name, name))
            return(lim->children[i]->data);
    return(0);
}

LIMpriv *
LIMaddchildlim(filt, lim, name)
    FILT *filt;
    LIMpriv *lim;
    char *name;
{
    int i;

    if (!lim->children) {
        lim->children = (LIMchdata**)malloc(2*sizeof(LIMchdata*));
        lim->children[1] = 0;
        i = 0;
    } else {
        for (i = 0; lim->children[i]; i++)
            if (!strcmp(lim->children[i]->name, name))
                return(lim->children[i]->data);
        lim->children = (LIMchdata**)realloc((void*)lim->children,
            (i+2)*sizeof(LIMchdata*));
        lim->children[i+1] = 0;
    }
    lim->children[i] = (LIMchdata*)malloc(sizeof(LIMchdata));
    lim->children[i]->data = (LIMpriv*)malloc(sizeof(LIMpriv));
    lim->children[i]->data->dbdata = 0;
    lim->children[i]->data->blockdata = lim->blockdata;
    lim->children[i]->data->children = 0;
    lim->children[i]->data->filt = filt;
    lim->children[i]->data->delfilt = 0;
    lim->children[i]->name = (char*)malloc(strlen(name)+1);
    strcpy(lim->children[i]->name, name);
    return(lim->children[i]->data);
}

LIMdbdata *
LIMfinddbdata(lim, name)
    LIMpriv *lim;
    char *name;
{
    int i;

    if (!lim->dbdata)
        return(0);
    for (i = 0; lim->dbdata[i]; i++)
        if (!strcmp(lim->dbdata[i]->name, name))
            return(lim->dbdata[i]);
    return(0);
}

LIMdbdata *
LIMadddbdata(lim, name)
    LIMpriv *lim;
    char *name;
{
    int i;

    if (!lim->dbdata) {
        lim->dbdata = (LIMdbdata**)malloc(2*sizeof(LIMdbdata*));
        i = 0;
    } else {
        for (i = 0; lim->dbdata[i]; i++)
            if (!strcmp(lim->dbdata[i]->name, name))
                return(lim->dbdata[i]);
        lim->dbdata = (LIMdbdata**)realloc(lim->dbdata,
            (i+2)*sizeof(LIMdbdata*));
    }
    lim->dbdata[i+1] = 0;
    lim->dbdata[i] = (LIMdbdata*)malloc(sizeof(LIMdbdata));
    lim->dbdata[i]->action = 0;
    lim->dbdata[i]->finish = 0;
    lim->dbdata[i]->sync = 0;
    lim->dbdata[i]->tmdata = 0;
    lim->dbdata[i]->newtmdata = 0;
    lim->dbdata[i]->statdata = 0;
    lim->dbdata[i]->newstatdata = 0;
    lim->dbdata[i]->name = (char*)malloc(strlen(name)+1);
    strcpy(lim->dbdata[i]->name, name);
    return(lim->dbdata[i]);
}

LIM *
LIMopendb(filt, name, parent)
    FILT *filt;
    char *name;
    LIM *parent;
{
    return((LIM*)LIMaddchildlim(filt, (LIMpriv*)parent, name));
}

static char *
LIMgetchildstr(src, name)
    char *src, *name;
{
    char tok[80], *iter, *lparen;
    char *ret;

    if (!src)
        return(0);
    iter = src;
    while (iter = PARSEgettoken(iter, tok)) {
        if ((*tok == '%') && !strcmp(tok+1, name)) {
            /* that's my man. */
            lparen = strchr(iter, '(');
            iter = STRfindrparen(lparen+1);
            ret = (char*)malloc(iter-lparen);
            strncpy(ret, lparen+1, iter-lparen-1);
            ret[iter-lparen-1] = 0;
            return(ret);
        } else if (*tok == '(') {
            /* don't decend into bad hierarchies. */
            iter = STRfindrparen(iter);
            if (iter)
                iter++;
        }
    }
}

static void
LIMparsedb(filt, dest, tmdata, statdata)
    FILT *filt;
    LIMpriv *dest;
    char **tmdata;
    char *statdata;
{
    int i, j, k;
    char tok[80], *iter, *last, *lparen, save;
    char **chtmdata, *chstdata;
    LIMdbdata *data;
    LIMpriv *child;

    chtmdata = (char**)malloc((numperiods()+1)*sizeof(char*));
    chtmdata[numperiods()] = 0;
    i = 0;
    if (statdata) {
        iter = last = statdata;
        while (iter = PARSEgettoken(iter, tok)) {
            if (*tok == '%') {
                /* nested limit space */
                lparen = strchr(iter, '(');
                iter = STRfindrparen(lparen+1);
                save = *iter;
                *iter = 0;
                chstdata = (char*)malloc(strlen(lparen));
                strcpy(chstdata, lparen+1);
                for (j = i; j < numperiods(); j++)
                    chtmdata[j] = LIMgetchildstr(tmdata[j], tok+1);
                child = LIMaddchildlim(filt, dest, tok+1);
                LIMparsedb(filt, child, chtmdata, chstdata);
                for (j = i; j < numperiods(); j++)
                    if (chtmdata[j])
                        free((void*)chtmdata[j]);
                free((void*)chstdata);
                iter++; /* skip past closing paren */
                continue;
            }
            /* ordinary limit space */
            data = LIMadddbdata(dest, tok);
            if (*iter == '(') {
                lparen = iter;
                iter = STRfindrparen(lparen+1);
                data->statdata = (char*)malloc(iter-lparen);
                strncpy(data->statdata, lparen+1, iter-lparen-1);
                data->statdata[iter-lparen-1] = 0;
                iter++; /* skip past closing paren */
            } else {
                data->statdata = (char*)malloc(1);
                data->statdata[0] = 0;
            }
            last = iter;
        }
    }
    for (i = 0; i < numperiods(); i++) {
        if (tmdata[i]) {
            iter = last = tmdata[i];
            while (iter = PARSEgettoken(iter, tok)) {
                if (*tok == '%') {
                    lparen = strchr(iter, '(');
                    iter = STRfindrparen(lparen+1);
                    if (child = LIMfindchildlim(dest, tok+1)) {
                        iter++;
                        continue;
                    }
                    /* nested limit space */
                    save = *iter;
                    *iter = 0;
                    chstdata = (char*)malloc(strlen(lparen));
                    strcpy(chstdata, lparen+1);
                    for (j = i; j < numperiods(); j++)
                        chtmdata[j] = LIMgetchildstr(tmdata[j], tok+1);
                    child = LIMaddchildlim(filt, dest, tok+1);
                    LIMparsedb(filt, child, chtmdata, 0);
                    for (j = i; j < numperiods(); j++)
                        if (chtmdata[j])
                            free((void*)chtmdata[j]);
                    iter++; /* skip past closing paren */
                    continue;
                }
                /* ordinary limit space */
                data = LIMadddbdata(dest, tok);
                if (!data->tmdata) {
                    data->tmdata = (char**)malloc((numperiods()+1)*sizeof(char*));
                    for (k = 0; k <= numperiods(); k++)
                        data->tmdata[k] = 0;
                }
                if (*iter == '(') {
                    lparen = iter;
                    iter = STRfindrparen(lparen+1);
                    data->tmdata[i] = (char*)malloc(iter-lparen);
                    strncpy(data->tmdata[i], lparen+1, iter-lparen-1);
                    data->tmdata[i][iter-lparen-1] = 0;
                    iter++; /* skip past closing paren */
                } else {
                    data->tmdata[i] = (char*)malloc(1);
                    data->tmdata[i][0] = 0;
                }
                last = iter;
            }
        }
    }
    free((void*)chtmdata);
}

int
LIMlendbstr(char *name, int del, char *value, char *oldvalue)
{
    int ret;

    /* entry = name(value) */
    ret = strlen(name);
    if (value) {
        if (strlen(value))
            ret += strlen(value)+2;
        else
            ret++;
    } else if (!del && oldvalue) {
        if (strlen(oldvalue))
            ret += strlen(oldvalue)+2;
        else
            ret++;
    } else
        return(0);
    return(ret);
}

void
LIMlenchildlim(LIMpriv *lim, int *lenstat, int *lentm)
{
    int chlenstat, chlentm;
    int i;

    *lenstat = 0;
    *lentm = 0;
    if (lim->children)
        for (i = 0; lim->children[i]; i++) {
            LIMlenchildlim(lim->children[i]->data, &chlenstat, &chlentm);
            /* entry = %name(childent) */
            if (chlenstat)
                *lenstat += chlenstat + 3 + strlen(lim->children[i]->name);
            if (chlentm)
                *lentm += chlentm + 3 + strlen(lim->children[i]->name);
        }
    if (!lim->dbdata)
        return;
    for (i = 0; lim->dbdata[i]; i++) {
        *lenstat += LIMlendbstr(lim->dbdata[i]->name,
            lim->dbdata[i]->action & kLIMdelstatdata,
            lim->dbdata[i]->newstatdata, lim->dbdata[i]->statdata);
        *lentm += LIMlendbstr(lim->dbdata[i]->name,
            lim->dbdata[i]->action & kLIMdeltmdata,
            lim->dbdata[i]->newtmdata,
            lim->dbdata[i]->tmdata ? lim->dbdata[i]->tmdata[0] : 0);
    }
}

void
LIMcatdbstr(char *dst, char *name, int del, char *value, char *oldvalue)
{
    /* entry = name(value) */
    if (!value && (!del && !oldvalue))
        return;
    strcat(dst, name);
    if (value) {
        if (strlen(value)) {
            strcat(dst, "(");
            strcat(dst, value);
            strcat(dst, ")");
        } else
            strcat(dst, ",");
    } else if (!del && oldvalue) {
        if (strlen(oldvalue)) {
            strcat(dst, "(");
            strcat(dst, oldvalue);
            strcat(dst, ")");
        } else
            strcat(dst, ",");
    }
}

void
LIMcatchildlim(LIMpriv *lim, char *stat, char *tmdata)
{
    int lenstat, lentm;
    int i;

    if (lim->children)
        for (i = 0; lim->children[i]; i++) {
            LIMlenchildlim(lim->children[i]->data, &lenstat, &lentm);
            if (lenstat) {
                strcat(stat, "%");
                strcat(stat, lim->children[i]->name);
                strcat(stat, "(");
            }
            if (lentm) {
                strcat(tmdata, "%");
                strcat(tmdata, lim->children[i]->name);
                strcat(tmdata, "(");
            }
            if (lenstat || lentm)
                LIMcatchildlim(lim->children[i]->data, stat, tmdata);
            if (lenstat)
                strcat(stat, ")");
            if (lentm)
                strcat(tmdata, ")");
        }
    if (!lim->dbdata)
        return;
    for (i = 0; lim->dbdata[i]; i++) {
        LIMcatdbstr(stat, lim->dbdata[i]->name,
            lim->dbdata[i]->action & kLIMdelstatdata,
            lim->dbdata[i]->newstatdata, lim->dbdata[i]->statdata);
        LIMcatdbstr(tmdata, lim->dbdata[i]->name,
            lim->dbdata[i]->action & kLIMdeltmdata,
            lim->dbdata[i]->newtmdata,
            lim->dbdata[i]->tmdata ? lim->dbdata[i]->tmdata[0] : 0);
    }
}

void
LIMwritedb(filt, lim)
    FILT *filt;
    LIMpriv *lim;
{
    int lentm, lenstat;
    char *newtm, *newstat;
    char *user;

    LIMlenchildlim(lim, &lenstat, &lentm);
    newtm = (char*)malloc(lentm+1);
    newstat = (char*)malloc(lenstat+1);
    *newtm = 0;
    *newstat = 0;
    LIMcatchildlim(lim, newstat, newtm);
    user = FILTgetuser(filt);
    setposts(user ? user : "_noauth", newtm);
    setstatic(user ? user : "_noauth", newstat);
    free((void*)newstat);
    free((void*)newtm);
}

/* test if all the "run" blocks are satisfied for this jobbie */
int
LIMshouldrun(list)
    LIMtestlist *list;
{
    int inpres;

    if (!list)
        return(1);
    inpres = list->pass;
    if (list->not)
        inpres = !inpres;
    if (list->next) {
        switch (list->op) {
          case kLIMopand:
            if (!inpres)
                return(0);
            return (LIMshouldrun(list->next));
          case kLIMopor:
            if (inpres)
                return(1);
            return (LIMshouldrun(list->next));
          default:
            return(0);
        }
    }
    return (inpres);
}

void
LIMchecklim(filt, lim)
    FILT *filt;
    LIM *lim;
{
    int i;
    LIMpriv *real;

    for (i = 0; LIMlimlist[i]; i++)
        LIMlimlist[i](filt, ((LIMpriv*)lim)->filt, lim);
    real = (LIMpriv*)lim;
    if (!real->dbdata)
        return;
}

static void
LIMbuildfiltreal(filt, decl)
    FILT *filt;
    LIMlimdecl *decl;
{
    int i;

    if (!LIMshouldrun(decl->run))
        return;
    FILTappend(filt, decl->filt);
    if (!decl->children)
        return;
    for (i = 0; decl->children[i]; i++)
        LIMbuildfiltreal(filt, decl->children[i]);
}

FILT *
LIMbuildfilt(filt, lim, decl)
    FILT *filt;
    LIM *lim;
    LIMlimdecl *decl;
{
    FILT *new;
    FILT *authdata;

    new = FILTalloc();
    authdata = FILTalloc();
    FILTappend(new, filt);
    LIMbuildfiltreal(authdata, decl);
    FILTappend(new, authdata);
    FILTaddfilt(new, "loaded", authdata);
    ((LIMpriv*)lim)->filt = new;
    ((LIMpriv*)lim)->delfilt = 1;
    return(new);
}

static void
LIMcallsync(filt, lim)
    FILT *filt;
    LIMpriv *lim;
{
    int i;

    if (lim->children)
        for (i = 0; lim->children[i]; i++)
            LIMcallsync(filt, lim->children[i]->data);
    if (!lim->dbdata)
        return;
    for (i = 0; lim->dbdata[i]; i++)
        if (lim->dbdata[i]->sync)
            lim->dbdata[i]->sync(filt, lim->filt, (LIM*)lim);
}

void
LIMtrackpost(filt)
    FILT *filt;
{
    char *user;
    char **tmdata;
    char statdata[256];
    LIMpriv lim;
    FILT *new;
    FILT *authdata;
    deplist *dep;
    int i;

    /* first, touch the "mesg" field in the deplist */
    dep = FILTgetdep(filt);
    dep_touch(find_dep(*dep, "mesg"), (void*)filt);

    /* now, build the new deplist */
    new = FILTalloc();
    authdata = FILTalloc();
    FILTappend(new, filt);
    for (i = 0; limits[i]; i++)
        LIMbuildfiltreal(authdata, limits[i]);
    FILTappend(new, authdata);
    FILTaddfilt(new, "loaded", authdata);

    /* pull the data from the database */
    user = FILTgetuser(filt);
    tmdata = (char**)malloc((numperiods()+1)*sizeof(char*));
    for (i = 0; i <= numperiods(); i++)
        tmdata[i] = 0;
    for (i = 0; i < sizeof(lim); i++)
        ((char*)&lim)[i] = 0;
    lim.blockdata = (LIMblockdata***)malloc(sizeof(LIMblockdata**));
    *lim.blockdata = 0;
    opendb();
    numposts(user ? user : "_unauth", tmdata);
    getstatic(user ? user : "_unauth", statdata, sizeof(statdata));
    LIMparsedb(new, &lim, tmdata, statdata);
    lim.filt = new;
    lim.delfilt = 1;

    /* check the limits and clean up */
    LIMchecklim(filt, (LIM*)&lim);
    LIMcallsync(filt, &lim);
    FILTadddata(filt, kFILTlimtype, sLIM, (void*)&lim, sizeof(lim));
}

int
LIMshouldsend(filt)
    FILT *filt;
{
    LIMpriv *lim;
    size_t s;

    lim = (LIMpriv*)FILTgetdata(filt, kFILTlimtype, sLIM, &s);
    if (!lim)
        return(1);
    return(*lim->blockdata == 0);
}

void
LIMcallfinish(filt, lim)
    FILT *filt;
    LIMpriv *lim;
{
    int i;

    if (lim->children)
        for (i = 0; lim->children[i]; i++)
            LIMcallfinish(filt, lim->children[i]->data);
    if (!lim->dbdata)
        return;
    for (i = 0; lim->dbdata[i]; i++)
        if (lim->dbdata[i]->finish)
            lim->dbdata[i]->finish(filt, lim->filt, (LIM*)lim);
}

void
LIMfreelim(lim)
    LIMpriv *lim;
{
    int i, k;

    if (lim->children) {
        for (i = 0; lim->children[i]; i++) {
            free((void*)lim->children[i]->name);
            LIMfreelim(lim->children[i]->data);
            free((void*)lim->children[i]->data);
            free((void*)lim->children[i]);
        }
        free((void*)lim->children);
    }
    if (!lim->dbdata)
        return;
    for (i = 0; lim->dbdata[i]; i++) {
        free((void*)lim->dbdata[i]->name);
        if (lim->dbdata[i]->statdata)
            free((void*)lim->dbdata[i]->statdata);
        if (lim->dbdata[i]->newstatdata)
            free((void*)lim->dbdata[i]->newstatdata);
        if (lim->dbdata[i]->tmdata) {
            for (k = 0; k < numperiods(); k++)
                if (lim->dbdata[i]->tmdata[k])
                    free((void*)lim->dbdata[i]->tmdata[k]);
            free((void*)lim->dbdata[i]->tmdata);
        }
        if (lim->dbdata[i]->newtmdata)
            free((void*)lim->dbdata[i]->newtmdata);
        free((void*)lim->dbdata[i]);
    }
    free((void*)lim->dbdata);
}

void
LIMfinish(filt)
    FILT *filt;
{
    LIMpriv *lim;
    size_t s;
    int i;

    lim = (LIMpriv*)FILTgetdata(filt, kFILTlimtype, sLIM, &s);
    if (!lim)
        return;
    LIMcallfinish(filt, lim);
    LIMwritedb(filt, lim);
    closedb();
    LIMfreelim(lim);
    if (*lim->blockdata) {
        for (i = 0; (*lim->blockdata)[i]; i++) {
            free((void*)(*lim->blockdata)[i]->name);
            free((void*)(*lim->blockdata)[i]);
        }
        free((void*)*lim->blockdata);
    }
    free((void*)lim->blockdata);
    FILTdeldata(filt, kFILTlimtype, sLIM);
}

/* get a line with continuation character */
static char *
LIMfgets(file)
    FILE *file;
{
    static char buf[1024];
    static char *semi = 0;
    char *foo;
    int done;

    foo = buf;
    done = 0;
    *buf = 0;
    if (semi) {
        foo = semi+1;
        semi = strchr(semi+1, ';');
        return(foo);
    }
    while (!done) {
        if (!fgets(foo, sizeof(buf)-strlen(buf), file))
            return(0);
        if (!MISCparseline(foo))
            continue;
        if (foo[strlen(foo)-1] != '\\')
            done = 1;
        else {
            foo[strlen(foo)-1] = ' ';
            foo += strlen(foo);
        }
    }
    if (semi = strchr(buf, ';'))
        *semi = 0;
    return(MISCparseline(buf));
}

typedef char *(*LIMconfcontfunc)(void*);

static char *
LIMgetconfpartreal(data, start, ret, func)
    void *data;
    char *start;
    LIMconfpart *ret;
    LIMconfcontfunc func;
{
    int inbrace, gottok, numlines;
    char *iter;
    char *tmpbuf;
    char tok[80];

    inbrace = 1;
    tmpbuf = 0;
    iter = start;
    gottok = 0;
    ret->lines = 0;
    numlines = 0;
    /* allow for braces nesting */
    while (inbrace) {
        iter = PARSEgettoken(iter, tok);
        if (!iter) {
            /* end of line.  add it to the list if we got a token. */
            if (gottok) {
                numlines++;
                if (ret->lines)
                    ret->lines = (char**)realloc((void*)ret->lines,
                                                 (numlines+1)*sizeof(char*));
                else
                    ret->lines = (char**)malloc(sizeof(char*)*2);
                ret->lines[numlines-1] = (char*)malloc(strlen(start)+1);
                strcpy(ret->lines[numlines-1], start);
                ret->lines[numlines] = 0;
            }
            tmpbuf = func(data);
            if (!tmpbuf) {
                if (ret->lines) {
                    for (inbrace = 0; inbrace < numlines; inbrace++)
                        free((void*)ret->lines[inbrace]);
                    free((void*)ret->lines);
                    ret->lines = 0;
                }
                return(0);
            }
            start = tmpbuf;
            iter = start;
            gottok = 0;
        } else {
            switch(*tok) {
              case '}':
                inbrace--;
                if (inbrace)
                    gottok = 1;
                break;
              case '{':
                inbrace++;
                /* fallthrough */
              default:
                gottok = 1;
            }
        }
    }
    if (gottok) {
        numlines++;
        while (*iter != '}')
            iter--;
        *iter++ = 0;
        if (ret->lines)
            ret->lines = (char**)realloc((void*)ret->lines,
                                         (numlines+1)*sizeof(char*));
        else
            ret->lines = (char**)malloc(sizeof(char*)*2);
        ret->lines[numlines-1] = (char*)malloc(strlen(start)+1);
        strcpy(ret->lines[numlines-1], start);
        ret->lines[numlines] = 0;
    }
    return(iter);
}

/* pull in a brace enclosed config segment, preserve newlines. */
char *
LIMgetconfpart(file, start, ret)
    FILE *file;
    char *start;
    LIMconfpart *ret;
{
    return(LIMgetconfpartreal((void*)file, start, ret,
        (LIMconfcontfunc)LIMfgets));
}

typedef struct _LIMconfpartpart {
    LIMconfpart *src;
    int *line;
} LIMconfpartpart;

char *
LIMconfpartpartcont(data)
    void *data;
{
    LIMconfpartpart *part = (LIMconfpartpart*)data;

    if (part->src->lines[*part->line+1])
        return(part->src->lines[++(*part->line)]);
    return(0);
}

/* get a loading part from a bigger one */
char *
LIMgetconfpartpart(src, line, start, ret)
    LIMconfpart *src;
    int *line;
    char *start;
    LIMconfpart *ret;
{
    LIMconfpartpart part;
    part.src = src;
    part.line = line;
    return (LIMgetconfpartreal((void*)&part, start, ret));
}

static LIMconfpart *
LIMargifyreal(data, start, func)
    void *data;
    char *start;
    LIMconfcontfunc func;
{
    char *iter, *last, tok[80];
    LIMconfpart *ret;
    int i, line, gottok, inbrace;

    i = 0;
    iter = start;
    gottok = 0;
    inbrace = 0;
    ret = (LIMconfpart*)malloc(2*sizeof(LIMconfpart));
    ret[0].lines = 0;
    ret[1].lines = 0;
    last = start;
    line = 0;
    while ((iter = PARSEgettoken(iter, tok)) || inbrace) {
        if (!iter) {
            if (gottok == 1) {
                if (ret[i].lines)
                    ret[i].lines = (char**)realloc((void*)ret[i].lines,
                        (line+2)*sizeof(char*));
                else
                    ret[i].lines = (char**)malloc(2*sizeof(char*));
                ret[i].lines[line] = (char*)malloc(strlen(start)+1);
                strcpy(ret[i].lines[line], start);
            } else if (!gottok) {
                ret[i].lines = (char**)malloc(sizeof(char*));
                ret[i].lines[0] = 0;
            }
            start = iter = func(data);
            line++;
            gottok = 3;
        }
        if (!iter)
            return(0);
        switch (*tok) {
          case ':':
	  case '=':
            if (!inbrace) {
                if (gottok == 1) {
                    /* one line big */
                    if (!ret[i].lines)
                        ret[i].lines = (char**)malloc(2*sizeof(char*));
                    else
                        ret[i].lines = (char**)realloc((void*)ret[i].lines,
                            (line+2)*sizeof(char*));
                    *last = 0;
                    ret[i].lines[line] = (char*)malloc(strlen(start)+1);
                    strcpy(ret[i].lines[line], start);
                    ret[i].lines[line+1] = 0;
                } else if (!gottok) {
                    /* no lines. */
                    ret[i].lines = (char**)malloc(2*sizeof(char*));
                    ret[i].lines[0] = (char*)malloc(1);
                    ret[i].lines[0][0] = 0;
                    ret[i].lines[1] = 0;
                }
                i++;
                ret = (LIMconfpart*)realloc((void*)ret,
                    (i+2)*sizeof(LIMconfpart));
                ret[i+1].lines = 0;
                start = iter;
                gottok = 0;
                line = 0;
            }
            break;
          case '{':
            if (!gottok) {
                iter = start = LIMgetconfpartreal(data, iter, &ret[i], func);
                gottok = 2;
            } else
                inbrace++;
            break;
          case '}':
            if (!inbrace)
                exit(1);
            inbrace--;
            gottok = 1;
            break;
          default:
            if (gottok == 2)
                exit(1);
            gottok = 1;
            break;
        }
        last = iter;
    }
    if (gottok == 1) {
        if (!ret[i].lines)
            ret[i].lines = (char**)malloc(2*sizeof(char*));
        else
            ret[i].lines = (char**)realloc((void*)ret[i].lines,
                (line+2)*sizeof(char*));
        *last = 0;
        ret[i].lines[line] = (char*)malloc(strlen(start)+1);
        strcpy(ret[i].lines[line], start);
        ret[i].lines[line+1] = 0;
    }
    return(ret);
}

/* pull in a brace enclosed config segment, preserve newlines. */
LIMconfpart *
LIMargifyfile(start, file)
    char *start;
    FILE *file;
{
    return(LIMargifyreal((void*)file, start, (LIMconfcontfunc)LIMfgets));
}

/* get a loading part from a bigger one */
LIMconfpart *
LIMargifypart(start, src, line)
    char *start;
    LIMconfpart *src;
    int *line;
{
    LIMconfpartpart part;
    part.src = src;
    part.line = line;
    return (LIMargifyreal((void*)&part, start, LIMconfpartpartcont));
}

/* free all the lines in a confpart */
void
LIMfreeconfpart(part)
    LIMconfpart *part;
{
    int i;

    if (!part->lines)
        return;
    for (i = 0; part->lines[i]; i++)
        free((void*)part->lines[i]);
    part->lines = 0;
}

/* load the config file */
int
LIMloadconfig(fname)
    char *fname;
{
    FILE *f;
    char *buf, *iter, *last;
    char tok[80];
    int done, i;
    LIMconfpart *parts;

    f = fopen(fname, "r");
    if (!f)
        return(-1);
    while (buf = LIMfgets(f)) {
        /* there are three parts to a declaration:  <prefix> ":" <suffix>,
         * where ":" seperates the other two. */
        iter = buf;
        last = iter;
        iter = PARSEgettoken(iter, tok);
        /* LIMfgets guarantees that at least one token exists on the line */
        for (i = 0; i < sizeof(LIMparselist)/sizeof(LIMparse); i++)
            if (STRtokinlist(tok, LIMparselist[i].name))
                break;
        if (i >= sizeof(LIMparselist)/sizeof(LIMparse))
            i = -1;
        parts = LIMargifyfile(iter, f);
        /* got lhs and rhs.  invoke the parser. */
        LIMparselist[i].func(LIMparselist[i].name[0], parts);
        /* free up the ends */
        for (i = 0; parts[i].lines; i++) {
            int j;
            for (j = 0; parts[i].lines[j]; j++)
                free((void*)parts[i].lines[j]);
            free((void*)parts[i].lines);
        }
        free((void*)parts);
    }
}

/* get the next token in a token block, ignore newlines */
char *
LIMgetparttok(iter, tok, line, block)
    char *iter, *tok;
    int *line;
    LIMconfpart *block;
{
    if (!iter) {
        *line = 0;
        iter = block->lines[0];
        if (!iter)
            return(0);
    }
    iter = PARSEgettoken(iter, tok);
    if (!iter) {
        if (!block->lines[*line+1])
            return(0);
        (*line)++;
        iter = block->lines[*line];
        iter = PARSEgettoken(iter, tok);
    }
    return(iter);
}

/* ================================================================== */
/* parse functions */

typedef struct _LIMlimblock {
    LIMconfpart *block;
    LIMtestlist *run;
} LIMlimblock;

/* limit declaration */
static int
LIMparselim(name, blocks)
    char *name;
    LIMconfpart *blocks;
{
    int i;
    LIMlimdecl *store;

    store = (LIMlimdecl*)malloc(sizeof(LIMlimdecl));
    if (!limits) {
        limits = (LIMlimdecl**)malloc(2*sizeof(LIMlimdecl*));
        limits[1] = 0;
        limits[0] = store;
    } else {
        for (i = 0; limits[i]; i++)
            continue;
        limits = (LIMlimdecl**)realloc((void*)limits, (i+2)*sizeof(LIMlimdecl*));
        limits[i+1] = 0;
        limits[i] = store;
    }
    return(LIMparselimitreal(store, blocks));
}

/* group declaration */
static int
LIMparsegroup(name, blocks)
    char *name;
    LIMconfpart *blocks;
{
    int i, line;
    LIMgroup **add, **hold;
    LIMgroupdecl *decl;
    char tok[80], *iter;
    char *chstart, *at;
    int chline, curadd;

    i = 0;
    iter = 0;
    if (!blocks[0].lines)
        return(-1);
    if (!blocks[1].lines)
        return(-1);
    if (blocks[2].lines)
        return(-1);
    iter = LIMgetparttok(iter, tok, &line, &blocks[0]);
    /* first, figure out the names of all the groups being declared here. */
    while (iter) {
        if (!i) {
            add = (LIMgroup**)malloc(2*sizeof(LIMgroup*));
            if (!add)
                return(-1);
        } else {
            hold = (LIMgroup**)realloc(add, (i+2)*sizeof(LIMgroup*));
            if (!hold) {
                for (i = 0; add[i]; i++) {
                    free((void*)add[i]->name);
                    free((void*)add[i]);
                }
                free((void*)add);
                return(-1);
            }
            add = hold;
        }
        add[i+1] = 0;
        add[i] = (LIMgroup*)malloc(sizeof(LIMgroup));
        chstart = 0;
        chline = 0;
        at = strchr(tok, '@');
        if (at)
            *at++ = 0;
        add[i]->users = LIMparsegroupdecl(&blocks[1], &chstart, &chline);
        add[i]->name = (char*)malloc(strlen(tok)+1);
        if (at) {
            add[i]->domain = (char*)malloc(strlen(at)+1);
            strcpy(add[i]->domain, at);
        } else
            add[i]->domain = 0;
        strcpy(add[i]->name, tok);
        i++;
        iter = LIMgetparttok(iter, tok, &line, &blocks[0]);
    }
    if (!i)
        /* no group names declared!  argh! */
        return(-1);
    for (i = 0; add[i]; i++) {
        if (add[i]->domain) {
            char *temp = (char*)malloc(strlen(add[i]->name)+strlen(add[i]->domain)+2);
            sprintf(temp, "%s@%s", add[i]->name, add[i]->domain);
            LIMdelgroup(temp);
            free((void*)temp);
        } else
            LIMdelgroup(add[i]->name);
    }
    if (groups) {
        for (line = 0; groups[line]; line++)
            ;
        hold = (LIMgroup**)realloc(groups, (i+line+1)*sizeof(LIMgroup*));
        if (!hold)
            return(-1);
        groups = hold;
        for (i = 0; add[i]; i++)
            groups[line+i] = add[i];
        groups[line+i] = 0;
        free((void*)add);
    } else
        groups = add;
    return(0);
}

/* alias declaration */
static int
LIMparsealias(name, blocks)
    char *name;
    LIMconfpart *blocks;
{
    char *iter, tok[80], alias[80];
    int line, numaliases;
    LIMgroup *add;
    depstruc *struc;

    if (!blocks[0].lines)
        return(-1);
    if (!blocks[1].lines)
        return(-1);
    if (blocks[2].lines)
        return(-1);
    if (!(iter = LIMgetparttok(0, alias, &line, &blocks[0])))
        return(-1);
    if (iter = LIMgetparttok(iter, tok, &line, &blocks[0]))
        return(-1);
    add = (LIMgroup*)malloc(sizeof(LIMgroup));
    add->name = (char*)malloc(strlen(alias)+1);
    strcpy(add->name, alias);
    add->domain = 0;
    line = 0;
    add->users = LIMparsegroupdecl(&blocks[1], &iter, &line);
    if (!aliases) {
        aliases = (LIMgroup**)malloc(2*sizeof(LIMgroup*));
        aliases[1] = 0;
        aliases[0] = add;
    } else {
        for (numaliases = 0; aliases[numaliases]; numaliases++)
            ;
        aliases = (LIMgroup**)realloc((void*)aliases, (numaliases+2)*sizeof(LIMgroup*));
        aliases[numaliases] = add;
        aliases[numaliases+1] = 0;
    }
    /* TODO: Add a dependancy, so that when the username changes, the alias
     * gets called again. */
    struc = new_dep(&dep_glob, 0);
    struc->data = 0;
    struc->action = LIMaliasdep;
    add_dep(find_dep(dep_glob, "filt"), struc);
    return(0);
}

/* ======================= */
/* group loading functions */
static void
LIMdelgroup(name)
    char *name;
{
    LIMgroup *del;
    int i;

    if (!groups)
        return;
    del = LIMfindgroup(name);
    if (!del)
        return;
    for (i = 0; groups[i] != del; i++)
        ;
    for (; groups[i]; i++)
        groups[i] = groups[i+1];
    LIMfreegroup(del->users);
    free((void*)del->name);
    free((void*)del);
    if (i == 1) {
        free((void*)groups);
        groups = 0;
    } else
        groups = (LIMgroup**)realloc((void*)groups, i*sizeof(LIMgroup*));
}

static LIMgroup *
LIMfindgroup(name)
    char *name;
{
    int i;
    int found = -1;
    char *domain;

    domain = strchr(name, '@');
    if (domain)
        *domain++ = 0;
    for (i = 0; groups[i]; i++) {
        if (!strcmp(groups[i]->name, name)) {
            if (domain) {
                if (groups[i]->domain && !strcmp(groups[i]->domain, domain))
                    return(groups[i]);
            } else if (found == -1)
                found = i;
            else
                return(0);
        }
    }
    if (found != -1)
        return(groups[found]);
    return(0);
}

static void
LIMfreegroup(group)
    LIMgroupdecl *group;
{
    LIMgroupdecl *del;

    while (group) {
        del = group;
        group = group->next;
        if (del->isuser)
            free((void*)del->data.user);
        else
            LIMfreegroup(del->data.group);
        free((void*)del);
    }
}

static LIMgroupdecl *
LIMcopygroupdecl(decl, domain)
    LIMgroupdecl *decl;
    char *domain;
{
    LIMgroupdecl *ret, *hold;
    char *at;

    hold = 0;
    ret = 0;
    while (decl) {
        if (!hold) {
            hold = (LIMgroupdecl*)malloc(sizeof(LIMgroupdecl));
            ret = hold;
        } else {
            hold->next = (LIMgroupdecl*)malloc(sizeof(LIMgroupdecl));
            hold = hold->next;
        }
        hold->isuser = decl->isuser;
        hold->op = decl->op;
        hold->next = 0;
        if (hold->isuser) {
            if (domain && !(at = strchr(decl->data.user, '@'))) {
                hold->data.user = (char*)malloc(strlen(decl->data.user)+strlen(domain)+2);
                sprintf(hold->data.user, "%s@%s", decl->data.user, domain);
            } else {
                hold->data.user = (char*)malloc(strlen(decl->data.user)+1);
                strcpy(hold->data.user, decl->data.user);
            }
        } else
            hold->data.group = LIMcopygroupdecl(decl->data.group, domain);
        decl = decl->next;
    }
    return(ret);
}

/* load declarations for given groups. */
static LIMgroupdecl *
LIMparsegroupdecl(block, start, line)
    LIMconfpart *block;
    char **start;
    int *line;
{
    char tok[80];
    char *iter = *start;
    char *at;
    LIMgroupdecl *hold, *ret, *new, *groupiter;
    LIMgroup *copy;
    char savop = '+';
    int i, not;

    ret = 0;
    hold = 0;
    while (iter = LIMgetparttok(iter, tok, line, block)) {
        not = 0;
        while (iter && !strcmp(tok, "!")) {
            not = !not;
            iter = LIMgetparttok(iter, tok, line, block);
        }
        if (!iter)
            break;
        new = (LIMgroupdecl*)malloc(sizeof(LIMgroupdecl));
        new->next = 0;
        new->op = savop;
        new->not = not;
        if (!ret)
            ret = new;
        if (hold)
            hold->next = new;
        hold = new;
        if (!strcmp(tok, "(")) {
            hold = LIMparsegroupdecl(block, &iter, line);
            new->isuser = 0;
            new->data.group = hold;
        } else if (*tok != '$') {
            new->isuser = 1;
            new->data.user = (char*)malloc(strlen(tok)+1);
            strcpy(new->data.user, tok);
        } else {
            /* group declaration.  copy the current data in the group, so as
             * to allow a second group decl to modify the first, and to
             * destroy any chance of infinite loops when we parse this. */
            if (tok[1] == '*') {
                /* special, "everyone", unmodifiable group. */
                new->data.user = (char*)malloc(strlen(tok)+1);
                strcpy(new->data.user, tok);
                new->isuser = 1;
            } else {
                copy = LIMfindgroup(tok+1);
                if (copy) {
                    new->data.group = LIMcopygroupdecl(copy->users, copy->domain);
                    new->isuser = 0;
                }
            }
        }
        iter = LIMgetparttok(iter, tok, line, block);
        if (!iter || *tok == ')')
            break;
        savop = *tok;
    }
    *start = iter;
    return (ret);
}

static int
LIMuseringroupdecl(user, group, domain)
    char *user;
    LIMgroupdecl *group;
    char *domain;
{
    int ret = 0;
    int inpres, inlast;
    char *temp, *at;

    inlast = 0;
    while (group) {
        if (group->isuser) {
            at = strchr(group->data.user, '@');
            if (!strncmp(group->data.user, "$*", 2)) {
                if (at) {
                    temp = strchr(user, '@');
                    inpres = !strcmp(temp, at);
                } else
                    inpres = 1;
            } else if (!at && domain) {
                temp = (char*)malloc(strlen(group->data.user)+strlen(domain)+2);
                sprintf(temp, "%s@%s", group->data.user, domain);
                inpres = !strcmp(user, temp);
                free((void*)temp);
            } else
                inpres = !strcmp(group->data.user, user);
        } else
            inpres = LIMuseringroupdecl(user, group->data.group, domain);
        if (group->not)
            inpres = !inpres;
        if (inpres) {
            switch (group->op) {
            case '+':
                ret = 1;
                break;
            case '-':
                ret = 0;
                break;
            case '|':
                inpres = inlast;
                break;
            }
        }
        group = group->next;
        inlast = inpres;
    }
    return (ret);
}

static int
LIMuseringroup(user, group)
    char *user;
    LIMgroup *group;
{
    return LIMuseringroupdecl(user, group->users, group->domain);
}

static int
LIMuseringroupname(user, name)
    char *user, *name;
{
    LIMgroup *group;
    char *at;

    if (*name == '*') {
        at = strchr(name, '@');
        if (!at)
            return(1);
        user = strchr(user, '@');
        if (!user)
            return(0);
        return(!strcmp(user, at));
    }
    group = LIMfindgroup(name);
    if (!group)
        return(0);
    return (LIMuseringroup(user, group));
}

static void
LIMaliasdep(dep, data)
    depstruc *dep;
    void *data;
{
    FILT *filt = (FILT*)data;
    deplist *list;
    depstruc *new;

    list = FILTgetdep(filt);
    new = new_dep(list, 0);
    new->data = (void*)filt;
    new->action = LIMrunalias;
    add_dep(find_dep(*list, "user"), new);
}

static void
LIMrunalias(dep, data)
    depstruc *dep;
    void *data;
{
    FILT *filt = (FILT*)data;
    int i;
    char *user;
    char buf[20];

    if (!aliases)
        return;
    user = FILTgetuser(filt);
    if (!user)
        return;
    for (i = 0; aliases[i]; i++) {
        if (LIMuseringroup(user, aliases[i])) {
            sprintf(buf, ".alias%d", i);
            if (FILThasint(filt, buf))
                return;
            FILTaddint(filt, buf, 1);
            FILTsetuser(filt, aliases[i]->name);
            FILTdelint(filt, buf);
            break;
        }
    }
}

/* ======================= */
/* limit loading functions */

/* the meat of it all.  recursively load a block of limits, set up the tests,
 * set up the child FILTs. */
int
LIMparselimitreal(store, blocks)
    LIMlimdecl *store;
    LIMconfpart *blocks;
{
    FILT *new;
    LIMconfpart *limtest, *limblock;
    char *iter, *start, tok[80];
    depstruc *dep;
    int line, i;
    LIMlimblock *tester;
    LIMconfpart *filtblock;

    new = FILTalloc();
    store->filt = new;
    store->run = 0;
    store->children = 0;
    if (!blocks[0].lines || !blocks[1].lines)
        return(-1);
    if (!blocks[1].lines)
        return(-1);
    if (*blocks[0].lines[0]) {
        limtest = (LIMconfpart*)malloc(sizeof(LIMconfpart));
        limblock = &blocks[1];
        for (line = 0; blocks[0].lines[line]; line++)
            continue;
        limtest->lines = (char**)malloc((line+1)*sizeof(char*));
        /* the test block is basically meaningless until we have a FILT* to
         * get dependancies from */
        for (line = 0; blocks[0].lines[line]; line++) {
            limtest->lines[line] = (char*)malloc(strlen(blocks[0].lines[line])+1);
            strcpy(limtest->lines[line], blocks[0].lines[line]);
        }
        limtest->lines[line] = 0;
        tester = (LIMlimblock*)malloc(sizeof(LIMlimblock));
        tester->block = limtest;
        tester->run = (LIMtestlist*)malloc(sizeof(LIMtestlist));
        tester->run->next = 0;
        tester->run->pass = 0;
        tester->run->op = 0;
        tester->run->not = 0;
        store->run = tester->run;
        dep = new_dep(&dep_glob, 0);
        dep->data = (void*)tester;
        dep->action = LIMparsetest;
        add_dep(find_dep(dep_glob, "filt"), dep);
    } else {
        store->run = 0;
        limblock = &blocks[1];
    }
    line = 0;
    if (limblock[1].lines) {
        /* interpret as one statement */
        filtblock = limblock;
        start = iter = limblock[0].lines[0];
        iter = PARSEgettoken(iter, tok);
        if (!iter)
            return(-1);
        limblock[0].lines[0] = iter;
        for (i = 0; LIMloadlist[i].name; i++)
            if (STRtokinlist(tok, LIMloadlist[i].name))
                break;
        if (LIMloadlist[i].name)
            if (LIMloadlist[i].func(new, LIMloadlist[i].name[0], filtblock)) {
                sprintf(tok, sLIMload, LIMloadlist[i].name[0]);
                FILTaddint(new, tok, 1);
            }
        limblock[0].lines[0] = start;
        return(0);
    }
    while (limblock->lines[line]) {
        iter = limblock->lines[line];
        iter = PARSEgettoken(iter, tok);
        filtblock = LIMargifypart(iter, limblock, &line);
        if (!strcasecmp(tok, "limit")) {
            if (!store->children) {
                store->children = (LIMlimdecl**)malloc(2*sizeof(LIMlimdecl*));
                i = 0;
            } else {
                for (i = 0; store->children[i]; i++)
                    continue;
                store->children = (LIMlimdecl**)realloc((void*)store->children, (i+2)*sizeof(LIMlimdecl*));
            }
            store->children[i] = (LIMlimdecl*)malloc(sizeof(LIMlimdecl));
            store->children[i+1] = 0;
            LIMparselimitreal(store->children[i], filtblock);
        } else {
            for (i = 0; LIMloadlist[i].name; i++)
                if (STRtokinlist(tok, LIMloadlist[i].name))
                    break;
            if (LIMloadlist[i].name) {
                if (LIMloadlist[i].func(new, LIMloadlist[i].name[0], filtblock)) {
                    sprintf(tok, sLIMload, LIMloadlist[i].name[0]);
                    FILTaddint(new, tok, 1);
                }
            }
        }
        for (i = 0; filtblock[i].lines; i++)
            LIMfreeconfpart(&filtblock[i]);
        free((void*)filtblock);
        line++;
    }
    return(0);
}

/* parse a test block */
void
LIMparsetest(dep, data)
    depstruc *dep;
    void *data;
{
    LIMconfpart *block, child;
    LIMtestlist *itertest;
    FILT *filt;
    char *iter, tok[80], testtok[80], sav, *last, *startdecl;
    int line, i, opfound, testfound;
    LIMlimblock *duh;

    filt = (FILT*)data;
    line = 0;
    duh = (LIMlimblock*)dep->data;
    block = duh->block;
    itertest = duh->run;
    startdecl = 0;
    iter = block->lines[0];
    last = iter;
    *testtok = 0;
    child.lines = 0;
    while (iter = LIMgetparttok(iter, tok, &line, block)) {
        opfound = -1;
        for (i = 0; i < sizeof(LIMoplist) / sizeof(LIMop); i++) {
            if (STRtokinlist(tok, LIMoplist[i].name)) {
                opfound = i;
                break;
            }
        }
        if (opfound >= 0) {
            /* found an operator. */
            switch (LIMoplist[opfound].type) {
              case kLIMopnot:
                itertest->not = !itertest->not;
                break;
              default:
                itertest->op = LIMoplist[opfound].type;
            }
            if (*testtok) {
                /* terminate the previous declaration, set up the
                 * test-parse block */
                sav = *last;
                *last = 0;
                LIMdeplist[testfound].func(filt, &child, &itertest->pass);
                if (child.lines) {
                    for (i = 0; child.lines[i]; i++)
                        free((void*)child.lines[i]);
                    free((void*)child.lines);
                    child.lines = 0;
                }
                *last = sav;
                itertest->next = (LIMtestlist*)malloc(sizeof(LIMtestlist));
                itertest = itertest->next;
                itertest->next = 0;
                itertest->pass = 1;
                itertest->not = 0;
                itertest->op = 0;
            }
            /* look for a new test */
            *testtok = 0;
        } else if (!strcmp(tok, "{")) {
            /* embedded declaration */
            if (*testtok && !child.lines) {
                iter = LIMgetconfpartpart(block, iter, &line, &child);
                LIMdeplist[testfound].func(filt, &child, &itertest->pass);
                child.lines = 0;
            }
        } else if (!*testtok) {
            startdecl = iter;
            strcpy(testtok, tok);
            testfound = -1;
            for (i = 0; i < sizeof(LIMdeplist) / sizeof(LIMdep); i++) {
                if (STRtokinlist(testtok, LIMdeplist[i].name)) {
                    testfound = i;
                    break;
                }
            }
        } else {
            if (!child.lines || !child.lines[0]) {
                if (!child.lines) {
                    child.lines = (char**)malloc(2*sizeof(char*));
                    child.lines[1] = 0;
                }
                child.lines[0] = (char*)malloc(strlen(tok)+1);
                strcpy(child.lines[0], tok);
            } else {
                child.lines[0] = (char*)realloc((void*)child.lines[0],
                    strlen(child.lines[0])+strlen(tok)+2);
                sprintf(child.lines[0]+strlen(child.lines[0]), " %s", tok);
            }
        }
        last = iter;
    }
    if (*testtok) {
        /* terminate the previous declaration, set up the test-parse block */
        sav = *last;
        *last = 0;
        LIMdeplist[testfound].func(filt, &child, &itertest->pass);
        if (child.lines) {
            for (i = 0; child.lines[i]; i++)
                free((void*)child.lines[i]);
            free((void*)child.lines);
            child.lines = 0;
        }
        *last = sav;
    }
}

/* ================================================================== */
/* condition tests */

typedef struct _LIMtestdep {
    int *store;
    void *data;
} LIMtestdep;

/* dependancy for user */
static void
LIMuserdep(filt, args, store)
    FILT *filt;
    LIMconfpart *args;
    int *store;
{
    deplist *list;
    depstruc *dep;
    LIMtestdep *data;
    int line, len;
    char *copy;

    list = FILTgetdep(filt);
    if (!args->lines)
        return;
    dep = new_dep(list, 0);
    dep->action = LIMtestuser;
    data = (LIMtestdep*)malloc(sizeof(LIMtestdep));
    data->store = store;
    len = 0;
    for (line = 0; args->lines[line]; line++)
        len += strlen(args->lines[line])+1;
    copy = (char*)malloc(len);
    *copy = 0;
    for (line = 0; args->lines[line]; line++) {
        strcat(copy, args->lines[line]);
        if (args->lines[line+1])
            strcat(copy, " ");
    }
    data->data = (void*)copy;
    dep->data = data;
    add_dep(find_dep(*list, "user"), dep);
}

/* user */
static void
LIMtestuser(dep, data)
    depstruc *dep;
    void *data;
{
    char tok[80];
    LIMtestdep *depdata;
    char *iter;
    FILT *filt = (FILT*)data;
    char *user;

    depdata = (LIMtestdep*)dep->data;
    iter = (char*)depdata->data;
    *depdata->store = 0;
    user = FILTgetuser(filt);
    if (!user)
        return;
    while (iter = PARSEgettoken(iter, tok)) {
        if (!strcmp(tok, user)) {
            *depdata->store = 1;
            return;
        }
    }
    *depdata->store = 0;
}

/* dependancy for group */
static void
LIMgroupdep(filt, args, store)
    FILT *filt;
    LIMconfpart *args;
    int *store;
{
    deplist *list;
    depstruc *dep;
    LIMtestdep *data;
    int line, len;
    char *copy;

    list = FILTgetdep(filt);
    if (!args->lines)
        return;
    dep = new_dep(list, 0);
    dep->action = LIMtestgroup;
    data = (LIMtestdep*)malloc(sizeof(LIMtestdep));
    data->store = store;
    len = 0;
    for (line = 0; args->lines[line]; line++)
        len += strlen(args->lines[line])+1;
    copy = (char*)malloc(len);
    *copy = 0;
    for (line = 0; args->lines[line]; line++) {
        strcat(copy, args->lines[line]);
        if (args->lines[line+1])
            strcat(copy, " ");
    }
    data->data = (void*)copy;
    dep->data = data;
    add_dep(find_dep(*list, "user"), dep);
}

/* group */
static void
LIMtestgroup(dep, data)
    depstruc *dep;
    void *data;
{
    char tok[80];
    LIMtestdep *depdata;
    char *iter;
    FILT *filt = (FILT*)data;
    char *user;

    depdata = (LIMtestdep*)dep->data;
    iter = (char*)depdata->data;
    *depdata->store = 0;
    user = FILTgetuser(filt);
    if (!user)
        return;
    while (iter = PARSEgettoken(iter, tok)) {
        if (LIMuseringroupname(user, tok)) {
            *depdata->store = 1;
            return;
        }
    }
}

/* dependancy for message dest */
static void
LIMdestdep(filt, args, store)
    FILT *filt;
    LIMconfpart *args;
    int *store;
{
    deplist *list;
    depstruc *dep;
    LIMtestdep *data;
    int line, len;
    char *copy;

    if (!args->lines || !args->lines[0] || args->lines[1])
        return;
    list = FILTgetdep(filt);
    dep = new_dep(list, 0);
    dep->action = LIMtestdest;
    data = (LIMtestdep*)malloc(sizeof(LIMtestdep));
    data->store = store;

    copy = (char*)malloc(strlen(args->lines[0])+1);
    strcpy(copy, args->lines[0]);

    data->data = (void*)copy;
    dep->data = data;
    add_dep(find_dep(*list, "mesg"), dep);
}

/* message destination */
static void
LIMtestdest(dep, data)
    depstruc *dep;
    void *data;
{
    FILT *filt = (FILT*)data;
    LIMtestdep *depdata = (LIMtestdep*)dep->data;
    char **nglist = 0, **patlist = 0;
    char *ng, *ngsave, *patsave;
    int index;

    ng = FILTgetheader(filt, "Newsgroups:");
    ngsave = (char*)malloc(strlen(ng)+1);
    strcpy(ngsave, ng);
    LIMnggetlist(&nglist, ngsave);
    patsave = (char*)malloc(strlen((char*)depdata->data)+1);
    strcpy(patsave, (char*)depdata->data);
    LIMnggetlist(&patlist, patsave);
    *depdata->store = LIMpermmatch(patlist, nglist);
    free((void*)ngsave);
    free((void*)nglist);
    free((void*)patsave);
    free((void*)patlist);
}

/* ================================================================= */
/* database functions */

char **
LIMdata(lim, name)
    LIM *lim;
    char *name;
{
    LIMdbdata *data = LIMfinddbdata((LIMpriv*)lim, name);
    if (!data)
        return(0);
    return(data->tmdata);
}

char *
LIMnewdata(lim, name)
    LIM *lim;
    char *name;
{
    LIMdbdata *data = LIMfinddbdata((LIMpriv*)lim, name);
    if (!data)
        return(0);
    return(data->newtmdata);
}

void
LIMsetdata(lim, name, value)
    LIM *lim;
    char *name, *value;
{
    int i;
    LIMdbdata *data = LIMadddbdata((LIMpriv*)lim, name);

    if (data->newtmdata)
        free((void*)data->newtmdata);
    if (!value) {
        data->newtmdata = 0;
        data->action |= kLIMdeltmdata;
        return;
    }
    data->newtmdata = (char*)malloc(strlen(value)+1);
    strcpy(data->newtmdata, value);
}

void
LIMresetdata(lim, name)
    LIM *lim;
    char *name;
{
    LIMdbdata *data = LIMfinddbdata((LIMpriv*)lim, name);
    if (!data)
        return;
    data->action &= ~kLIMdeltmdata;
    if (!data->newtmdata)
        return;
    free((void*)data->newtmdata);
    data->newtmdata = 0;
}

char *
LIMstatdata(lim, name)
    LIM *lim;
    char *name;
{
    LIMdbdata *data = LIMfinddbdata((LIMpriv*)lim, name);
    if (!data)
        return(0);
    return(data->statdata);
}

void
LIMblock(lim, name)
    LIM *lim;
    char *name;
{
    LIMblockdata **blockdata;
    int i;

    blockdata = *((LIMpriv*)lim)->blockdata;
    if (!blockdata) {
        blockdata = (LIMblockdata**)malloc(2*sizeof(LIMblockdata*));
        i = 0;
    } else {
        for (i = 0; blockdata[i]; i++)
            if (!strcmp(blockdata[i]->name, name))
                return;
        blockdata = (LIMblockdata**)realloc((void*)blockdata,
            (i+2)*sizeof(LIMblockdata*));
    }
    blockdata[i+1] = 0;
    blockdata[i] = (LIMblockdata*)malloc(sizeof(LIMblockdata*));
    blockdata[i]->name = (char*)malloc(strlen(name)+1);
    strcpy(blockdata[i]->name, name);
    *((LIMpriv*)lim)->blockdata = blockdata;
}

int
LIMisblock(lim, name)
    LIM *lim;
    char *name;
{
    LIMblockdata **blockdata;
    int i;
    int blockthis;

    blockdata = *((LIMpriv*)lim)->blockdata;
    if (!blockdata)
        return(kLIMnotblocked);
    /* blockdata exists, so blocking is implicit. */
    blockthis = 0;
    for (i = 0; blockdata[i]; i++)
        if (!strcmp(blockdata[i]->name, name))
            blockthis = kLIMblockthis;
    if (i == 1)
        return(blockthis);
    return(kLIMblockelse | blockthis);
}

void
LIMaddsync(lim, name, sf)
    LIM *lim;
    char *name;
    LIMsyncfunc sf;
{
    LIMdbdata *data;

    data = LIMadddbdata(lim, name);
    data->sync = sf;
}

void
LIMaddfinish(lim, name, fin)
    LIM *lim;
    char *name;
    LIMfinishfunc fin;
{
    LIMdbdata *data;

    data = LIMadddbdata(lim, name);
    data->finish = fin;
}

int
LIMparmloaded(filt, name)
    FILT *filt;
    char *name;
{
    char buf[80];
    FILT *child;

    if (!(child = FILTgetfilt(filt, "loaded")))
        return(0);
    sprintf(buf, sLIMload, name);
    return (FILThasint(child, buf));
}
