#include <stdarg.h>

#include "gucs_err.h"
#include "gucs_err_code.h"

static GUCS_error_type *gucs_err_code_set_impl(GUCS_error_type *, int, int);
static GUCS_error_type *gucs_err_locked(void);

static GUCS_error_type *gucs_no_mem_err;
static GUCS_error_type *gucs_usage_err;
static GUCS_error_type *gucs_lock_err;

#define GUCS_ERROR_HASH_BITS	4
#define SIZE_GUCS_ERROR_HASH	(1<<GUCS_ERROR_HASH_BITS)
#define GUCS_ERROR_HASH_MASK	(SIZE_GUCS_ERROR_HASH-1)
#define GUCS_HASH_BITS_PER_BYTE	(8/GUCS_ERROR_HASH_BITS)

typedef struct _GUCS_error_data_el
{
	char *name;
	int raw;
	void *data;
	GUCS_error_accessor_type accessor;
	GUCS_error_cleanup_type cleanup;
	struct _GUCS_error_data_el *next;
} GUCS_error_data_el;

struct _GUCS_error_type
{
	int refcount;
	int locked;
	GUCS_error_data_el data_els[SIZE_GUCS_ERROR_HASH];
};

static int GUCS_err_hash_name(char *name);
static GUCS_error_data_el *GUCS_data_el_find(GUCS_error_type *, char *, int);

static int
GUCS_err_hash_name(char *name)
{
	int i, j;
	int res;

	res = 0;
	for (i = 0; name[i]; i++)
	{
		/* name is a string, so the high-order bits should be pretty
		 * uniform.  xor the low order bits, and the high order will
		 * fall out in the return statement. */
		res = ((res<<(GUCS_ERROR_HASH_BITS-1)) | (res>>1)) ^ name[i];
	}
	return (res & GUCS_ERROR_HASH_MASK);
}

static GUCS_error_data_el *
GUCS_data_el_find(GUCS_error_type *error, char *name, int hash)
{
	GUCS_error_data_el *iter;

	iter = &error->data_els[hash];
	if (iter->name == (char *) 0)
		return (0);

	while (strcmp(iter->name, name) != 0)
	{
		if (iter->next == (GUCS_error_data_el *) 0)
			return (0);
		iter = iter->next;
	}
	return (iter);
}

int
gucs_err_init()
{
	GUCS_error_type *err;

	/* initialize gucs_no_mem global. */
	gucs_no_mem_err = gucs_err_alloc();
	if (gucs_no_mem_err == (GUCS_error_type *) 0)
		return (-1);
	err = gucs_err_code_set(gucs_no_mem_err,
	                        GUCS_ERR_INTERNAL, GUCS_ERR_NO_MEM, -1);
	if (err != (GUCS_error_type *) 0)
		return (-1);
	err = gucs_err_commit(gucs_no_mem_err);
	if (err != (GUCS_error_type *) 0)
		return (-1);

	/* initialize gucs_usage_err global. */
	gucs_usage_err = gucs_err_alloc();
	if (gucs_usage_err == (GUCS_error_type *) 0)
		return (-1);
	err = gucs_err_code_set(gucs_usage_err,
	                        GUCS_ERR_INTERNAL, GUCS_ERR_USAGE, -1);
	if (err != (GUCS_error_type *) 0)
		return (-1);
	err = gucs_err_commit(gucs_no_mem_err);
	if (err != (GUCS_error_type *) 0)
		return (-1);

	/* initialize gucs_lock_err global. */
	gucs_lock_err = gucs_err_alloc();
	if (gucs_lock_err == (GUCS_error_type *) 0)
		return (-1);
	err = gucs_err_code_set(gucs_lock_err,
	                        GUCS_ERR_INTERNAL, GUCS_ERR_LOCKED, -1);
	if (err != (GUCS_error_type *) 0)
		return (-1);
	err = gucs_err_commit(gucs_lock_err);
	if (err != (GUCS_error_type *) 0)
		return (-1);

	/* all done. */
	return (0);
}

GUCS_error_type *
gucs_err_alloc()
{
	GUCS_error_type *ret;
	int i;

	ret = (GUCS_error_type *) malloc(sizeof(GUCS_error_type));
	if (ret == (GUCS_error_type *) 0)
		goto err1;
	for (i = 0; i < SIZE_GUCS_ERROR_HASH; i++)
		ret->data_els[i].name = 0;
	ret->refcount = 0;
	ret->locked = 0;
	gucs_err_ref(ret);
	return (ret);

	err1:
		return (0);
}

GUCS_error_type *
gucs_err_commit(GUCS_error_type *err)
{
	if (err->locked != 0)
		return (gucs_err_locked());
	err->locked = 1;
	return ((GUCS_error_type *) 0);
}

void
gucs_err_ref(GUCS_error_type *err)
{
	err->refcount++;
}

void
gucs_err_unref(GUCS_error_type *err)
{
	GUCS_error_data_el *iter, *del;
	int i, first;

	err->refcount--;
	if (err->refcount > 0)
		return;

	for (i = 0; i < SIZE_GUCS_ERROR_HASH; i++)
	{
		if (err->data_els[i].name != (char *) 0)
		{
			iter = &err->data_els[i];
			first = 1;
			while (iter != (GUCS_error_data_el *) 0)
			{
				del = iter;
				iter = iter->next;
				if (del->cleanup != (GUCS_error_cleanup_type) 0)
					(*del->cleanup)(err,
					                del->name, del->data);
				free((void *) del->name);
				if (! first)
					free((void *) del);
				first = 0;
			}
		}
	}

	free((void *) err);
}

void *
gucs_error_data(GUCS_error_type *error, char *name)
{
	GUCS_error_data_el *el;
	int hash;

	hash = GUCS_err_hash_name(name);
	el = GUCS_data_el_find(error, name, hash);
	if (el == (GUCS_error_data_el *) 0)
		return (0);

	if (el->accessor != (GUCS_error_accessor_type) 0)
		return ((*el->accessor)(error, name, el->data));
	return (el->data);
}

GUCS_error_type *
gucs_err_code_set(GUCS_error_type *error, ...)
{
	GUCS_error_type *ret;
	va_list args;
	int code, level;

	if (error->locked != 0)
		return (gucs_err_locked());

	va_start(args, error);
	code = va_arg(args, int);
	level = 0;
	while (code != -1)
	{
		ret = gucs_err_code_set_impl(error, level, code);
		if (ret != (GUCS_error_type *) 0)
			goto err1;
		level++;
		code = va_arg(args, int);
	}
	va_end(args);

	return ((GUCS_error_type *) 0);

	err1:
		return (ret);
}

GUCS_error_type *
gucs_err_data_set(GUCS_error_type *error, char *name,
                  GUCS_error_accessor_type accessor,
                  GUCS_error_cleanup_type cleanup, void *data)
{
	GUCS_error_data_el *el, *iter;
	int hash;

	/* make sure the data is locked before writing. */
	if (error->locked != 0)
		return (gucs_err_locked());

	/* look up the old data el in the top-level error only, to see if we're
	 * overwriting it. */
	hash = GUCS_err_hash_name(name);
	el = GUCS_data_el_find(error, name, hash);
	if (el != (GUCS_error_data_el *) 0)
	{
		/* we're overwriting an existing el, so we only need to clean
		 * up its data, and fill it in anew. */
		if (el->cleanup != (GUCS_error_cleanup_type) 0)
			(*el->cleanup)(error, name, el->data);
	} else
	{
		/* need to allocate a new el for this data. */
		if (error->data_els[hash].name != (char *) 0)
		{
			el = (GUCS_error_data_el *)
			  malloc(sizeof(GUCS_error_data_el));
			if (el == (GUCS_error_data_el *) 0)
				goto err1;

			/* need to append to the hash chain. */
			iter = &error->data_els[hash];
			while (iter->next != (GUCS_error_data_el *) 0)
				iter = iter->next;
		} else
		{
			/* the head element is always allocated. */
			iter = 0;
			el = &error->data_els[hash];
		}

		/* name the element we've added. */
		el->name = (char *) malloc(strlen(name)+1);
		if (el->name == (char *) 0)
			goto err2;
		strcpy(el->name, name);

		if (iter != (GUCS_error_data_el *) 0)
			iter->next = el;
		el->next = 0;
	}

	/* el exists, and is named. */
	el->cleanup = cleanup;
	el->accessor = accessor;
	el->data = data;
	return (0);

	err2:
		if (el != &error->data_els[hash])
			free((void *) el);
	err1:
		return (gucs_err_no_mem());
}

typedef struct _GUCS_error_code_type
{
	int *name;
	int name_used;
	int name_size;
} GUCS_error_code_type;

#define GUCS_ERROR_CODE_DEFAULT_SIZE	2

static void
GUCS_error_code_cleanup(GUCS_error_type *error, char *name, void *data)
{
	GUCS_error_code_type *error_code;

	error_code = (GUCS_error_code_type *) data;
	free((void *) error_code->name);
	free((void *) error_code);
}

/* get a level of the error_code hierarchy. */
int
gucs_error_code(GUCS_error_type *error, int level)
{
	GUCS_error_code_type *error_code;

	error_code =
	  (GUCS_error_code_type *) gucs_error_data(error, "GUCS.error_code");
	if (error_code == (GUCS_error_code_type *) 0 ||
	  error_code->name_used < level)
		return (-1);
	return (error_code->name[level]);
}

/* set a level of the error_code hierarchy. */
static GUCS_error_type *
gucs_err_code_set_impl(GUCS_error_type *error, int level, int code)
{
	GUCS_error_code_type *error_code;
	int *hold;

	error_code =
	  (GUCS_error_code_type *) gucs_error_data(error, "GUCS.error_code");

	if (level == 0)
	{
		if (error_code != (GUCS_error_code_type *) 0)
			return (gucs_err_usage());
		error_code =
		  (GUCS_error_code_type *) malloc(sizeof(GUCS_error_code_type));

		if (error_code == (GUCS_error_code_type *) 0)
			goto err1;

		/* allocate the name... */
		error_code->name_size = GUCS_ERROR_CODE_DEFAULT_SIZE;
		error_code->name =
		  (int *) malloc(error_code->name_size*sizeof(int));
		if (error_code->name == (int *) 0)
			goto err2;
		error_code->name_used = 0;
		/* store the error_code in the structure, as the last step
		 * which can fail. */
		if (gucs_err_data_set(error, "GUCS.error_code",
		                      0, GUCS_error_code_cleanup,
		                      (void *) error_code) < 0)
			goto err3;
	} else
	{
		if (error_code == (GUCS_error_code_type *) 0 ||
		  level != error_code->name_used + 1)
			return (gucs_err_usage());

		if (error_code->name_used == error_code->name_size)
		{
			/* need to grow the name. */
			hold =
			  (int *) realloc((void *) error_code->name,
			                  error_code->name_size*2*sizeof(int));
			if (hold == (int *) 0)
				goto err1;
			error_code->name_size *= 2;
			error_code->name = hold;
		}
	}
	/* at this point, the error_code has enough room to fill in the
	 * actual error data. */
	error_code->name[level] = code;
	error_code->name_used = level;
	return (0);

	err3:
		free((int *) error_code->name);
	err2:
		free((void *) error_code);
	err1:
		return (gucs_err_no_mem());
}

GUCS_error_type *
gucs_err_no_mem()
{
	gucs_err_ref(gucs_no_mem_err);
	return (gucs_no_mem_err);
}

GUCS_error_type *
gucs_err_usage()
{
	gucs_err_ref(gucs_usage_err);
	return (gucs_usage_err);
}

static GUCS_error_type *
gucs_err_locked()
{
	gucs_err_ref(gucs_lock_err);
	return (gucs_lock_err);
}

GUCS_error_type *
gucs_err_errno(int err)
{
	GUCS_error_type *ret, *status;

	if (err == -1)
	{
		status = gucs_err_usage();
		goto err1;
	}

	ret = gucs_err_alloc();
	if (ret == (GUCS_error_type *) 0)
	{
		status = gucs_err_no_mem();
		goto err1;
	}
	status = gucs_err_code_set(ret, GUCS_ERR_ERRNO, err, -1);
	if (status != (GUCS_error_type *) 0)
		goto err2;

	return (ret);

	err2:
		gucs_err_unref(ret);
	err1:
		return (status);
}
