#include <sys/types.h>
#include <netdb.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>

#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "libgucs.h"
#include "gucs_internal.h"
#include "gucs_err_code.h"

struct _GUCS_cf
{
	char *filename;
	int fd;

	GUCS_cf_buffer *buflist;
	GUCS_cf_position *included;
};

struct _GUCS_cf_buffer
{
	GUCS_cf *cf;
	GUCS_cf_buffer *prev, *next;

	int refcount;

	char line[256];
	long offset;
	size_t length;
};

struct _GUCS_cf_position
{
	GUCS_cf_buffer *curpos;
	char *iter;

	int lineno;
	long offset;
};

/* buffer functions */
static GUCS_cf_buffer *GUCS_fseek(GUCS_cf *cf, long offset);
static void GUCS_buf_ref(GUCS_cf_buffer *buf);
static void GUCS_buf_unref(GUCS_cf_buffer *buf);

/* error types */
static GUCS_error_type *GUCS_err_string(GUCS_cf_position *, GUCS_cf_position *);
static GUCS_error_type *GUCS_err_no_match(GUCS_cf_position *,
                                          GUCS_cf_position *);
static GUCS_error_type *GUCS_err_no_eof(GUCS_cf_position *, GUCS_cf_position *);

static void GUCS_err_pos_cleanup(GUCS_error_type *, char *, void *);

GUCS_parse_el GUCS_lbrace[] =
{
	{ GUCS_parse_token, 0, (void *) "{", 1, 0, 0 },
	{ GUCS_parse_to_eoln, 0, 0, 0, 0, 0 },
	{ 0, 0, 0, 0, 0, 0 }
};

GUCS_parse_el GUCS_rbrace[] =
{
	{ GUCS_parse_token, 0, (void *) "}", 1, 0, 0 },
	{ GUCS_parse_to_eoln, 0, 0, 0, 0, 0 },
	{ 0, 0, 0, 0, 0, 0 }
};

GUCS_parse_el GUCS_blank[] =
{
	{ GUCS_parse_to_eoln, 0, 0, 0, 0, 0 },
	{ 0, 0, 0, 0, 0, 0 }
};

GUCS_cf *
gucs_fopen(char *filename)
{
	GUCS_cf *ret;
	int fd;

	fd = open(filename, O_RDONLY, 0);
	if (fd < 0)
		goto err1;
	ret = malloc(sizeof(GUCS_cf));
	if (ret == (GUCS_cf *) 0)
		goto err2;
	ret->filename = (char *) malloc(strlen(filename)+1);
	if (ret->filename == (char *) 0)
		goto err3;
	strcpy(ret->filename, filename);
	ret->fd = fd;
	ret->buflist = 0;
	ret->included = 0;
	return (ret);

	/* error cases */
	err3:
		free((void *) ret);
	err2:
		close(fd);
	err1:
		return (0);
}

void
gucs_fclose(GUCS_cf *cf)
{
	close(cf->fd);
	free((void *) cf->filename);
	free((void *) cf);
}

static GUCS_cf_buffer *
GUCS_fseek(GUCS_cf *cf, long offset)
{
	GUCS_cf_buffer *hold, *iter;
	long curplace;

	iter = cf->buflist;
	hold = 0;
	curplace = 0;
	while (iter != (GUCS_cf_buffer *) 0 && curplace < offset)
	{
		curplace = iter->offset + iter->length;

		hold = iter;
		iter = iter->next;
	}
	/* Couldn't find our spot.  So, we need to read in a new block. */
	if (offset != curplace)
		/* XXX: error. */;

	if (iter != (GUCS_cf_buffer *) 0)
	{
		/* found it. */
		GUCS_buf_ref(iter);
		return (iter);
	}

	/* fill in the data from the file. */
	iter = hold;
	hold = (GUCS_cf_buffer *) malloc(sizeof(GUCS_cf_buffer));
	hold->cf = cf;
	hold->prev = iter;
	hold->next = 0;
	hold->refcount = 0;
	hold->offset = offset;
	/* an fseek should not be necessary. */
	hold->length = read(cf->fd, hold->line, sizeof(hold->line));
	if (hold->length == 0)
	{
		free((void *) hold);
		return (0);
	}
	if (iter != (GUCS_cf_buffer *) 0)
		iter->next = hold;
	else
		cf->buflist = hold;

	GUCS_buf_ref(hold);
	return (hold);
}

static void
GUCS_buf_ref(GUCS_cf_buffer *buf)
{
	buf->refcount++;
}

static void
GUCS_buf_unref(GUCS_cf_buffer *buf)
{
	GUCS_cf_buffer *iter;
	GUCS_cf *cf;

	buf->refcount--;
	if (buf->refcount == 0 && buf->prev == (GUCS_cf_buffer *) 0)
	{
		/* nothing referring to this buffer, anymore. */

		/* a GUCS_fseek implementation detail dictates that the
		 * file is read forward, and that all buffers in a
		 * cf->buflist be continuous, so we only free from the
		 * beginning. */
		cf = buf->cf;
		while (buf != (GUCS_cf_buffer *) 0 && buf->refcount == 0)
		{
			iter = buf;
			buf = buf->next;
			free((void *) iter);
		}
		if (buf != (GUCS_cf_buffer *) 0)
			buf->prev = (GUCS_cf_buffer *) 0;
		cf->buflist = buf;
	}
}

int
GUCS_pos_lineno(GUCS_cf_position *pos)
{
	return (pos->lineno);
}

char *
GUCS_pos_filename(GUCS_cf_position *pos)
{
	return (pos->curpos->cf->filename);
}

GUCS_cf_position *
GUCS_file_start(GUCS_cf *cf)
{
	GUCS_cf_buffer *buf;
	GUCS_cf_position *pos;

	buf = GUCS_fseek(cf, 0);
	pos = (GUCS_cf_position *) malloc(sizeof(GUCS_cf_position));
	if (pos == (GUCS_cf_position *) 0)
		goto err2;

	/* fill in GUCS_cf_position structure. */
	pos->curpos = buf;
	if (buf != (GUCS_cf_buffer *) 0)
		pos->iter = buf->line;
	else
		pos->iter = (char *) 0;
	pos->lineno = 1;
	pos->offset = 0;
	return (pos);

	err2:
		GUCS_buf_unref(buf);
	err1:
		return (0);
}

GUCS_cf_position *
GUCS_pos_copy(GUCS_cf_position *pos)
{
	GUCS_cf_position *ret;

	ret = (GUCS_cf_position *) malloc(sizeof(GUCS_cf_position));
	if (ret == (GUCS_cf_position *) 0)
		goto err1;
	ret->curpos = pos->curpos;
	ret->iter = pos->iter;
	ret->lineno = pos->lineno;
	ret->offset = pos->offset;
	if (ret->curpos != (GUCS_cf_buffer *) 0)
		GUCS_buf_ref(ret->curpos);
	return (ret);

	/* error cases */
	err1:
		return (0);
}

void
GUCS_pos_free(GUCS_cf_position *pos)
{
	if (pos->curpos != (GUCS_cf_buffer *) 0)
		GUCS_buf_unref(pos->curpos);
	free((void *) pos);
}

int
GUCS_pos_equal(GUCS_cf_position *p1, GUCS_cf_position *p2)
{
	if (p1->curpos == p2->curpos && p1->iter == p2->iter)
		return (1);
	return (0);
}

/* how much has been read between two positions? */
size_t
GUCS_pos_diff(GUCS_cf_position *p2, GUCS_cf_position *p1)
{
	return (p2->offset - p1->offset);
}

/* show the data in the file. */
int
GUCS_pos_peek(GUCS_cf_position *pos, void *data, size_t len)
{
	int ret;
	GUCS_cf_buffer *buf, *next;
	char *iter, *store;

	if (pos->curpos == (GUCS_cf_buffer *) 0)
		return (0);

	/* pull out the next bits of data from the buffer. */
	buf = pos->curpos;
	iter = pos->iter;
	store = (char *) data;

	GUCS_buf_ref(buf);
	for (ret = 0; ret < len; ret++)
	{
		store[ret] = *iter;
		iter++;
		if (iter == buf->line + buf->length)
		{
			/* need to skip to the next buffer. */
			next = GUCS_fseek(buf->cf, buf->offset + buf->length);
			GUCS_buf_unref(buf);
			if (next == (GUCS_cf_buffer *) 0)
				/* EOF. */
				return ret+1;
			buf = next;
			iter = buf->line;
		}
	}
	GUCS_buf_unref(buf);

	/* filled up the output buffer. */
	return (ret);
}

/* advance the current position by a length. */
void
GUCS_pos_advance(GUCS_cf_position *pos, int length)
{
	GUCS_cf_buffer *buf;

	buf = pos->curpos;
	pos->offset += length;
	while (length > 0)
	{
		length--;

		if (*pos->iter == '\n')
			pos->lineno++;
		pos->iter++;
		if (pos->iter == buf->line + buf->length)
		{
			pos->curpos = GUCS_fseek(buf->cf,
			                         buf->offset + buf->length);
			GUCS_buf_unref(buf);
			if (pos->curpos == (GUCS_cf_buffer *) 0 && length > 0)
			{
				fprintf(stderr, "d'oh!\n");
				exit(1);
			}
			buf = pos->curpos;
			if (buf != (GUCS_cf_buffer *) 0)
				pos->iter = buf->line;
		}
	}
}

/* have we hit EOF? */
int
GUCS_pos_eof(GUCS_cf_position *pos)
{
	char iter;

	if (GUCS_pos_peek(pos, &iter, 1) != 1)
		return (1);
	return (0);
}

GUCS_cf_expression *
GUCS_exp_alloc(GUCS_cf_position *start)
{
	GUCS_cf_expression *ret;

	ret = (GUCS_cf_expression *) malloc(sizeof(GUCS_cf_expression));
	if (ret == (GUCS_cf_expression *) 0)
		goto err1;
	ret->start = GUCS_pos_copy(start);
	if (ret->start == (GUCS_cf_position *) 0)
		goto err2;
	ret->cur = GUCS_pos_copy(start);
	if (ret->cur == (GUCS_cf_position *) 0)
		goto err3;
	return (ret);

	err3:
		GUCS_pos_free(ret->start);
	err2:
		free((void *) ret);
	err1:
		return (0);
}

void
GUCS_exp_free(GUCS_cf_expression *exp)
{
	GUCS_pos_free(exp->start);
	GUCS_pos_free(exp->cur);
	free((void *) exp);
}

GUCS_cf_position *
GUCS_exp_start(GUCS_cf_expression *exp)
{
	return (exp->start);
}

GUCS_cf_position *
GUCS_exp_cur(GUCS_cf_expression *exp)
{
	return (exp->cur);
}

void
GUCS_skip(GUCS_cf_position *pos, GUCS_parse_f parse)
{
	GUCS_error_type *err;
	GUCS_parse_el skip[] =
	{
		{ parse, 0, 0, 0, 0, 0 },
		{ 0, 0, 0, 0, 0, 0 },
	};

	err = GUCS_attempt_parse(pos, skip);
	if (err != (GUCS_error_type *) 0)
		gucs_err_unref(err);
}

/* mark as comment the rest of the line */
int
GUCS_parse_comment(GUCS_cf_expression *exp, int dlen, void *data,
                   int size, void **result, GUCS_error_type **err)
{
	char iter;

	if (GUCS_pos_peek(exp->cur, &iter, 1) != 1)
		return (-1);
	if (iter == '#')
	{
		while (iter != '\n')
		{
			GUCS_pos_advance(exp->cur, 1);
			if (GUCS_pos_peek(exp->cur, &iter, 1) != 1)
				return (0);
		}
		return (0);
	}
	return (-1);
}

/* mark as space the next bit of the line */
int
GUCS_parse_space(GUCS_cf_expression *exp, int dlen, void *data,
                 int size, void **result, GUCS_error_type **err)
{
	char iter, got;

	if (GUCS_pos_peek(exp->cur, &iter, 1) != 1)
		return (-1);
	got = 0;
	while (iter != '\n' && isspace(iter))
	{
		GUCS_pos_advance(exp->cur, 1);
		got++;
		if (GUCS_pos_peek(exp->cur, &iter, 1) != 1)
			return (0);
	}
	if (got == 0)
		return (-1);
	return (0);
}

/* mark as string the next bit of the line */
int
GUCS_parse_string(GUCS_cf_expression *exp, int dlen, void *data,
                  int size, void **result, GUCS_error_type **err)
{
	char iter, *store, term;
	GUCS_error_type *status;
	GUCS_cf_position *string;
	int len;

	if (GUCS_pos_peek(exp->cur, (void *) &iter, 1) != 1)
		return (-1);
	if (iter != '"' && iter != '\'')
		return (-1);
	term = iter;

	/* begins with '"', it's a string. */
	string = GUCS_pos_copy(exp->cur);
	GUCS_pos_advance(exp->cur, 1);
	if (GUCS_pos_peek(exp->cur, (void *) &iter, 1) != 1)
		goto err1;

	/* store our string */
	if (size != GUCS_PARSE_BEGIN)
	{
		store = (char *) malloc(size);
		*result = (void *) store;
	} else
		store = (char *) 0;
	len = 0;

	/* search for the closing quote. */
	while (iter != term)
	{
		GUCS_pos_advance(exp->cur, 1);
		if (iter == '\\')
		{
			/* escape the next character. */
			if (GUCS_pos_peek(exp->cur, (void *) &iter, 1) != 1)
				goto err1;
			if (store != (char *) 0)
				store[len] = iter;
			len++;

			GUCS_pos_advance(exp->cur, 1);
			if (GUCS_pos_peek(exp->cur, (void *) &iter, 1) != 1)
				goto err1;
		} else if (iter != '\n')
		{
			/* unescaped character. */
			if (store != (char *) 0)
				store[len] = iter;
			len++;
			if (GUCS_pos_peek(exp->cur, (void *) &iter, 1) != 1)
				goto err1;
		} else
			/* unescaped newline. */
			goto err1;
	}
	/* past close quote */
	GUCS_pos_advance(exp->cur, 1);

	/* terminate the string. */
	if (store != (char *) 0)
		store[len] = '\0';

	/* and done */
	GUCS_pos_free(string);
	return (len+1);

	err1:
		*err = GUCS_err_string(exp->start, string);
		GUCS_pos_free(string);
		return (-1);
}

/* mark as integer the next bit of the line */
int
GUCS_parse_number(GUCS_cf_expression *exp, int dlen, void *data,
                  int size, void **result, GUCS_error_type **err)
{
	char iter;
	long i;

	if (GUCS_pos_peek(exp->cur, &iter, 1) != 1)
		return (-1);
	i = 0;
	while (isdigit(iter))
	{
		i *= 10;
		i += (iter-'0');
		GUCS_pos_advance(exp->cur, 1);
		if (GUCS_pos_peek(exp->cur, &iter, 1) != 1)
		{
			if (result != (void *) 0)
				*((long *) result) = i;
			return (sizeof(long));
		}
	}
	if (iter != '\n' && ! isspace(iter))
		return (-1);

	if (size != GUCS_PARSE_BEGIN)
	{
		*result = (void *) malloc(size);
		*((long *) *result) = i;
	}
	return (sizeof(long));
}

/* mark as token the next part of the line. */
int
GUCS_parse_token(GUCS_cf_expression *exp, int dlen, void *data,
                 int size, void **result, GUCS_error_type **err)
{
	char iter;
	int len;
	char *store, *cmp;

	if (GUCS_pos_peek(exp->cur, &iter, 1) != 1)
		return (-1);
	/* XXX: currently define token as 'not a number, string, space, or
	 * comment'. */
	if (iter == '#' || iter == '"' || iter == '\'' ||
	  isdigit(iter) || isspace(iter))
		return (-1);

	if (size != GUCS_PARSE_BEGIN)
	{
		store = (char *) malloc(size);
		*result = (void *) store;
		cmp = (char *) 0;
	} else
	{
		store = (char *) 0;
		cmp = (char *) data;
	}
	len = 0;
	while (iter != '\n' && ! isspace(iter))
	{
		if (store != (char *) 0)
			store[len] = iter;
		else if (cmp != (char *) 0)
		{
			if (*cmp != iter)
				return (-1);
			cmp++;
			dlen--;
			if (dlen < 0)
				return (-1);
		}
		len++;
		GUCS_pos_advance(exp->cur, 1);
		if (GUCS_pos_peek(exp->cur, &iter, 1) != 1)
		{
			if (store != (char *) 0)
				store[len] = '\0';
			else if (cmp != (char *) 0 && dlen != 0)
				return (-1);
			return (len+1);
		}
	}
	if (store != (char *) 0)
		store[len] = '\0';
	else if (cmp != (char *) 0 && dlen != 0)
		return (-1);
	return (len+1);
}

/* clear out all temporarily held data for this parse iteration. */
void
GUCS_reset_parse(GUCS_parse_el *els)
{
	int i;

	for (i = 0; els[i].parse_f; i++)
		if (els[i].ret)
		{
			if (els[i].free_f)
				(*els[i].free_f)(els[i].ret_len, els[i].ret);
			else
				free((void *) els[i].ret);
			els[i].ret = 0;
		}
}

/* attempt to parse a line according to an expected format. */
GUCS_error_type *
GUCS_attempt_parse(GUCS_cf_position *pos, GUCS_parse_el *els)
{
	GUCS_error_type *ret, *code;
	GUCS_cf_expression exp;
	int i;

	exp.start = pos;
	exp.cur = GUCS_pos_copy(pos);
	if (exp.cur == (GUCS_cf_position *) 0)
		goto err1;

	ret = 0;
	/* validate the types of the line's data */
	for (i = 0; els[i].parse_f; i++)
	{
		els[i].ret_len = (*els[i].parse_f)(&exp, els[i].parse_data_len,
		                                   els[i].parse_data,
		                                   GUCS_PARSE_BEGIN, 0, &ret);
		els[i].ret = (void *) 0;
		if (els[i].ret_len < 0)
			goto err2;
	}
	GUCS_pos_free(exp.cur);

	exp.cur = GUCS_pos_copy(pos);
	if (exp.cur == (GUCS_cf_position *) 0)
		goto err1;

	/* fill in the real data */
	for (i = 0; els[i].parse_f; i++)
		(*els[i].parse_f)(&exp, els[i].parse_data_len,
		                  els[i].parse_data,
		                  els[i].ret_len, &els[i].ret, &ret);
	GUCS_pos_advance(pos, GUCS_pos_diff(exp.cur, pos));
	GUCS_pos_free(exp.cur);
	return (0);

	err2:
		GUCS_reset_parse(els);
		if (ret == (GUCS_error_type *) 0)
			ret = GUCS_err_no_match(pos, pos);
		GUCS_pos_free(exp.cur);
	err1:
		return (ret);
}

/* attempt to parse a line according to a variety of syntaxes, return which
 * syntax worked. */
GUCS_error_type *
GUCS_parse(GUCS_cf_position *pos, GUCS_parse_el **arr, int *which)
{
	GUCS_error_type *ret;
	int i;

	for (i = 0; arr[i]; i++)
	{
		ret = GUCS_attempt_parse(pos, arr[i]);
		if (ret == (GUCS_error_type *) 0)
		{
			*which = i;
			return (0);
		} else if (gucs_error_code(ret, 0) != GUCS_ERR_PARSE ||
		  gucs_error_code(ret, 1) != GUCS_ERR_PARSE_NOMATCH)
			return (ret);
		else
			gucs_err_unref(ret);
	}
	return (GUCS_err_no_match(pos, pos));
}

/* skip spaces and comments until we reach EOLN */
int
GUCS_seek_eoln(GUCS_cf_position *pos)
{
	GUCS_cf_expression exp;
	char iter;

	if (GUCS_pos_peek(pos, &iter, 1) != 1)
		/* EOF.  Treat like EOLN. */
		return (0);

	exp.start = (GUCS_cf_position *) 0;	/* doesn't throw exceptions */
	exp.cur = pos;
	while (iter != '\n')
	{
		if (GUCS_parse_space(&exp, 0, 0, GUCS_PARSE_BEGIN, 0, 0) < 0 &&
		  GUCS_parse_comment(&exp, 0, 0, GUCS_PARSE_BEGIN, 0, 0) < 0)
			return (-1);
		if (GUCS_pos_peek(exp.cur, &iter, 1) != 1)
			/* EOF. */
			return (0);
	}
	/* iter == '\n' */
	GUCS_pos_advance(exp.cur, 1);
	return (0);
}

int
GUCS_parse_to_eoln(GUCS_cf_expression *exp, int dlen, void *data,
                   int size, void **result, GUCS_error_type **err)
{
	return (GUCS_seek_eoln(exp->cur));
}

/* for generated code */
GUCS_error_type *
GUCS_include(GUCS_cf *cf, GUCS_realm_parser parser, void *data, void *state)
{
	GUCS_cf_position *pos;
	GUCS_error_type *ret;

	pos = GUCS_file_start(cf);
	if (pos == (GUCS_cf_position *) 0)
	{
		ret = gucs_err_errno(errno);
		goto err1;
	}

	ret = (*parser)(pos, data, 1, state);
	if (ret != (GUCS_error_type *) 0)
		goto err2;
	if (! GUCS_pos_eof(pos))
		ret = GUCS_err_no_eof(pos, pos);

	/*FALLTHROUGH*/
	err2:
		GUCS_pos_free(pos);
	err1:
		return (ret);
}

/* error functions */
static void
GUCS_err_pos_cleanup(GUCS_error_type *err, char *name, void *data)
{
	GUCS_pos_free(data);
}

GUCS_error_type *
GUCS_parse_err(GUCS_cf_position *start, GUCS_cf_position *pos, int code)
{
	GUCS_error_type *ret, *status;
	GUCS_cf_position *pos_copy;

	ret = gucs_err_alloc();
	if (ret == (GUCS_error_type *) 0)
	{
		status = gucs_err_no_mem();
		goto err1;
	}

	status = gucs_err_code_set(ret, GUCS_ERR_PARSE, code, -1);
	if (status != (GUCS_error_type *) 0)
		goto err2;

	pos_copy = GUCS_pos_copy(pos);
	if (pos_copy == (GUCS_cf_position *) 0)
		goto err2;
	status = gucs_err_data_set(ret, GUCS_ERR_DATA_POS, 0,
	                           GUCS_err_pos_cleanup, (void *) pos_copy);
	if (status != (GUCS_error_type *) 0)
		goto err3;

	pos_copy = GUCS_pos_copy(start);
	if (pos_copy == (GUCS_cf_position *) 0)
		goto err2;
	status = gucs_err_data_set(ret, GUCS_ERR_DATA_START, 0,
	                           GUCS_err_pos_cleanup, (void *) pos_copy);
	if (status != (GUCS_error_type *) 0)
		goto err3;

	return (ret);

	err3:
		GUCS_pos_free(pos_copy);
	err2:
		gucs_err_unref(ret);
	err1:
		return (status);
}

static GUCS_error_type *
GUCS_err_string(GUCS_cf_position *start, GUCS_cf_position *pos)
{
	return (GUCS_parse_err(start, pos, GUCS_ERR_PARSE_STRING));
}

static GUCS_error_type *
GUCS_err_no_match(GUCS_cf_position *start, GUCS_cf_position *pos)
{
	return (GUCS_parse_err(start, pos, GUCS_ERR_PARSE_NOMATCH));
}

static GUCS_error_type *
GUCS_err_no_eof(GUCS_cf_position *start, GUCS_cf_position *pos)
{
	return (GUCS_parse_err(start, pos, GUCS_ERR_PARSE_NO_EOF));
}
